/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _ALGORITHMS_H
#define _ALGORITHMS_H 1

#include <complex>
#include <vector>
#include <list>
#include <cfloat>
#include <stdexcept>
#include <map>

#include "matrix_template.h"
#include "layer.h"

const std::complex<double> kEPSILON_0 = std::complex<double>(8.854187816e-12, 0.0); /*!< free space permittivity in farads per m ((s^4 * A^2) / (m^3 * kg)) */
const std::complex<double> kMU_0 = std::complex<double>(4e-7 * M_PI, 0.0); /*!< free space permeability in henry per m ((m * kg) / (s^2 * A^2)) */
const double kC = 299792458; /*!< speed of light in m / s */
const double nodes[] = {-0.991455371120813, -0.949107912342758, \
	                      -0.864864423359769, -0.741531185599394, \
												-0.586087235467691, -0.405845151377397, \
												-0.207784955007898,  0.000000000000000, \
												 0.207784955007898,  0.405845151377397, \
												 0.586087235467691,  0.741531185599394, \
												 0.864864423359769,  0.949107912342758, \
												 0.991455371120813}; /*!< 15 point Gauss-Kronrod rule nodes */
const double kronrod15Weigths[] = {0.022935322010529, 0.063092092629979, \
                                   0.104790010322250, 0.140653259715525, \
																	 0.169004726639267, 0.190350578064785, \
																	 0.204432940075298, 0.209482141084728, \
																	 0.204432940075298, 0.190350578064785, \
																	 0.169004726639267, 0.140653259715525, \
																	 0.104790010322250, 0.063092092629979, \
																	 0.022935322010529}; /*!< 15 point Kronrod rule weights */
const double gauss7Weigths[] = {0.129484966168870, 0.279705391489277, 0.381830050505119, \
                                0.417959183673469, 0.381830050505119, 0.279705391489277, \
																0.129484966168870}; /*!< 7 point Gauss rule weigths */


//! Structure holding an interval or region.
template<class numT>
struct interval
{
	numT start_T;	/*!< start value of the interval */
	numT end_T;	/*!< end value of the interval */
};


//! Enumeration holding information about a part of a complex number.
enum complexNumberPart
{
	kREAL_PART, /*!< Enum value representing the real part of a complex number. */
	kIMAG_PART /*!< Enum value representing the imaginary part of complex number. */
};

//! Structure holding the information required by the integration algorithm to
//integrate the characteristic equation.
struct characteristicEquationIntegrationParameters
{
	const unsigned int& modeOrder_uint;
	std::vector<layerDescription<double> >& layers_vector;
	complexNumberPart variable_complexNumberPart;
};

//! Structure holding the information required by the path integration algorithm.
template <class numT>
struct pathIntegralParameters
{
	std::complex<numT> (*pathIntFunction)(const std::complex<numT>& arg, void* param); /*!< pointer to the function to be integrated */
	void* pathIntFunctionParam; /*!< pointer to the argument of the function to be integrated */
	std::complex<numT> startPoint_complexDouble; /*!< start point of the path */
	std::complex<numT> pathDeriv_complexDouble; /*!< derivative of the path function */
	const numT& eps; /*!< precision */
};

//! Structure to hold an interval for the Gauss-Kronrod integration
template <class numT>
struct gkInterval
{
	numT start; /*!< start of the interval */
	numT end; /*!< end of the interval */
	std::complex<numT> gauss7; /*!< Gauss estimate */
	std::complex<numT> kronrod15; /*!< Kronrod estimate */
};

//! Integration function based on a 7/15 point Gauss-Kronrod rule.
/*!
 * The function is based on 15 point Kronrod and 7 point Gauss rule. The
 * algorithm is globally adaptive. The required nodes and weights were obtained
 * from Emporia State University,
 * http://mathcsjava.emporia.edu/oneilcat/ExperimentApplet3/what_is_gauss.html,
 * available 12 October 2007.
 *
 * \param intFunc function to be integrated
 * \param param parameters for the function to be integrated
 * \param a start value of the intergration interval
 * \param b end value of the integration interval
 * \param check error flag (0 = no errors, 1 = error)
 * \param tol tolerance
 * \param maxIterations maximum number of iterations
 * \sa gauss7_kronrod15_estimate
 */
template <class numT>
std::complex<numT> integral(std::complex<numT>(*intFunc)(const numT& x, void* param), void* param, const numT& a, const numT& b, unsigned int& check, numT tol = static_cast<numT>(1e-14), const unsigned int& maxIterations = static_cast<unsigned int>(INFINITY))
{
	if (tol < static_cast<numT>(DBL_EPSILON))
		tol = static_cast<numT>(DBL_EPSILON);

	std::complex<numT> gauss7, kronrod15;
	numT improvementRel = NAN;
	numT min_double = std::min(a, b);
	numT max_double = std::max(a, b);
	unsigned int maxIterations_uint = maxIterations;
	std::multimap<numT, gkInterval<numT> > interval_map;
	check = 0;
	
	gauss7_kronrod15_estimate(intFunc, param, min_double, max_double, gauss7, kronrod15, check);
	numT errorTemp = static_cast<numT>(200.0) * std::abs(gauss7 - kronrod15);
	numT error = errorTemp * std::sqrt(errorTemp);
	interval_map.insert(std::pair<numT, gkInterval<numT> >(error, (gkInterval<numT>){min_double, max_double, gauss7, kronrod15}));
	
	while ((std::isnan(improvementRel) || ((improvementRel * std::sqrt(interval_map.size())) > tol)) && (maxIterations_uint > 0) && (interval_map.size() > 0))
	{
		gkInterval<numT> worstInterval = (--(interval_map.end()))->second;
		
		gkInterval<numT> interval1 = {worstInterval.start, worstInterval.start + (worstInterval.end - worstInterval.start) / static_cast<numT>(2.0), static_cast<numT>(0.0), static_cast<numT>(0.0)};
		gkInterval<numT> interval2 = {worstInterval.start + (worstInterval.end - worstInterval.start) / static_cast<numT>(2.0), worstInterval.end, static_cast<numT>(0.0), static_cast<numT>(0.0)};
		
		gauss7_kronrod15_estimate(intFunc, param, interval1.start, interval1.end, interval1.gauss7, interval1.kronrod15, check);
		gauss7_kronrod15_estimate(intFunc, param, interval2.start, interval2.end, interval2.gauss7, interval2.kronrod15, check);
		
		// Update global Kronrod estimates and the improvement estimate
		std::complex<numT> tempKronrod15 = kronrod15;
		kronrod15 -= worstInterval.kronrod15;
		kronrod15 += interval1.kronrod15;
		kronrod15 += interval2.kronrod15;
		improvementRel = std::abs((tempKronrod15 - kronrod15) / kronrod15);

		// Update the interval map			
		interval_map.erase(--(interval_map.end()));
		if ((interval1.start - interval1.end) / static_cast<numT>(20.0) + interval1.start != interval1.start)
		{
			numT error1Temp = static_cast<numT>(200.0) * std::abs(interval1.gauss7 - interval1.kronrod15);
			numT error1 = error1Temp * std::sqrt(error1Temp);
			interval_map.insert(std::pair<numT, gkInterval<numT> >(error1, interval1));
		}
		else
			check = 1;

		if ((interval2.start - interval2.end) / static_cast<numT>(20.0) + interval2.start != interval2.start)
		{
			numT error2Temp = static_cast<numT>(200.0) * std::abs(interval2.gauss7 - interval2.kronrod15);
			numT error2 = error2Temp * std::sqrt(error2Temp);
			interval_map.insert(std::pair<numT, gkInterval<numT> >(error2, interval2));
		}
		else
			check = 1;
	}
	
	return kronrod15;
}

//! Approximation of an interval of an integral using a 7/15 point Guass-Kronrod estimate.
/*!
 * \sa integral
 */
template <class numT>
void gauss7_kronrod15_estimate(std::complex<numT>(*intFunc)(const numT& x, void* param), void* param, const numT& a, const numT& b, std::complex<numT>& gauss7, std::complex<numT>& kronrod15, unsigned int& check)
{
	numT bmao2 = (b - a) / static_cast<numT>(2.0);
	numT bpao2 = (b + a) / static_cast<numT>(2.0);
	
	gauss7 = static_cast<numT>(0.0);
	kronrod15 = static_cast<numT>(0.0);
	
	for (unsigned int counter = 0; counter < 15; counter++)
	{
		std::complex<numT> functionValue = bmao2 * intFunc(bmao2 * static_cast<numT>(nodes[counter]) + bpao2, param);
		if (counter % 2)
			gauss7 += static_cast<numT>(gauss7Weigths[(counter - 1) / 2]) * functionValue;
		
		kronrod15 += static_cast<numT>(kronrod15Weigths[counter]) * functionValue;
	}
}

//! Subfunction for the integration of a straight path.
/*!
 * \sa straight_path_integral
 */
template <class numT>
std::complex<numT> straight_path_integral_subfunction(const numT& arg, void* param)
{
	pathIntegralParameters<numT> parameters = *(pathIntegralParameters<numT>*)param;

	std::complex<numT> temp = (*parameters.pathIntFunction)(parameters.startPoint_complexDouble + arg * parameters.pathDeriv_complexDouble, parameters.pathIntFunctionParam) * parameters.pathDeriv_complexDouble;

	if (std::isnan(std::real(temp)) || std::isnan(std::imag(temp)))
		throw std::runtime_error("straight_paht_intergral_subfunction: Function value is NAN");

	return temp;
}

//! Integration of a straight path in complex space.
/*!
 * This function takes two points in complex space and integrates a function
 * along a line between these two points. Integration is done by
 * parameterisation.
 *
 * \sa straight_path_integral_subfunction
 */
template <class numT>
std::complex<numT> straight_path_integral(std::complex<numT> (*pathIntFunction)(const std::complex<numT>& arg, void* param), void* pathIntFunctionParam, const std::complex<numT>& a, const std::complex<numT>& b, unsigned int& check, const numT& eps = static_cast<numT>(1e-20), const unsigned int& maxIterations = static_cast<unsigned int>(INFINITY))
{
	pathIntegralParameters<numT> param = {pathIntFunction, pathIntFunctionParam, a, b - a, eps};
	return integral(&straight_path_integral_subfunction, &param, static_cast<numT>(0.0), static_cast<numT>(1.0), check, eps, maxIterations); // This is the required parameterisation.
}

//! Function to yield the number of roots of a complex function in a given interval.
/*!
 * The function is based on an algorithm described in M. Dellnitz, O. Schütze,
 * Q. Zheng, "Locating all the zeros of an analytic function in one complex
 * variable," Journal of Computational and Applied Mathematics, 138, 325-333
 * (2002).
 *
 * \sa straight_path_integral
 */
template <class numT>
numT root_test(std::complex<numT> (*pathIntFunction)(const std::complex<numT>& arg, void* param), void* pathIntFunctionParam, const interval<std::complex<numT> >& beta_interval, const numT& eps = static_cast<numT>(1e-15))
{
	std::complex<numT> min_complexDouble(std::min(std::real(beta_interval.start_T), std::real(beta_interval.end_T)), std::min(std::imag(beta_interval.start_T), std::imag(beta_interval.end_T)));
	std::complex<numT> max_complexDouble(std::max(std::real(beta_interval.start_T), std::real(beta_interval.end_T)), std::max(std::imag(beta_interval.start_T), std::imag(beta_interval.end_T)));
	std::complex<numT> factor_complexDouble(static_cast<numT>(0.0), static_cast<numT>(2.0 * M_PI));

	std::complex<numT> corner1_complexDouble(min_complexDouble);
	std::complex<numT> corner2_complexDouble(std::real(max_complexDouble), std::imag(min_complexDouble));
	std::complex<numT> corner3_complexDouble(max_complexDouble);
	std::complex<numT> corner4_complexDouble(std::real(min_complexDouble), std::imag(max_complexDouble));

	numT tol1_double = eps;
	numT tol2_double = eps;
	numT tol3_double = eps;
	numT tol4_double = eps;

	unsigned int check1 = 0, check2 = 0, check3 = 0, check4 = 0;
	unsigned int maxIterations = static_cast<unsigned int>(INFINITY); //100;
	numT integralOld_double = static_cast<numT>(-1.0);
	numT integralNew_double = static_cast<numT>(0.0);
	numT int1_double = static_cast<numT>(NAN);
	numT int2_double = static_cast<numT>(NAN);
	numT int3_double = static_cast<numT>(NAN);
	numT int4_double = static_cast<numT>(NAN);
	numT scaleFactor_double = static_cast<numT>(1.0);
			
	int1_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner1_complexDouble, corner2_complexDouble, check1, tol1_double, maxIterations) / factor_complexDouble);
	int2_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner2_complexDouble, corner3_complexDouble, check2, tol2_double, maxIterations) / factor_complexDouble);
	int3_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner3_complexDouble, corner4_complexDouble, check3, tol3_double, maxIterations) / factor_complexDouble);
	int4_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner4_complexDouble, corner1_complexDouble, check4, tol4_double, maxIterations) / factor_complexDouble);

	while (std::isnan(int1_double) || std::isnan(int2_double) || std::isnan(int3_double) || std::isnan(int4_double))
	{
//		scaleFactor_double *= 0.9999;
		scaleFactor_double *= static_cast<numT>(1.1);

		min_complexDouble /= scaleFactor_double;
		max_complexDouble *= scaleFactor_double;
		
		corner1_complexDouble = std::complex<numT>(min_complexDouble);
		corner2_complexDouble = std::complex<numT>(std::real(max_complexDouble), std::imag(min_complexDouble));
		corner3_complexDouble = std::complex<numT>(max_complexDouble);
		corner4_complexDouble = std::complex<numT>(std::real(min_complexDouble), std::imag(max_complexDouble));

		int1_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner1_complexDouble, corner2_complexDouble, check1, tol1_double, maxIterations) / factor_complexDouble);
		int2_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner2_complexDouble, corner3_complexDouble, check2, tol2_double, maxIterations) / factor_complexDouble);
		int3_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner3_complexDouble, corner4_complexDouble, check3, tol3_double, maxIterations) / factor_complexDouble);
		int4_double = std::real(straight_path_integral(pathIntFunction, pathIntFunctionParam, corner4_complexDouble, corner1_complexDouble, check4, tol4_double, maxIterations) / factor_complexDouble);
	}
	
	integralNew_double = int1_double + int2_double + int3_double + int4_double;
	
	return integralNew_double;
}

//! Function to determine a single root of a complex function in a given interval.
/*!  
 * This function uses two different root finding algorithms to try an find a
 * single root of a complex function in a given interval.
 *
 * \sa newton
 * \sa barlow
 */
template <class numT>
std::complex<numT> determine_root_in_interval(interval<std::complex<numT> > nEff_interval, void (*fun)(const std::complex<numT>&, void*, std::complex<numT>*, std::complex<numT>*), void* param)
{
	std::complex<numT> nEff;

	try
	{
		nEff = newton(nEff_interval, fun, param);
	}
	catch (...)
	{
		nEff = barlow(nEff_interval, fun, param);
	}
	
	return nEff;
}

//! Function to find the root of a complex function based on Newton's method.
/*!
 * The function uses Netwon's method as described in E.W. Weisstein, "Newton's
 * Method," MathWorld, http://mathworld.wolfram.com/NewtonsMethod.html,
 * available 12 October 2007, with nine semi-random starting points.
 *
 * \sa determine_root_in_interval
 */
template <class numT>
std::complex<numT> newton(const interval<std::complex<numT> >& interval, void (*fun)(const std::complex<numT>&, void*, std::complex<numT>*, std::complex<numT>*), void* param, const numT& threshold = static_cast<numT>(1e-15)) throw (std::runtime_error)
{
	std::complex<numT> xDiff_complexDouble(interval.end_T - interval.start_T);
	std::complex<numT> xOut_complexDouble;
	std::complex<numT> f_complexDouble;
	std::complex<numT> fDiff_complexDouble;
	std::complex<numT> ratio_complexDouble(static_cast<numT>(1.0), static_cast<numT>(1.0));
	numT threshold_double = threshold;
	numT realUpper(std::max(std::real(interval.start_T), std::real(interval.end_T))), realLower(std::min(std::real(interval.start_T), std::real(interval.end_T)));
	numT imagUpper(std::max(std::imag(interval.start_T), std::imag(interval.end_T))), imagLower(std::min(std::imag(interval.start_T), std::imag(interval.end_T)));
//	double realPoints_doublePtr[9] = {1.00, 0.50, 0.00, 1.00, 0.50, 0.00, 1.00, 0.50, 0.00};
//	double imagPoints_doublePtr[9] = {0.00, 0.00, 0.00, 0.50, 0.50, 0.50, 1.00, 1.00, 1.00};
	numT realPoints_doublePtr[9] = {static_cast<numT>(0.48), static_cast<numT>(0.97), \
                                  static_cast<numT>(0.53), static_cast<numT>(0.01), \
																	static_cast<numT>(0.99), static_cast<numT>(0.04), \
																	static_cast<numT>(0.98), static_cast<numT>(0.47), \
																	static_cast<numT>(0.03)};
	numT imagPoints_doublePtr[9] = {static_cast<numT>(0.51), static_cast<numT>(0.02), \
                                  static_cast<numT>(0.04), static_cast<numT>(0.03), \
																	static_cast<numT>(0.51), static_cast<numT>(0.48), \
																	static_cast<numT>(0.96), static_cast<numT>(0.98), \
																	static_cast<numT>(0.94)};
	unsigned int maxIterations_uint = 50;
//	unsigned int maxIterations_uint = INFINITY;

	for (unsigned int pointCounter_uint = 0; pointCounter_uint < 9; pointCounter_uint++)
	{
		unsigned int realUpper_uint = 0;
		unsigned int realLower_uint = 0;
		unsigned int imagUpper_uint = 0;
		unsigned int imagLower_uint = 0;
		xOut_complexDouble = std::complex<numT>(std::real(interval.start_T) + std::real(xDiff_complexDouble) * realPoints_doublePtr[pointCounter_uint], std::imag(interval.start_T) + std::imag(xDiff_complexDouble) * imagPoints_doublePtr[pointCounter_uint]);
		maxIterations_uint = 200;
		
		while ((std::abs(ratio_complexDouble / xOut_complexDouble) > threshold_double) && (maxIterations_uint > 0))
		{
			(*fun)(xOut_complexDouble, param, &f_complexDouble, &fDiff_complexDouble);
			
			ratio_complexDouble = (f_complexDouble / fDiff_complexDouble);
			xOut_complexDouble -= ratio_complexDouble;

			// Bound checking
			if (std::real(xOut_complexDouble) > realUpper)
			{
				if (realUpper_uint == 0)
				{
					xOut_complexDouble = std::complex<numT>(realUpper, std::imag(xOut_complexDouble));
					realUpper_uint = 1;
				}
				else
					xOut_complexDouble = std::complex<numT>(realLower, std::imag(xOut_complexDouble));
			}
			else if (std::real(xOut_complexDouble) < realLower)
			{
				if (realLower_uint == 0)
				{
					xOut_complexDouble = std::complex<numT>(realLower, std::imag(xOut_complexDouble));
					realLower_uint = 1;
				}
				else
					xOut_complexDouble = std::complex<numT>(realUpper, std::imag(xOut_complexDouble));
			}	
			if (std::imag(xOut_complexDouble) > imagUpper)
			{
				if (imagUpper_uint == 0)
				{
					xOut_complexDouble = std::complex<numT>(std::real(xOut_complexDouble), imagUpper);
					imagUpper_uint = 1;
				}
				else
					xOut_complexDouble = std::complex<numT>(std::real(xOut_complexDouble), imagLower);
			}
			else if (std::real(xOut_complexDouble) < imagLower)
			{
				if (imagLower_uint == 0)
				{
					xOut_complexDouble = std::complex<numT>(std::real(xOut_complexDouble), imagLower);
					imagLower_uint = 1;
				}
				else
					xOut_complexDouble = std::complex<numT>(std::real(xOut_complexDouble), imagUpper);			
			}
			maxIterations_uint--;
		}
		if (maxIterations_uint != 0)
			return xOut_complexDouble;
	}
	if (maxIterations_uint == 0)
	{
// 		if (this->_streams_map["log"] != NULL)
// 		{
// 			this->_streams_map["log"]->lock();
// 			*(this->_streams_map["log"]->getObjectPtr()) << "HHFibre::newton: problem with root finding: " << xOut_complexDouble / this->_k0Scaled_double << "\n";
// 			this->_streams_map["log"]->unlock();
// 		}
 		throw std::runtime_error("HHFibre:newton: No root found.");
	}
	
	return xOut_complexDouble;
}

//! Function to determine the root of a complex function using the secant method.
/*!
 * The function uses the secant method as described in E.W. Weisstein, "Secant
 * Method," MathWorld, http://mathworld.wolfram.com/SecantMethod.html,
 * available 12 October 2007, with nine semi-random starting points.
 *
 * \sa determine_root_in_interval
 */
template <class numT>
std::complex<numT> barlow(const interval<std::complex<numT> >& interval, void (*fun)(const std::complex<numT>&, void*, std::complex<numT>*, std::complex<numT>*), void* param, const numT& threshold = static_cast<numT>(1e-15))
{
	numT imagMax_double = std::max(std::imag(interval.start_T), std::imag(interval.end_T));
	numT imagMin_double = std::min(std::imag(interval.start_T), std::imag(interval.end_T));
	numT realMax_double = std::max(std::real(interval.start_T), std::real(interval.end_T));
	numT realMin_double = std::min(std::real(interval.start_T), std::real(interval.end_T));

//	double imagStart_doublePtr[8] = {0.1, 0.5, 0.9, 0.1, 0.9, 0.1, 0.5, 0.9};
//	double realStart_doublePtr[8] = {0.1, 0.1, 0.1, 0.5, 0.5, 0.9, 0.9, 0.9};
//	double imagEnd_doublePtr[8] =   {0.4, 0.5, 0.6, 0.4, 0.6, 0.4, 0.5, 0.6};
//	double realEnd_doublePtr[8] =   {0.4, 0.4, 0.4, 0.5, 0.5, 0.6, 0.6, 0.6};

	numT imagStart_doublePtr[8] = {static_cast<numT>(0.94), static_cast<numT>(0.55), \
                                 static_cast<numT>(0.93), static_cast<numT>(0.08), \
																 static_cast<numT>(0.87), static_cast<numT>(0.12), \
																 static_cast<numT>(0.48), static_cast<numT>(0.92)};
	numT realStart_doublePtr[8] = {static_cast<numT>(0.08), static_cast<numT>(0.07), \
                                 static_cast<numT>(0.13), static_cast<numT>(0.46), \
																 static_cast<numT>(0.54), static_cast<numT>(0.92), \
																 static_cast<numT>(0.86), static_cast<numT>(0.91)};
	numT imagEnd_doublePtr[8] =   {static_cast<numT>(0.36), static_cast<numT>(0.45), \
                                 static_cast<numT>(0.64), static_cast<numT>(0.44), \
																 static_cast<numT>(0.62), static_cast<numT>(0.38), \
																 static_cast<numT>(0.53), static_cast<numT>(0.62)};
	numT realEnd_doublePtr[8] =   {static_cast<numT>(0.39), static_cast<numT>(0.38), \
                                 static_cast<numT>(0.44), static_cast<numT>(0.48), \
																 static_cast<numT>(0.51), static_cast<numT>(0.59), \
																 static_cast<numT>(0.62), static_cast<numT>(0.57)};

	unsigned int rootFound_uint = 0;
	std::complex<numT> root_complexDouble(static_cast<numT>(NAN), static_cast<numT>(NAN));

	for (unsigned int pointCounter_uint = 0; (pointCounter_uint < 8) && (rootFound_uint == 0); pointCounter_uint++)
	{
		unsigned int boundaryProblem_uint = 0;
		std::complex<numT> x1(std::complex<numT>(realStart_doublePtr[pointCounter_uint] * (realMax_double - realMin_double) + realMin_double, imagStart_doublePtr[pointCounter_uint] * (imagMax_double - imagMin_double) + imagMin_double));
		std::complex<numT> y1;
		(*fun)(x1, param, &y1, NULL);
		std::complex<numT> x2(std::complex<numT>(realEnd_doublePtr[pointCounter_uint] * (realMax_double -realMin_double) + realMin_double, imagEnd_doublePtr[pointCounter_uint] * (imagMax_double - imagMin_double) + imagMin_double));
		std::complex<numT> y2;
		(*fun)(x2, param, &y2, NULL);

		unsigned int loopCounter_uint;
		for (loopCounter_uint = 0; (loopCounter_uint < 200) && (boundaryProblem_uint < 7) && (rootFound_uint == 0) && !std::isnan(std::real(x2)) && !std::isnan(std::imag(x2)); loopCounter_uint++)
		{
			std::complex<numT> change_complexDouble = (x2 - x1) * y2 / (y2 - y1);
			std::complex<numT> temp = x2 - change_complexDouble;
			
			// Threshold check
			if ((std::abs(std::log(y1/y2)) < static_cast<numT>(25.0)) && (std::abs(std::imag(change_complexDouble)) < (threshold * std::abs(temp))) && (std::abs(std::real(change_complexDouble)) < (threshold * std::abs(temp))) && !std::isnan(std::real(change_complexDouble)) && !std::isnan(std::imag(change_complexDouble)))
			{
				rootFound_uint = 1;
				root_complexDouble = temp;
			}

			x1 = x2;
			y1 = y2;
			x2 = temp;

			// Boundary check
			if (std::real(x2) > (realMax_double + threshold))
			{
				boundaryProblem_uint++;
				x2 = std::complex<numT>(realMin_double, std::imag(x2));
				rootFound_uint = 0;
				root_complexDouble = std::complex<numT>(static_cast<numT>(NAN), static_cast<numT>(NAN));
			}
			else if (std::real(x2) < (realMin_double - threshold))
			{
				boundaryProblem_uint++;
				x2 = std::complex<numT>(realMax_double, std::imag(x2));
				rootFound_uint = 0;
				root_complexDouble = std::complex<numT>(static_cast<numT>(NAN), static_cast<numT>(NAN));
			}
			if (std::imag(x2) > (imagMax_double + threshold))
			{
				boundaryProblem_uint++;
				x2 = std::complex<numT>(std::real(x2), imagMin_double);
				rootFound_uint = 0;
				root_complexDouble = std::complex<numT>(static_cast<numT>(NAN), static_cast<numT>(NAN));
			}
			else if (std::imag(x2) < (imagMin_double - threshold))
			{
				boundaryProblem_uint++;
				x2 = std::complex<numT>(std::real(x2), imagMax_double);
				rootFound_uint = 0;
				root_complexDouble = std::complex<numT>(static_cast<numT>(NAN), static_cast<numT>(NAN));
			}
			(*fun)(x2, param, &y2, NULL);
			
			if (std::abs(y2) == 0.0)
			{
				rootFound_uint = 1;
				root_complexDouble = x2;
			}
		}
	}
	
	if (rootFound_uint == 1)
		return root_complexDouble;
	else
		return std::complex<numT>(static_cast<numT>(NAN), static_cast<numT>(NAN));
}

//! Function to split an interval into a number of subintervals.
/*!
 * The longer side of the complex interval is split into the number of
 * subintervals given. If the shorter side is longer than the new interval side
 * it is split as well.
 */
template <class numT>
std::list<interval<std::complex<numT> > > split_interval(const unsigned int& divisions, const interval<std::complex<numT> >& input_interval)
{
	std::complex<numT> diff_complexDouble(input_interval.end_T - input_interval.start_T);
	std::list<interval<std::complex<numT> > > output_list;
	unsigned int divisionsLarge_uint = divisions;
	
	if (divisionsLarge_uint < 1)
		divisionsLarge_uint = 1;
	
	if (std::abs(std::real(diff_complexDouble)) > std::abs(std::imag(diff_complexDouble)))
	{
		numT divisionSize_double = std::abs(std::real(diff_complexDouble)) / static_cast<numT>(divisionsLarge_uint);
		unsigned int divisionsSmall_uint = static_cast<unsigned int>(intValue(floor(std::abs(std::abs(std::imag(diff_complexDouble)) - static_cast<numT>(0.1) * divisionSize_double) / divisionSize_double))) + 1;
		
		for (unsigned int counterLarge_uint = 0; counterLarge_uint < divisionsLarge_uint; counterLarge_uint++)
		{
			for (unsigned int counterSmall_uint = 0; counterSmall_uint < divisionsSmall_uint; counterSmall_uint++)
			{
				interval<std::complex<numT> > temp_interval;
				temp_interval.start_T = std::complex<numT>(std::real(input_interval.start_T) + static_cast<numT>(counterLarge_uint) / static_cast<numT>(divisionsLarge_uint) * std::real(diff_complexDouble), std::imag(input_interval.start_T) + static_cast<numT>(counterSmall_uint) / static_cast<numT>(divisionsSmall_uint) * (std::imag(diff_complexDouble)));
				temp_interval.end_T = std::complex<numT>(std::real(input_interval.start_T) + static_cast<numT>(counterLarge_uint + 1) / static_cast<numT>(divisionsLarge_uint) * std::real(diff_complexDouble), std::imag(input_interval.start_T) + static_cast<numT>(counterSmall_uint + 1) / static_cast<numT>(divisionsSmall_uint) * (std::imag(diff_complexDouble)));
				output_list.push_back(temp_interval);
			}
		}
	}
	else
	{
		numT divisionSize_double = std::abs(std::imag(diff_complexDouble)) / static_cast<numT>(divisionsLarge_uint);
		unsigned int divisionsSmall_uint = static_cast<unsigned int>(intValue(floor(std::abs(std::abs(std::real(diff_complexDouble)) - static_cast<numT>(0.1) * divisionSize_double) / divisionSize_double))) + 1;
		
		for (unsigned int counterLarge_uint = 0; counterLarge_uint < divisionsLarge_uint; counterLarge_uint++)
		{
			for (unsigned int counterSmall_uint = 0; counterSmall_uint < divisionsSmall_uint; counterSmall_uint++)
			{
				interval<std::complex<numT> > temp_interval;
				temp_interval.start_T = std::complex<numT>(std::real(input_interval.start_T) + static_cast<numT>(counterSmall_uint) / static_cast<numT>(divisionsSmall_uint) * std::real(diff_complexDouble), std::imag(input_interval.start_T) + static_cast<numT>(counterLarge_uint) / static_cast<numT>(divisionsLarge_uint) * (std::imag(diff_complexDouble)));
				temp_interval.end_T = std::complex<numT>(std::real(input_interval.start_T) + static_cast<numT>(counterSmall_uint + 1) / static_cast<numT>(divisionsSmall_uint) * std::real(diff_complexDouble), std::imag(input_interval.start_T) + static_cast<numT>(counterLarge_uint + 1) / static_cast<numT>(divisionsLarge_uint) * (std::imag(diff_complexDouble)));
				output_list.push_back(temp_interval);
			}
		}
	}

	return output_list;
}

//! Function to test if two values are approximately equal.
template <class numT>
bool thresholdEqual(const numT& num1, const numT& num2, const numT& threshold = static_cast<numT>(1e-15))
{
	if (num1 == static_cast<numT>(0.0))
		if (std::abs(num2) < threshold)
			return true;
		else
			return false;

	if (num2 == static_cast<numT>(0.0))
		if (std::abs(num1) < threshold)
			return true;
		else
			return false;
	
	numT ratio = std::abs(num1 / num2);
	if (ratio == static_cast<numT>(1.0))
		return true;
	else if (ratio < static_cast<numT>(1.0))
		ratio = static_cast<numT>(1.0) - ratio;
	else if (ratio > static_cast<numT>(1.0))
		ratio = ratio - static_cast<numT>(1.0);
	
	if (ratio < threshold)
		return true;
	else
		return false;
}

//! Function to test if two values are approximately equal.
template <class numT>
bool thresholdEqual(const std::complex<numT>& num1, const std::complex<numT>& num2, const numT& threshold = static_cast<numT>(1e-15))
{
	if (thresholdEqual(std::real(num1), std::real(num2), threshold) && thresholdEqual(std::real(num1), std::real(num2), threshold))
		return true;
	else
		return false;
}

//! Function to test if two values are approximately equal.
template <class numT>
bool thresholdEqual(const std::complex<numT>& num1, const numT& num2, const numT& threshold = static_cast<numT>(1e-15))
{
	return thresholdEqual(num1, std::complex<numT>(num2), threshold);
}

//! Function to test if two values are approximately equal.
template <class numT>
bool thresholdEqual(const numT& num1, const std::complex<numT>& num2, const numT& threshold = static_cast<numT>(1e-15))
{
	return thresholdEqual(std::complex<numT>(num1), num2, threshold);
}

//! Function to increase the size of an interval by a set amount.
template <class numT>
interval<std::complex<numT> > growInterval(const interval<std::complex<numT> >& inputInterval, const numT& amount)
{
	interval<std::complex<numT> > out;
	
	out.start_T = std::complex<numT>(std::min(std::real(inputInterval.start_T), std::real(inputInterval.end_T)) - amount, std::min(std::imag(inputInterval.start_T), std::imag(inputInterval.end_T)) - amount);
	out.end_T = std::complex<numT>(std::max(std::real(inputInterval.start_T), std::real(inputInterval.end_T)) + amount, std::max(std::imag(inputInterval.start_T), std::imag(inputInterval.end_T)) + amount);
	
	return out;
}

//! Function to solve a stiff ODE system.
/*!
 * The function is based on a Runge-Kutta formula described in J.R. Cash, “A
 * Semi-Implicit Runge-Kutta Formula for the Integration of Stiff Systems of
 * Ordinary Differential Equations,” The Chemical Engineering Journal, 20,
 * 219-224, (1980).
 *
 * \sa ode_integration_step
 */
template <class numT>
matrix<std::complex<numT> > solve_ode_system(const matrix<std::complex<numT> >& initialisation, const numT& x_start, const numT& x_end, const numT& scaling, matrix<std::complex<numT> >(*jacobian_function)(const numT& x, void* param), void* param, const numT& eps = static_cast<numT>(1e-6))
{
	matrix<std::complex<numT> > y_matrix(initialisation);
	matrix<std::complex<numT> > yTemp_matrix(initialisation.rows());
	numT x_border, step_size, x, TE = static_cast<numT>(0.0), transmittance;

	// Initialisation
// 	y_matrix(0) = 1.0;
// 	for (unsigned int counter = 1; counter < this->_modeDescriptions_vector.size(); counter++)
// 	{
// 		y_matrix(counter) = 0.0;
// 	}

	x_border = x_end * scaling;
	step_size = (x_end - x_start) * scaling / static_cast<numT>(10000.0);
	x = x_start;

// 	if (this->_streams_map["transmission"] != NULL)
// 	{
// 		this->_streams_map["transmission"]->lock();
// 		*(this->_streams_map["transmission"]->getObjectPtr()) << x << "\t" << 10 * log10(std::real(y_matrix(0)) * std::real(y_matrix(0)) + std::imag(y_matrix(0)) * std::imag(y_matrix(0))) << "\n";
// 		this->_streams_map["transmission"]->unlock();
// 	}

	while (x <= x_border)
	{
		ode_integration_step(y_matrix, yTemp_matrix, TE, step_size, x, jacobian_function, param);
		if (std::abs(TE) > eps)
			step_size /= static_cast<numT>(2.0);
		else
		{
			if (std::abs(TE) < (eps / static_cast<numT>(20.0)))
				step_size *= static_cast<numT>(2.0);
				
			if ((x < x_border) && ((x + step_size) > x_border))
				x = x_border;
			else
				x += step_size;
				
			y_matrix = yTemp_matrix;
			
// 			transmittance = 10.0 * log10(std::abs(y_matrix(0)) * std::abs(y_matrix(0)));
// 			
// 			if (this->_streams_map["transmission"] != NULL)
// 			{
// 				this->_streams_map["transmission"]->lock();
// 				*(this->_streams_map["transmission"]->getObjectPtr()) << x << "\t" << transmittance << "\n";
// 				this->_streams_map["transmission"]->unlock();
// 			}
		}
	}
	
// 	if (this->_streams_map["transmittance"] != NULL)
// 	{
// 		this->_streams_map["transmittance"]->lock();
// 		if (!isnan(this->_variable_double))
// 		{
// 			*(this->_streams_map["transmittance"]->getObjectPtr()) << this->_variable_double << "\t";
// 		}
// 		*(this->_streams_map["transmittance"]->getObjectPtr()) << transmittance << "\n";
// 		this->_streams_map["transmittance"]->unlock();
// 	}
// 	this->_transmission_double = (double)(10.0 * log10(std::abs(y_matrix(0)) * std::abs(y_matrix(0))));
	return y_matrix;
}

//! Subfunction to solve a stiff ODE system.
/*!
 * \sa solve_ode_system
 */
template <class numT>
void ode_integration_step(const matrix<std::complex<numT> >& y, matrix<std::complex<numT> >& yTemp, numT& TE, const numT& step_size, const numT& x, matrix<std::complex<numT> >(*jacobian_function)(const numT& x, void* param), void* param)
{
	numT a  = static_cast<numT>(1.0 - std::sqrt(0.5));
	numT c  = static_cast<numT>(-0.5514700764);
	numT d  = static_cast<numT>(-0.1555595513);
	numT e  = static_cast<numT>(0.9993640515);
	numT w1 = static_cast<numT>(0.6662422963);
	numT w2 = static_cast<numT>(0.3337577037);
	numT c_hat = -(a * a);
	unsigned int size_uint(y.rows());
	matrix<std::complex<numT> > fy_matrix(jacobian_function(x, param));
	matrix<std::complex<numT> > fTemp_matrix(size_uint, size_uint);
	matrix<std::complex<numT> > f1_matrix(size_uint);
	matrix<std::complex<numT> > f2_matrix(size_uint);
	matrix<std::complex<numT> > temp_matrix(size_uint);

	for (unsigned int counter1 = 0; counter1 < size_uint; counter1++)
	{
		f1_matrix(counter1) = static_cast<numT>(0.0);
		for (unsigned int counter2 = 0; counter2 < size_uint; counter2++) {
			f1_matrix(counter1) += fy_matrix(counter1, counter2) * y(counter2);
			if (counter1 == counter2)
				fTemp_matrix(counter1, counter2) = std::complex<numT>(static_cast<numT>(1.0), static_cast<numT>(0.0)) - fy_matrix(counter1, counter2) * a * step_size;
			else
				fTemp_matrix(counter1, counter2) = std::complex<numT>(static_cast<numT>(0.0), static_cast<numT>(0.0)) - fy_matrix(counter1, counter2) * a * step_size;
		}
	}
	
	matrix<std::complex<numT> > LU_matrix(fTemp_matrix.ludecomp());
	
	for (unsigned int counter1 = 0; counter1 < size_uint; counter1++)
	{
		temp_matrix(counter1) = static_cast<numT>(0.0);
		for (unsigned int counter2 = 0; counter2 < size_uint; counter2++)
			temp_matrix(counter1) += fy_matrix(counter1, counter2) * f1_matrix(counter2);

		temp_matrix(counter1) = (f1_matrix(counter1) + temp_matrix(counter1) * c * step_size) * step_size;
	}
	
	matrix<std::complex<numT> > temp2_matrix(LU_matrix.fbsubs(temp_matrix));
	matrix<std::complex<numT> > k1_matrix(LU_matrix.fbsubs(temp2_matrix));
	
	for (unsigned int counter1 = 0; counter1 < size_uint; counter1++)
	{
		f2_matrix(counter1) = static_cast<numT>(0.0);
		for (unsigned int counter2 = 0; counter2 < size_uint; counter2++)
			f2_matrix(counter1) += fy_matrix(counter1, counter2) * (y(counter2) + k1_matrix(counter2) * e);
	}
	
	for (unsigned int counter1 = 0; counter1 < size_uint; counter1++)
	{
		for (unsigned int counter2 = 0; counter2 < size_uint; counter2++)
		{
			if (counter1 == counter2)
				fTemp_matrix(counter1, counter2) = std::complex<numT>(static_cast<numT>(1.0), static_cast<numT>(0.0)) + fy_matrix(counter1, counter2) * d * step_size;
			else
				fTemp_matrix(counter1, counter2) = std::complex<numT>(static_cast<numT>(0.0), static_cast<numT>(0.0)) + fy_matrix(counter1, counter2) * d * step_size;
		}
	}

	for (unsigned int counter1 = 0; counter1 < size_uint; counter1++)
	{
		temp_matrix(counter1) = static_cast<numT>(0.0);
		for (unsigned int counter2 = 0; counter2 < size_uint; counter2++)
			temp_matrix(counter1) += fTemp_matrix(counter1, counter2) * f2_matrix(counter2);
	
		temp_matrix(counter1) *= step_size;
	}

	temp2_matrix = LU_matrix.fbsubs(temp_matrix);
	matrix<std::complex<numT> > k2_matrix(LU_matrix.fbsubs(temp2_matrix));
	
	// y_new_hat
	for (unsigned int counter1 = 0; counter1 < size_uint; counter1++)
	{
		temp_matrix(counter1) = static_cast<numT>(0.0);
		for (unsigned int counter2 = 0; counter2 < size_uint; counter2++)
			temp_matrix(counter1) += fy_matrix(counter1, counter2) * f1_matrix(counter2);
		
		temp_matrix(counter1) = f1_matrix(counter1) + temp_matrix(counter1) * step_size * c_hat;
	}
	
	temp2_matrix = LU_matrix.fbsubs(temp_matrix);
	temp_matrix = LU_matrix.fbsubs(temp2_matrix);
	
	TE = static_cast<numT>(0.0);
	
	for (unsigned int counter = 0; counter < size_uint; counter++)
	{
		numT TE_check;
		
		yTemp(counter) = (k1_matrix(counter) * w1 + k2_matrix(counter) * w2) + y(counter);
		TE_check = std::abs(yTemp(counter) - (temp_matrix(counter) * step_size + y(counter)));
		if (TE_check > TE)
			TE = TE_check;
	}
}

#endif

