/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _GRATINGMODELTM_H
#define _GRATINGMODELTM_H 1

#include <complex>
#include <vector>

#include "algorithms.h"
#include "sectionSolver.h"
#include "matrix_template.h"
#include "section.h"
#include "mode.h"

template<class modeModel, class numT>
class gratingModelTM : public sectionSolver<numT>
{
 	private:
		section* f_sectionPtr;
 		std::vector<mode<modeModel, numT> > f_mode_vector;
		matrix<std::complex<numT> >* f_zeta_matrix;
		numT f_normalisation;
		numT f_scaling;
		numT f_lambda;

	private:
		void set_zeta_matrix();
	
	public:
		gratingModelTM(const std::vector<mode<modeModel, numT> >& mode_vector, const section& theSection, const numT& lambda, const numT& scaling = static_cast<numT>(1e6), const numT& normalisation = static_cast<numT>(1.0));
		gratingModelTM(const gratingModelTM<modeModel, numT>& object);
		~gratingModelTM();
//		static matrix<std::complex<double> > jacobian(const double& x, void* param);
		numT get_scaling();
		numT get_lambda();
		section* get_section_ptr();
		std::vector<mode<modeModel, numT> >* get_mode_vector_ptr();
		std::complex<numT> get_zeta_matrix_element(const unsigned int& row);
		matrix<std::complex<numT> > amplitudes(const matrix<std::complex<numT> >& init);
};

template<class modeModel, class numT>
void gratingModelTM<modeModel, numT>::set_zeta_matrix()
{
	layerDescription<numT> coreLayer = f_mode_vector[0].get_layer_description_for_radius(static_cast<numT>(0.0));
	numT k0Scaled_double = static_cast<numT>(2.0 * M_PI) / (f_lambda * f_scaling);
	
	for (int counter = 0; counter < f_mode_vector.size(); counter++)
	{
		f_mode_vector[counter].set_scaling(f_scaling);
		f_mode_vector[counter].set_normalisation(f_normalisation);
	}
	
	if (f_zeta_matrix != NULL)
		delete f_zeta_matrix;
	
	f_zeta_matrix = new matrix<std::complex<numT> >(f_mode_vector.size(), f_mode_vector.size());

	for (unsigned int rowCounter_uint = 0; rowCounter_uint < f_mode_vector.size(); rowCounter_uint++)
	{
		numT azimuthalIntegral = static_cast<numT>(2.0 * M_PI);
		std::complex<numT> radialIntegral = coupling_integral_radial(f_mode_vector[0], f_mode_vector[rowCounter_uint], static_cast<numT>(0.0), coreLayer.r);
		(*f_zeta_matrix)(rowCounter_uint) = (k0Scaled_double * static_cast<std::complex<numT> >(std::sqrt(kEPSILON_0 / kMU_0)) / (f_scaling * f_scaling)) / (static_cast<numT>(2.0) * f_normalisation) * coreLayer.n * static_cast<numT>(f_sectionPtr->get_p_change()) * azimuthalIntegral * radialIntegral;
	}
	
	for (unsigned int rowCounter_uint = 1; rowCounter_uint < f_mode_vector.size(); rowCounter_uint++)
	{
		numT azimuthalIntegral = static_cast<numT>(2.0 * M_PI);
		std::complex<numT> radialIntegral = coupling_integral_radial(f_mode_vector[rowCounter_uint], f_mode_vector[rowCounter_uint], static_cast<numT>(0.0), coreLayer.r);
		(*f_zeta_matrix)(rowCounter_uint, rowCounter_uint) = (k0Scaled_double * static_cast<std::complex<numT> >(std::sqrt(kEPSILON_0 / kMU_0)) / (f_scaling * f_scaling)) / (static_cast<numT>(2.0) * f_normalisation) * coreLayer.n * static_cast<numT>(f_sectionPtr->get_p_change()) * azimuthalIntegral * radialIntegral;		
	}
}

// template<class modeModel>
// matrix<std::complex<double> > gratingModelTM<modeModel>::jacobian(const double& x, void* param)
// {
// 	gratingModelTM* modeSolver = (gratingModelTM*)param;
// 	
// 	std::vector<mode<modeModel> >* mode_vectorPtr = modeSolver->get_mode_vector_ptr();
// 	double k0Scaled = 2.0 * M_PI / (modeSolver->get_lambda() * modeSolver->get_scaling());
// 	double sigma = modeSolver->get_section_ptr()->get_sigma(x / modeSolver->get_scaling());
// 
// 	matrix<std::complex<double> > fy(mode_vectorPtr->size(), mode_vectorPtr->size());
// 	
// 	for (int counter1 = 0; counter1 < mode_vectorPtr->size(); counter1++)
// 	{
// 		for (int counter2 = 0; counter2 < mode_vectorPtr->size(); counter2++)
// 		{
// 			if (counter1 == counter2)
// 			{
// 				fy(counter1, counter2) = sigma * std::complex<double>(0.0, -modeSolver->get_section_ptr()->get_dc_change()) * modeSolver->get_zeta_matrix_element(counter1, counter2) * std::exp(-2.0 * std::abs(k0Scaled * std::abs(std::imag((*mode_vectorPtr)[counter1].get_nEff())) * x));
// 			}
// 			else
// 			{
// 				double accumulatedPhase_double = modeSolver->get_section_ptr()->get_accumulated_phase(x / modeSolver->get_scaling());
// 					
// 				std::complex<double> exponent;
// 				if (std::real((*mode_vectorPtr)[counter1].get_nEff()) < std::real((*mode_vectorPtr)[counter2].get_nEff()))
// 				{
// 					// +
// 					exponent = std::complex<double>(0.0, x * k0Scaled * (std::real((*mode_vectorPtr)[counter1].get_nEff()) - std::real((*mode_vectorPtr)[counter2].get_nEff())) + accumulatedPhase_double);
// 				}
// 				else
// 				{
// 					// -
// 					exponent = std::complex<double>(0.0, x * k0Scaled * (std::real((*mode_vectorPtr)[counter1].get_nEff()) - std::real((*mode_vectorPtr)[counter2].get_nEff())) - accumulatedPhase_double);
// 				}
// 				fy(counter1, counter2) = sigma * std::complex<double>(0.0, -modeSolver->get_section_ptr()->get_ac_change() / 2.0) * modeSolver->get_zeta_matrix_element(counter1, counter2) * std::exp(exponent) * std::exp(-x * k0Scaled * std::abs(std::imag((*mode_vectorPtr)[counter1].get_nEff() + (*mode_vectorPtr)[counter2].get_nEff())));
// 			}
// 		}
// 	}
// 	
// 	return fy;
// }

template <class modeModel, class numT>
gratingModelTM<modeModel, numT>::gratingModelTM(const std::vector<mode<modeModel, numT> >& mode_vector, const section& theSection, const numT& lambda, const numT& scaling, const numT& normalisation)
{
	f_mode_vector = mode_vector;
	f_sectionPtr = new section(theSection);
	f_zeta_matrix = 0;
	f_lambda = lambda;
	f_scaling = scaling;
	f_normalisation = normalisation;
}

template <class modeModel, class numT>
gratingModelTM<modeModel, numT>::gratingModelTM(const gratingModelTM<modeModel, numT>& object)
{
	f_mode_vector = object.f_mode_vector;
	f_sectionPtr = object._theGrating->clone();

	if (object._zeta_matrix != 0)
		f_zeta_matrix = new matrix<std::complex<numT> >(*(object._zeta_matrix));
	else
		f_zeta_matrix = 0;

	f_lambda = object.f_lambda;
	f_scaling = object.f_scaling;
	f_normalisation = object.f_normalisation;
}

template <class modeModel, class numT>
gratingModelTM<modeModel, numT>::~gratingModelTM()
{
	delete f_sectionPtr;
	delete f_zeta_matrix;
}

template<class modeModel, class numT>
numT gratingModelTM<modeModel, numT>::get_scaling()
{
	return f_scaling;
}

template<class modeModel, class numT>
numT gratingModelTM<modeModel, numT>::get_lambda()
{
	return f_lambda;
}

template<class modeModel, class numT>
section* gratingModelTM<modeModel, numT>::get_section_ptr()
{
	return f_sectionPtr;
}

template<class modeModel, class numT>
std::vector<mode<modeModel, numT> >* gratingModelTM<modeModel, numT>::get_mode_vector_ptr()
{
	return &f_mode_vector;
}

template<class modeModel, class numT>
std::complex<numT> gratingModelTM<modeModel, numT>::get_zeta_matrix_element(const unsigned int& row)
{
	if (f_zeta_matrix == 0) {
		set_zeta_matrix();
	}

	return (*f_zeta_matrix)(row);
}

template<class modeModel, class numT>
matrix<std::complex<numT> > gratingModelTM<modeModel, numT>::amplitudes(const matrix<std::complex<numT> >& init)
{
	if (init.rows() <= 0)
		return matrix<std::complex<numT> >(0);
	else if (init.rows() == 1)
		return matrix<std::complex<numT> >(init);

	if (f_zeta_matrix == 0) {
		set_zeta_matrix();
	}

	std::complex<numT> min_sigma = (f_mode_vector[0].get_nEff() - f_mode_vector[1].get_nEff() - f_lambda / static_cast<numT>(f_sectionPtr->get_period())) * static_cast<numT>(M_PI) / (f_lambda * f_scaling) + ((*f_zeta_matrix)(0) * static_cast<numT>(f_sectionPtr->get_dc_change()) - static_cast<numT>(f_sectionPtr->get_ac_change()) * (*f_zeta_matrix)(1, 1) /  static_cast<numT>(2.0)) / static_cast<numT>(2.0);
	int min_number = 1;
	std::complex<numT> i(static_cast<numT>(0.0), static_cast<numT>(1.0));

	for (int counter = 2; counter < f_zeta_matrix->rows(); counter++)
	{
		std::complex<numT> temp_sigma = (f_mode_vector[0].get_nEff() - f_mode_vector[counter].get_nEff() - f_lambda / static_cast<numT>(f_sectionPtr->get_period())) * static_cast<numT>(M_PI) / (f_lambda * f_scaling) + ((*f_zeta_matrix)(0) * static_cast<numT>(f_sectionPtr->get_dc_change()) - static_cast<numT>(f_sectionPtr->get_ac_change()) * (*f_zeta_matrix)(counter,counter) /  static_cast<numT>(2.0)) / static_cast<numT>(2.0);

		if (std::abs(min_sigma) > std::abs(temp_sigma))
		{
			min_sigma = temp_sigma;
			min_number = counter;
		}
	}
	
	matrix<std::complex<numT> > amps(2);

	amps(0) = init(0);
	amps(1) = init(min_number);
	
	std::complex<numT> gamma = std::sqrt(min_sigma * min_sigma + std::abs(static_cast<numT>(f_sectionPtr->get_ac_change()) * (*f_zeta_matrix)(min_number) / static_cast<numT>(2.0)) * std::abs(static_cast<numT>(f_sectionPtr->get_ac_change()) * (*f_zeta_matrix)(min_number) / static_cast<numT>(2.0)));
	numT k0 = static_cast<numT>(2.0 * M_PI) / (f_lambda * f_scaling);
	
	numT z = f_sectionPtr->get_length() * f_scaling;
	
	matrix<std::complex<numT> > T(2,2);
	T(0,0) = std::cos(gamma * z) + i * min_sigma / gamma * std::sin(gamma * z);
	T(0,1) = i * static_cast<numT>(f_sectionPtr->get_ac_change()) * (*f_zeta_matrix)(min_number) / static_cast<numT>(2.0) / gamma * std::sin(gamma * z);
	T(1,0) = i * static_cast<numT>(f_sectionPtr->get_ac_change()) * std::conj((*f_zeta_matrix)(min_number)) / static_cast<numT>(2.0) / gamma * std::sin(gamma * z);
	T(1,1) = std::cos(gamma * z) + i * min_sigma / gamma * std::sin(gamma * z);

	return T * amps;
}

#endif

