/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "layer.h"

layer::layer(const std::wstring& type) throw (std::runtime_error)
{
	f_type = type;
	f_r = 0;
	f_e = 0;
	f_u = 0;

	std::vector<std::wstring> vec = tokenise(f_type, L'_');
	
	if ((vec.size() >= 1) && (vec[0] == L"constant"))
	{		
		layer_constant::init(*this);
		if ((vec.size() >= 2) && (vec[1] == L"dielectric"))
		{
			layer_constant_dielectric::init(*this);
			if ((vec.size() >= 3) && (vec[2] == L"delta"))
			{
				layer_constant_dielectric_delta::init(*this);
			}
		}
		return;
	}
	else if ((vec.size() >= 1) && (vec[0] == L"linear"))
	{
		layer_linear::init(*this);
		if ((vec.size() >= 2) && (vec[1] == L"dielectric"))
		{
			layer_linear_dielectric::init(*this);
		}
		return;
	}
	
	throw std::runtime_error("Layer type unknown.");
}

layer::layer(const layer& obj)
{
	f_type = obj.f_type;
	f_r = new eo<std::complex<double> >(*obj.f_r);
	f_e = new eo<std::complex<double> >(*obj.f_e);
	f_u = new eo<std::complex<double> >(*obj.f_u);
}

std::vector<std::wstring> layer::get_types()
{
	std::vector<std::wstring> output;
	output.push_back(L"constant");
	output.push_back(L"constant_dielectric");
	output.push_back(L"constant_dielectric_delta");
	output.push_back(L"linear");
	output.push_back(L"linear_dielectric");
	return output;
}

std::map<std::wstring, std::complex<double> > layer::get_vars() const
{
	std::map<std::wstring, std::complex<double> > vars;
	
	if (f_r != 0)
		f_r->get_vars(vars);
	
	if (f_e != 0)
		f_e->get_vars(vars);
	
	if (f_u != 0)
		f_u->get_vars(vars);
	
	return vars;
}

void layer::set_vars(const std::map<std::wstring, std::complex<double> >& vars)
{
	if (f_r != 0)
		f_r->set_vars(vars);
	
	if (f_e != 0)
		f_e->set_vars(vars);
	
	if (f_u != 0)
		f_u->set_vars(vars);
}

void layer::rename_vars(const std::map<std::wstring, std::wstring >& vars)
{
	if (f_r != 0)
		f_r->rename_vars(vars);
	
	if (f_e != 0)
		f_e->rename_vars(vars);
	
	if (f_u != 0)
		f_u->rename_vars(vars);	
}

void layer::set_type(const std::wstring& type)
{
	f_type = type;
}

std::wstring layer::get_type()
{
	return f_type;
}

void layer::set_r(const eo<std::complex<double> >& r)
{
	delete f_r;
	f_r = new eo<std::complex<double> >(r);
}

eo<std::complex<double> > layer::get_r()
{
	return eo<std::complex<double> >(*f_r);
}

void layer::set_e(const eo<std::complex<double> >& e)
{
	delete f_e;
	f_e = new eo<std::complex<double> >(e);
}

void layer::set_u(const eo<std::complex<double> >& u)
{
	delete f_u;
	f_u = new eo<std::complex<double> >(u);
}

void layer::simplify()
{
	if (f_e != 0)
		f_e->simplify();
	
	if (f_u != 0)
		f_u->simplify();
}

layer::~layer()
{
	delete f_r;
	delete f_e;
	delete f_u;
}

void layer_constant::init(layer& obj)
{
	obj.set_type(L"constant");
	obj.set_r(eo<std::complex<double> >(L"radius"));
	obj.set_e(eo<std::complex<double> >(L"electricIndex"));
	obj.set_u(eo<std::complex<double> >(L"magneticIndex"));
	obj.simplify();
}

void layer_constant_dielectric::init(layer& obj)
{
	obj.set_type(L"constant_dielectric");
	obj.set_r(eo<std::complex<double> >(L"radius"));
	obj.set_e(eo<std::complex<double> >(L"refractiveIndex"));
	obj.set_u(eo<std::complex<double> >(1.0));
	obj.simplify();
}

void layer_constant_dielectric_delta::init(layer& obj)
{
	obj.set_type(L"constant_dielectric_delta");
	obj.set_r(eo<std::complex<double> >(L"radius"));
	obj.set_e(eo<std::complex<double> >(L"refractiveIndex") / (1.0 - eo<std::complex<double> >(L"delta")));
	obj.set_u(eo<std::complex<double> >(1.0));
	obj.simplify();
}

void layer_linear::init(layer& obj)
{
	obj.set_type(L"linear");
	obj.set_r(eo<std::complex<double> >(L"radius"));
	
	eo<std::complex<double> > ke = (eo<std::complex<double> >(L"electricIndex1") - eo<std::complex<double> >(L"electricIndex2")) / (eo<std::complex<double> >(L"lambda1") - eo<std::complex<double> >(L"lambda2"));
	eo<std::complex<double> > de = eo<std::complex<double> >(L"electricIndex1") - ke * eo<std::complex<double> >(L"lambda1");
	obj.set_e(ke * eo<std::complex<double> >(L"lambda") + de);

	eo<std::complex<double> > ku = (eo<std::complex<double> >(L"magneticIndex1") - eo<std::complex<double> >(L"magneticIndex2")) / (eo<std::complex<double> >(L"lambda1") - eo<std::complex<double> >(L"lambda2"));
	eo<std::complex<double> > du = eo<std::complex<double> >(L"magneticIndex1") - ku * eo<std::complex<double> >(L"lambda1");
	obj.set_u(ku * eo<std::complex<double> >(L"lambda") + du);

	obj.simplify();
}

void layer_linear_dielectric::init(layer& obj)
{
	obj.set_type(L"linear_dielectric");
	obj.set_r(eo<std::complex<double> >(L"radius"));
	
	eo<std::complex<double> > ke = (eo<std::complex<double> >(L"refractiveIndex1") - eo<std::complex<double> >(L"refractiveIndex2")) / (eo<std::complex<double> >(L"lambda1") - eo<std::complex<double> >(L"lambda2"));
	eo<std::complex<double> > de = eo<std::complex<double> >(L"refractiveIndex1") - ke * eo<std::complex<double> >(L"lambda1");
	obj.set_e(ke * eo<std::complex<double> >(L"lambda") + de);

	obj.set_u(eo<std::complex<double> >(1.0));

	obj.simplify();
}

