/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _LAYER_H
#define _LAYER_H 1

#include <string>
#include <vector>
#include <complex>
#include <stdexcept>

#include "eo.h"
#include "stringFunctions.h"
#include "matrix_template.h"

template<class numT>
struct layerDescription
{
	std::complex<numT> n; /**< refractive index n = sqrt(epsilon / epsilon_0) * sqrt(mu / mu_0) */
	std::complex<numT> e; /**< electric index e = sqrt(epsilon / epsilon_0) */
	std::complex<numT> u; /**< magnetic index u = sqrt(mu / mu_0) */
	numT r; /**< radius */
	matrix<std::complex<numT> > fieldExpansionCoefficients; /** matrix holding the field expansion coeffcients */
	bool fieldExpansionCoeffcientsSet; /** whether the field expansion coefficients are set */
	std::wstring type;
};

class layer
{	
	private:
		std::wstring f_type;
		eo<std::complex<double> >* f_r;
		eo<std::complex<double> >* f_e;
		eo<std::complex<double> >* f_u;
		
	public:
		layer(const std::wstring& type) throw (std::runtime_error);
		layer(const layer& obj);
		static std::vector<std::wstring> get_types();
		std::map<std::wstring, std::complex<double> > get_vars() const;
		void set_vars(const std::map<std::wstring, std::complex<double> >& vars);
		void rename_vars(const std::map<std::wstring, std::wstring >& vars);
		void set_type(const std::wstring& type);
		std::wstring get_type();
		void set_r(const eo<std::complex<double> >& r);
		eo<std::complex<double> > get_r();
		void set_e(const eo<std::complex<double> >& e);
		void set_u(const eo<std::complex<double> >& u);
		void simplify();
		template<class numT>
		layerDescription<numT> description() const throw (std::runtime_error);
		~layer();

	private:
		layer();
};

template<class numT>
layerDescription<numT> layer::description() const throw (std::runtime_error)
{
	if ((f_r != 0) && (f_e != 0) && (f_u != 0))
	{
		layerDescription<numT> out = {static_cast<std::complex<numT> >((std::complex<double>)(*f_e) * (std::complex<double>)(*f_u)), \
														static_cast<std::complex<numT> >((std::complex<double>)(*f_e)), \
														static_cast<std::complex<numT> >((std::complex<double>)(*f_u)), \
														static_cast<numT>(std::abs((std::complex<double>)*f_r)), \
														matrix<std::complex<numT> >(), false, f_type};
		return out;
	}
	
	throw std::runtime_error("Not all layer functions are set.");
}



class layer_constant
{
	public:
		static void init(layer& obj);
};

class layer_constant_dielectric
{
	public:
		static void init(layer& obj);
};

class layer_constant_dielectric_delta
{
	public:
		static void init(layer& obj);
};

class layer_linear
{
	public:
		static void init(layer& obj);
};

class layer_linear_dielectric
{
	public:
		static void init(layer& obj);
};
#endif

