/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _MODE_H
#define _MODE_H 1

#include <complex>
#include <vector>
#include <ostream>
//#include "mpfr++.h"
#include "algorithms.h"
#include "ad.h"
//#include "modeModelHE.h"
//#include "modeModelLP.h"

template <class modeModel, class numT>
class mode : public modeModel
{
	private:
		unsigned int _order_uint;															/**< mode order */
		numT _lambda_double;
		numT _scaling_double;
		std::complex<numT> _nEff_complexDouble;							/**< effective refractive index of the mode */
		std::vector<layerDescription<numT> > _layerDescription_vector;
		numT _power;
	
	public:
		mode(const std::complex<numT>& nEff, const numT& lambda, const numT& scaling, const std::vector<layerDescription<numT> >& layerDescription);
		std::complex<numT> get_nEff() const;
		numT get_lambda() const;
		numT get_scaling() const;
		numT get_power();
		unsigned get_order() const;
		bool are_field_expansion_coeffcients_set();
		void set_field_expansion_coefficients();
		layerDescription<numT> get_layer_description_for_radius(const numT& r) const;
		std::vector<layerDescription<numT> > get_layer_description_vector() const;
		void calculate_mode_power();
		void set_normalisation(const numT& power = static_cast<numT>(1.0));
		void set_scaling(const numT& scaling = static_cast<numT>(1e6));
		//std::ostream& plot_mode_field_radial(std::ostream& os);
};

template <class modeKind, class numT>
bool operator<(const mode<modeKind, numT>& mode1, const mode<modeKind, numT>& mode2);

template <class modeKind, class numT>
mode<modeKind, numT>::mode(const std::complex<numT>& nEff, const numT& lambda, const numT& scaling, const std::vector<layerDescription<numT> >& layerDescription)
{
	_nEff_complexDouble = nEff;
	_layerDescription_vector = layerDescription;
	_lambda_double = lambda;
	_scaling_double = scaling;
	_power = NAN;
	_order_uint = (unsigned int)intValue(modeKind::get_core_mode_order());
}

template <class modeKind, class numT>
std::complex<numT> mode<modeKind, numT>::get_nEff() const
{
	return _nEff_complexDouble;
}

template <class modeKind, class numT>
numT mode<modeKind, numT>::get_lambda() const
{
	return _lambda_double;
}

template <class modeKind, class numT>
numT mode<modeKind, numT>::get_scaling() const
{
	return _scaling_double;
}

template <class modeKind, class numT>
numT mode<modeKind, numT>::get_power()
{
	if (std::isnan(_power))
		calculate_mode_power();

	return _power;
}

template <class modeKind, class numT>
unsigned int mode<modeKind, numT>::get_order() const
{
	return _order_uint;
}

template <class modeKind, class numT>
bool mode<modeKind, numT>::are_field_expansion_coeffcients_set()
{
	for (typename std::vector<layerDescription<numT> >::iterator it = _layerDescription_vector.begin(); it != _layerDescription_vector.end(); it++)
		if (!(*it).fieldExpansionCoeffcientsSet)
			return false;

	return true;
}

template <class modeKind, class numT>
void mode<modeKind, numT>::set_field_expansion_coefficients()
{
	// incremental inside out
	for (int counter = 0; counter < _layerDescription_vector.size(); counter++)
	{
		if (counter == 0)
			_layerDescription_vector[counter].fieldExpansionCoefficients = modeKind::field_expansion_coefficients_centre(*this);
		else
			_layerDescription_vector[counter].fieldExpansionCoefficients = modeKind::M(_nEff_complexDouble, _layerDescription_vector[counter].e, _layerDescription_vector[counter].u, _layerDescription_vector[counter - 1].e, _layerDescription_vector[counter - 1].u, _layerDescription_vector[counter - 1].r, _lambda_double, _order_uint, _scaling_double) * _layerDescription_vector[counter - 1].fieldExpansionCoefficients;
		
		_layerDescription_vector[counter].fieldExpansionCoeffcientsSet = true;
	}
	
	modeKind::ensure_bound_modes_in_outermost_layer(_layerDescription_vector);
}

template <class modeKind, class numT>
layerDescription<numT> mode<modeKind, numT>::get_layer_description_for_radius(const numT& r) const
{
	unsigned int layerCounter_uint = 0;
	for (; (layerCounter_uint < _layerDescription_vector.size()) && (_layerDescription_vector[layerCounter_uint].r < r); layerCounter_uint++) {}

	return _layerDescription_vector[layerCounter_uint];
}

template <class modeKind, class numT>
std::vector<layerDescription<numT> > mode<modeKind, numT>::get_layer_description_vector() const
{
	return _layerDescription_vector;
}


template <class modeKind, class numT>
void mode<modeKind, numT>::calculate_mode_power()
{
	if (!are_field_expansion_coeffcients_set())
		set_field_expansion_coefficients();
	
	unsigned int check = 0;
	_power = modeKind::normalisation_prefactor(*this) * std::real(integral(&(modeKind::normalisation_integration_function), this, static_cast<numT>(0.0), _layerDescription_vector[_layerDescription_vector.size() - 1].r, check));
}

template <class modeKind, class numT>
void mode<modeKind, numT>::set_normalisation(const numT& power)
{
	std::complex<numT> factor = std::sqrt(power / get_power());
	
	for (typename std::vector<layerDescription<numT> >::iterator it = _layerDescription_vector.begin(); it != _layerDescription_vector.end(); it++)
		it->fieldExpansionCoefficients = it->fieldExpansionCoefficients * factor;
	
	_power = power;
}

template <class modeKind, class numT>
void mode<modeKind, numT>::set_scaling(const numT& scaling)
{
	if (!std::isnan(_power))
	{
		numT factor = scaling / _scaling_double;
		factor *= factor;
		factor *= factor;
		_power *= factor;
	}
	_scaling_double = scaling;
}

template <class modeKind, class numT>
bool operator<(const mode<modeKind, numT>& mode1, const mode<modeKind, numT>& mode2)
{
	if (std::real(mode1.get_nEff()) < std::real(mode2.get_nEff()))
		return true;
	else
		return false;
}

//template <class modeModelHE<numT>, class numT>
//std::ostream& mode<modeModelHE<numT>, numT>::plot_mode_field_radial(std::ostream& os)
//{
//	numT r_start = static_cast<numT>(1e-14);
//	numT r_end = _layerDescription_vector[_layerDescription_vector.size() - 1].r;
//	numT r_step = r_end / 1000.0;
//
//	os << "radius [m]\treal(n)\timag(n)\treal(E_phi)\timag(E_phi)\treal(n^2*E_r)\timag(n^2*E_r)\treal(E_z)\timag(E_z)\treal(H_phi)\timag(H_phi)\treal(n^2*H_r)\timag(n^2*H_r)\treal(H_z)\timag(H_z)\n";
//	for (numT radius = r_start; radius < r_end; radius += r_step)
//	{
//		std::complex<numT> E_phi, E_r, E_z, H_phi, H_r, H_z;
//		layerDescription layer = get_layer_description_for_radius(radius);
//		modeModelHE<numT>::field_radial(radius, _nEff_complexDouble, layer.e, layer.u, layer.fieldExpansionCoefficients, _lambda_double, _order_uint, _scaling_double, &E_phi, &E_r, &E_z, &H_phi, &H_r, &H_z);
//		os << doubleValue(radius) << "\t" << doubleValue(std::real(layer.n)) << "\t" << doubleValue(std::imag(layer.n)) << "\t" << doubleValue(std::real(E_phi)) << "\t" << doubleValue(std::imag(E_phi)) << "\t" << doubleValue(std::real(layer.n * layer.n * E_r)) << "\t" << doubleValue(std::imag(layer.n * layer.n * E_r)) << "\t" << doubleValue(std::real(E_z)) << "\t" << doubleValue(std::imag(E_z)) << "\t" << doubleValue(std::real(H_phi)) << "\t" << doubleValue(std::imag(H_phi)) << "\t" << doubleValue(std::real(layer.n * layer.n * H_r)) << "\t" << doubleValue(std::imag(layer.n * layer.n * H_r)) << "\t" << doubleValue(std::real(H_z)) << "\t" << doubleValue(std::imag(H_z)) << "\n";
//	}
//	
//	return os;
//}

//template <, class numT>
//std::ostream& mode<modeModelLP, numT>::plot_mode_field_radial(std::ostream& os)
//{
//	numT r_start = static_cast<numT>(1e-14);
//	numT r_end = _layerDescription_vector[_layerDescription_vector.size() - 1].r;
//	numT r_step = r_end / 1000.0;
//
//	os << "radius [m]\treal(n)\timag(n)\treal(R)\timag(R)\n";
//	for (numT radius = r_start; radius < r_end; radius += r_step)
//	{
//		std::complex<numT> R;
//		layerDescription layer = get_layer_description_for_radius(radius);
//		modeModelLP::field_radial(radius, _nEff_complexDouble, layer.e, layer.u, layer.fieldExpansionCoefficients, _lambda_double, _order_uint, _scaling_double, &R);
//		os << doubleValue(radius) << "\t" << doubleValue(std::real(layer.n)) << "\t" << doubleValue(std::imag(layer.n)) << "\t" << doubleValue(std::real(R)) << "\t" << doubleValue(std::imag(R)) << "\n";
//	}
//	
//	return os;
//}
#endif

