/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _MODEMODELHE_H
#define _MODEMODELHE_H 1

#include <complex>
#include "algorithms.h"
#include "ad.h"
#include "matrix_template.h"
#include "bessel_amos.h"
#include "modeSolver.h"
#include "mode.h"

//! Class implementing the hybrid mode model
/*!
 * This class implements the necessary functions for the hybrid mode model. The
 * underlying theory is outlined in the following publications:
 *
 * P. Yeh, A. Yariv, E. Marom, "Theory of Bragg fiber," Journal of the Optical
 * Society of America, 68(9), 1196-1201 (1978).
 *
 * I. del Villar, "Theoretical analysis and fabrication of nanostructures with
 * Electrostatic Self-Assembly Monolayer process," Universidad Pública de
 * Navarra (2006).
 *
 * T. Erdogan, "Cladding-Mode Resonances in Short- and Long-Period Fiber
 * Grating Filters," Journal of the Otical Society of America A, 14(8),
 * 1760-1773 (1997).
 *
 * A.W. Snyder, J.D.  Love, "Optical Waveguide Theory," Chapman and Hall: New
 * York, USA (1983).
 */
template<class numT>
class modeModelHE
{
	private:
		//! Enum to indicate which, if any, of the layers supplied to the M function is the outermost.
		enum outermostLayerPosition
		{
			kNONE,
			kFIRST,
			kSECOND
		};
		
	public:
		//! Function to supply the core mode order.
		/*!
		 * This function is part of the modeSolver interface.
		 */
		static numT get_core_mode_order(void);
		//! Function to calculate the value and derivative of the characteristic equation.
		/*!
		 * This function is part of the modeSolver interface.
		 */
		static void characteristic_equation(const std::complex<numT>& nEff, void* param, std::complex<numT>* f, std::complex<numT>* fDiff = NULL);
		//! Integration function for the normalisation
		/*!
		 * This function is part of the mode interface.
		 *
		 * \sa M, MAD
		 */
		static std::complex<numT> normalisation_integration_function(const numT& r, void* param);
		//! Function to calculate the prefactor for the normalisation
		/*!
		 * This function is part of the mode interface.
		 */
		static numT normalisation_prefactor(const mode<modeModelHE<numT>, numT>& mode);
		//! Function to calculate the field expansion coefficients for the centre of the fibre
		/*!
		 * This function is part of the mode interface.
		 */
		static matrix<std::complex<numT> > field_expansion_coefficients_centre(const mode<modeModelHE<numT>, numT>& mode);
		//! Function to supply the transfer matrix in the radial direction
		/*!
		 * This function is part of the mode interface.
		 *
		 * \sa characteristic_equation
		 */
		static matrix<std::complex<numT> > M(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition = kNONE);
		//! Function to set the field expansion coefficients yielding oscillating or infinite field at infinite radius to zero.
		/*!
		 * This function is part of the mode interface.
		 */
		void ensure_bound_modes_in_outermost_layer(std::vector<layerDescription<numT> >& layers);
		//! Function to calculate the transfer matrix and its derivative using automatic derivatives.
		/*!
		 * \sa characteristic_equation
		 */
		static matrix<AD<std::complex<numT> > > MAD(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition = kNONE);
		//! Function to supply the radial field
		static void field_radial(const numT& radius, const std::complex<numT>& nEff, const std::complex<numT>& e, const std::complex<numT>& u, const matrix<std::complex<numT> >& fieldExpansionCoefficients, const numT& _lambda, const numT& order, const numT& _scaling, std::complex<numT>* E_phi, std::complex<numT>* E_r, std::complex<numT>* E_z, std::complex<numT>* H_phi, std::complex<numT>* H_r, std::complex<numT>* H_z);
};

template <class numT>
numT modeModelHE<numT>::get_core_mode_order(void)
{
	return static_cast<numT>(1.0);
}

template <class numT>
void modeModelHE<numT>::characteristic_equation(const std::complex<numT>& nEff, void* param, std::complex<numT>* f, std::complex<numT>* fDiff)
{
	modeSolver<modeModelHE<numT>, numT>* solver = (modeSolver<modeModelHE<numT>, numT>*)param;
	numT lambda_double = solver->get_lambda_double();
	std::vector<layerDescription<numT> > layers_vector = solver->get_layers_vector();
	unsigned int modeOrder_uint = static_cast<unsigned int>(intValue(get_core_mode_order()));
	numT scaling_double = solver->get_scaling_double();
	
	if (fDiff != NULL)
	{
		matrix<AD<std::complex<numT> > > m(4, 4, AD<std::complex<numT> >(static_cast<numT>(1.0)), matrix<AD<std::complex<numT> > >::diagonal_fillPattern);
		
		for (int counter = 0; counter < layers_vector.size() - 1; counter++)
 			if (counter == layers_vector.size() - 2)
	 			m = m * MAD(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, static_cast<numT>(modeOrder_uint), scaling_double, kSECOND);
 			else
	 			m = m * MAD(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, static_cast<numT>(modeOrder_uint), scaling_double, kNONE);
		
		AD<std::complex<numT> > result(m(1,1) * m(3,3) - m(1,3) * m(3,1));

		*f = result.fun();
		*fDiff = result.diff();
	}
	else
	{
		matrix<std::complex<numT> > m(4, 4, std::complex<numT>(static_cast<numT>(1.0)), matrix<std::complex<numT> >::diagonal_fillPattern);
		
		for (int counter = 0; counter < layers_vector.size() - 1; counter++)
			if (counter == layers_vector.size() - 2)
				m = m * M(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, static_cast<numT>(modeOrder_uint), scaling_double, kSECOND);
			else
				m = m * M(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, static_cast<numT>(modeOrder_uint), scaling_double, kNONE);
		
		*f = m(1,1) * m(3,3) - m(1,3) * m(3,1);
	}
}

template <class numT>
std::complex<numT> modeModelHE<numT>::normalisation_integration_function(const numT& r, void* param)
{
	mode<modeModelHE<numT>, numT>* theMode = (mode<modeModelHE<numT>, numT>*)param;

	if (!theMode->are_field_expansion_coeffcients_set())
		theMode->set_field_expansion_coefficients();
	
	std::complex<numT> E_r, E_phi, H_r, H_phi;
	
	layerDescription<numT> layer = theMode->get_layer_description_for_radius(r);
	
	field_radial(r, theMode->get_nEff(), layer.e, layer.u, layer.fieldExpansionCoefficients, theMode->get_lambda(), get_core_mode_order(), theMode->get_scaling(), &E_phi, &E_r, NULL, &H_phi, &H_r, NULL);
	return (E_r * std::conj(H_phi) - H_r * std::conj(E_phi)) * r * theMode->get_scaling();
}

template <class numT>
numT modeModelHE<numT>::normalisation_prefactor(const mode<modeModelHE, numT>& mode)
{
	return static_cast<numT>(M_PI);
}

template <class numT>
matrix<std::complex<numT> > modeModelHE<numT>::field_expansion_coefficients_centre(const mode<modeModelHE<numT>, numT>& mode)
{
	matrix<std::complex<numT> > m(4, 4, std::complex<numT>(static_cast<numT>(1.0)), matrix<std::complex<numT> >::diagonal_fillPattern);

	std::vector<layerDescription<numT> > layerDescription_vector(mode.get_layer_description_vector());

	for (int counter = 0; counter < layerDescription_vector.size() - 1; counter++)
		if (counter == layerDescription_vector.size() - 2)
			m = m * M(mode.get_nEff(), layerDescription_vector[counter].e, layerDescription_vector[counter].u, layerDescription_vector[counter + 1].e, layerDescription_vector[counter + 1].u, layerDescription_vector[counter].r, mode.get_lambda(), mode.get_order(), mode.get_scaling(), kSECOND);
		else
			m = m * M(mode.get_nEff(), layerDescription_vector[counter].e, layerDescription_vector[counter].u, layerDescription_vector[counter + 1].e, layerDescription_vector[counter + 1].u, layerDescription_vector[counter].r, mode.get_lambda(), mode.get_order(), mode.get_scaling(), kNONE);

	std::complex<numT> D_B = -(m(1,1) / m(1,3));
	std::complex<numT> factor = std::sqrt((m(0,1) + m(0,3) * D_B) / (m(2,1) + m(2,3) * D_B));

 	matrix<std::complex<numT> > coeffs(4,1);
	coeffs(0)	= factor;
	coeffs(2)	= static_cast<numT>(1.0) / factor;

	return coeffs;
}

template <class numT>
matrix<std::complex<numT> > modeModelHE<numT>::M(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition)
{
	numT lambda = _lambda * _scaling;
	numT r = _r * _scaling;
	numT Z0 = static_cast<numT>(119.9169832 * M_PI) * _scaling * _scaling;
	numT k0 = static_cast<numT>(2.0 * M_PI) / lambda;
	std::complex<numT> n1 = e1 * u1;
	std::complex<numT> n2 = e2 * u2;
	std::complex<numT> i(static_cast<numT>(0.0), static_cast<numT>(1.0));
	std::complex<numT> beta = nEff * k0;
	std::complex<numT> gamma1;
	std::complex<numT> gamma2;
	std::complex<numT> C1, C2;
	std::complex<numT> C1_p, C2_p;
	std::complex<numT> D1, D2;
	std::complex<numT> D1_p, D2_p;
	std::complex<numT> sign1, sign2;
	std::complex<numT> W;

	if (std::real(n1) > std::real(nEff))
	{
		if (layerPosition != kFIRST)
		{
			gamma1 = std::sqrt(k0 * k0 * (n1 * n1 - nEff * nEff));
	
			besselj(doubleValue(order), gamma1 * r, C1, C1_p);
 			C1_p  = order / (gamma1 * r) * C1 - C1_p;
	
			bessely(doubleValue(order), gamma1 * r, D1, D1_p);
 			D1_p  = order / (gamma1 * r) * D1 - D1_p;
			
			sign1 = static_cast<numT>(1.0);
			W = static_cast<numT>(2.0) / (static_cast<numT>(M_PI) * gamma1 * r);
		}
		else
		{
			gamma1 = std::sqrt(k0 * k0 * (n1 * n1 - nEff * nEff));
	
			besselh2(doubleValue(order), gamma1 * r, C1, C1_p);
 			C1_p  = order / (gamma1 * r) * C1 - C1_p;
	
			besselh1(doubleValue(order), gamma1 * r, D1, D1_p);
 			D1_p  = order / (gamma1 * r) * D1 - D1_p;
			
			sign1 = static_cast<numT>(1.0);
			W = static_cast<numT>(-4.0) * i / (static_cast<numT>(M_PI) * gamma1 * r);
		}
	}
	else
	{
		gamma1 = std::sqrt(k0 * k0 * (nEff * nEff - n1 * n1));

		besseli(doubleValue(order), gamma1 * r, C1, C1_p);
		C1_p  = order / (gamma1 * r) * C1 + C1_p;

		besselk(doubleValue(order), gamma1 * r, D1, D1_p);
		D1_p  = order / (gamma1 * r) * D1 - D1_p;
		
		sign1 = static_cast<numT>(-1.0);
		W = static_cast<numT>(-1.0) / (gamma1 * r);
	}

	if (std::real(n2) > std::real(nEff))
	{
		if (layerPosition != kSECOND)
		{
			gamma2 = std::sqrt(k0 * k0 * (n2 * n2 - nEff * nEff));
	
			besselj(doubleValue(order), gamma2 * r, C2, C2_p);
 			C2_p  = order / (gamma2 * r) * C2 - C2_p;
	
			bessely(doubleValue(order), gamma2 * r, D2, D2_p);
 			D2_p  = order / (gamma2 * r) * D2 - D2_p;
			
			sign2 = static_cast<numT>(1.0);
		}
		else
		{
			gamma2 = std::sqrt(k0 * k0 * (n2 * n2 - nEff * nEff));
	
			besselh2(doubleValue(order), gamma2 * r, C2, C2_p);
 			C2_p  = order / (gamma2 * r) * C2 - C2_p;
	
			besselh1(doubleValue(order), gamma2 * r, D2, D2_p);
 			D2_p  = order / (gamma2 * r) * D2 - D2_p;
			
			sign2 = static_cast<numT>(1.0);
		}
	}
	else
	{
		gamma2 = std::sqrt(k0 * k0 * (nEff * nEff - n2 * n2));

		besseli(doubleValue(order), gamma2 * r, C2, C2_p);
		C2_p  = order / (gamma2 * r) * C2 + C2_p;

		besselk(doubleValue(order), gamma2 * r, D2, D2_p);
		D2_p  = order / (gamma2 * r) * D2 - D2_p;
		
		sign2 = static_cast<numT>(-1.0);
	}

	std::complex<numT> sgammaRatio = (sign1 * gamma1) / (sign2 * gamma2);
	std::complex<numT> overSgammaDiff = order * nEff / (i * r) * (sign1 * gamma1 / (sign2 * gamma2 * gamma2) - static_cast<numT>(1.0) / gamma1);

	std::complex<numT> fa = (e2 * e2) / (e1 * e1) * sgammaRatio;
	std::complex<numT> fb = Z0 / (e1 * e1) * overSgammaDiff;
	std::complex<numT> fc = static_cast<numT>(-1.0) / (Z0 * u1 * u1) * overSgammaDiff;
	std::complex<numT> fd = (u2 * u2) / (u1 * u1) * sgammaRatio;
		
	matrix<std::complex<numT> > m(4,4);
	
	m(0,0) = (C2 * D1_p - fa * C2_p * D1) / W;
	m(0,1) = (D2 * D1_p - fa * D2_p * D1) / W;
	m(0,2) = (fb * C2 * D1) / W;
	m(0,3) = (fb * D2 * D1) / W;
	
	m(1,0) = (fa * C2_p * C1 - C2 * C1_p) / W;
	m(1,1) = (fa * D2_p * C1 - D2 * C1_p) / W;
	m(1,2) = (-fb * C2 * C1) / W;
	m(1,3) = (-fb * D2 * C1) / W;
	
	m(2,0) = (fc * C2 * D1) / W;
	m(2,1) = (fc * D2 * D1) / W;
	m(2,2) = (C2 * D1_p - fd * C2_p * D1) / W;
	m(2,3) = (D2 * D1_p - fd * D2_p * D1) / W;
	
	m(3,0) = (-fc * C2 * C1) / W;
	m(3,1) = (-fc * D2 * C1) / W;
	m(3,2) = (fd * C2_p * C1 - C2 * C1_p) / W;
	m(3,3) = (fd * D2_p * C1 - D2 * C1_p) / W;

	return m;
}

template <class numT>
void modeModelHE<numT>::ensure_bound_modes_in_outermost_layer(std::vector<layerDescription<numT> >& layers)
{
	layers[layers.size() - 1].fieldExpansionCoefficients(0,0) = static_cast<numT>(0.0);
	layers[layers.size() - 1].fieldExpansionCoefficients(2,0) = static_cast<numT>(0.0);
}

template <class numT>
matrix<AD<std::complex<numT> > > modeModelHE<numT>::MAD(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition)
{
	std::complex<numT> n1(e1 * u1);
	std::complex<numT> n2(e2 * u2);
	AD<std::complex<numT> > e1_AD(e1);
	AD<std::complex<numT> > e2_AD(e2);
	AD<std::complex<numT> > u1_AD(u1);
	AD<std::complex<numT> > u2_AD(u2);
	AD<std::complex<numT> > n1_AD(e1 * u1);
	AD<std::complex<numT> > n2_AD(e2 * u2);
	AD<std::complex<numT> > nEff_AD(nEff, static_cast<numT>(1.0));
	AD<std::complex<numT> > r_AD(_r * _scaling);
	AD<std::complex<numT> > lambda_AD(_lambda * _scaling);
	AD<std::complex<numT> > k0_AD = AD<std::complex<numT> >(static_cast<numT>(2.0 * M_PI)) / lambda_AD;
	AD<std::complex<numT> > beta_AD = nEff_AD * k0_AD;
	AD<std::complex<numT> > order_AD(order);
	AD<std::complex<numT> > Z0_AD = AD<std::complex<numT> >(static_cast<numT>(119.9169832 * M_PI) * _scaling * _scaling);

	AD<std::complex<numT> > i_AD(std::complex<numT>(static_cast<numT>(0.0), static_cast<numT>(1.0)));

	AD<std::complex<numT> > gamma1_AD;
	AD<std::complex<numT> > gamma2_AD;
	AD<std::complex<numT> > C1_AD, C2_AD;
	AD<std::complex<numT> > C1_p_AD, C2_p_AD;
	AD<std::complex<numT> > D1_AD, D2_AD;
	AD<std::complex<numT> > D1_p_AD, D2_p_AD;
	AD<std::complex<numT> > sign1_AD, sign2_AD;
	AD<std::complex<numT> > W_AD;

	if (std::real(n1) > std::real(nEff))
	{
		gamma1_AD = std::sqrt(k0_AD * k0_AD * (n1_AD * n1_AD - nEff_AD * nEff_AD));
		sign1_AD = AD<std::complex<numT> >(static_cast<numT>(1.0));

		if (layerPosition != kFIRST)
		{
			AD<std::complex<numT> > arg1(gamma1_AD * r_AD);
			std::complex<numT> C1, C1_p1;
			besselj(doubleValue(order), arg1.fun(), C1, C1_p1);
			std::complex<numT> C1p   = order / arg1.fun() * C1 - C1_p1;
			std::complex<numT> C1_p  = arg1.diff() * C1p;
			std::complex<numT> C1p_p = (arg1.diff() * (-order * C1 + order * order * C1 - C1 * arg1.fun() * arg1.fun() + C1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
			C1_AD    = AD<std::complex<numT> >(C1, C1_p);
			C1_p_AD  = AD<std::complex<numT> >(C1p, C1p_p);

			std::complex<numT> D1, D1_p1;
			bessely(doubleValue(order), arg1.fun(), D1, D1_p1);
			std::complex<numT> D1p   = order / arg1.fun() * D1 - D1_p1;
			std::complex<numT> D1_p  = arg1.diff() * D1p;
			std::complex<numT> D1p_p = (arg1.diff() * (-order * D1 + order * order * D1 - D1 * arg1.fun() * arg1.fun() + D1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
			D1_AD    = AD<std::complex<numT> >(D1, D1_p);
			D1_p_AD  = AD<std::complex<numT> >(D1p, D1p_p);

			W_AD = AD<std::complex<numT> >(static_cast<numT>(2.0) / static_cast<numT>(M_PI)) / (gamma1_AD * r_AD);
		}
		else
		{
			AD<std::complex<numT> > arg1(gamma1_AD * r_AD);
			std::complex<numT> C1, C1_p1;
			besselh2(doubleValue(order), arg1.fun(), C1, C1_p1);
			std::complex<numT> C1p   = order / arg1.fun() * C1 - C1_p1;
			std::complex<numT> C1_p  = arg1.diff() * C1p;
			std::complex<numT> C1p_p = (arg1.diff() * (-order * C1 + order * order * C1 - C1 * arg1.fun() * arg1.fun() + C1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());

			C1_AD    = AD<std::complex<numT> >(C1, C1_p);
			C1_p_AD  = AD<std::complex<numT> >(C1p, C1p_p);
	
			std::complex<numT> D1, D1_p1;
			besselh1(doubleValue(order), arg1.fun(), D1, D1_p1);
			std::complex<numT> D1p   = order / arg1.fun() * D1 - D1_p1;
			std::complex<numT> D1_p  = arg1.diff() * D1p;
			std::complex<numT> D1p_p = (arg1.diff() * (-order * D1 + order * order * D1 - D1 * arg1.fun() * arg1.fun() + D1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
			D1_AD    = AD<std::complex<numT> >(D1, D1_p);
			D1_p_AD  = AD<std::complex<numT> >(D1p, D1p_p);
			
			W_AD = AD<std::complex<numT> >(static_cast<numT>(-4.0) / static_cast<numT>(M_PI)) * i_AD / (gamma1_AD * r_AD);
		}
	}
	else
	{
		gamma1_AD = std::sqrt(k0_AD * k0_AD * (nEff_AD * nEff_AD - n1_AD * n1_AD));
		sign1_AD = AD<std::complex<numT> >(static_cast<numT>(-1.0));

		AD<std::complex<numT> > arg1(gamma1_AD * r_AD);
		std::complex<numT> C1, C1_p1;
		besseli(doubleValue(order), arg1.fun(), C1, C1_p1);
		std::complex<numT> C1p   = order / arg1.fun() * C1 + C1_p1;
		std::complex<numT> C1_p  = arg1.diff() * C1p;
		std::complex<numT> C1p_p = (arg1.diff() * (C1 * arg1.fun() * arg1.fun() - C1_p1 * arg1.fun() - order * C1 + order * order * C1)) / (arg1.fun() * arg1.fun());
		C1_AD    = AD<std::complex<numT> >(C1, C1_p);
		C1_p_AD  = AD<std::complex<numT> >(C1p, C1p_p);

		std::complex<numT> D1, D1_p1;
		besselk(doubleValue(order), arg1.fun(), D1, D1_p1);
		std::complex<numT> D1p   = order / arg1.fun() * D1 - D1_p1;		
		std::complex<numT> D1_p  = arg1.diff() * D1p;
		std::complex<numT> D1p_p = (arg1.diff() * (-order * D1 + order * order * D1 + D1 * arg1.fun() * arg1.fun() + D1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
		D1_AD    = AD<std::complex<numT> >(D1, D1_p);
		D1_p_AD  = AD<std::complex<numT> >(D1p, D1p_p);
		
		W_AD = AD<std::complex<numT> >(static_cast<numT>(-1.0)) / (gamma1_AD * r_AD);
	}

	if (std::real(n2) > std::real(nEff))
	{
		gamma2_AD = std::sqrt(k0_AD * k0_AD * (n2_AD * n2_AD - nEff_AD * nEff_AD));
		sign2_AD = AD<std::complex<numT> >(static_cast<numT>(1.0));

		if (layerPosition != kSECOND)
		{	
			AD<std::complex<numT> > arg2(gamma2_AD * r_AD);
			std::complex<numT> C2, C2_p1;
			besselj(doubleValue(order), arg2.fun(), C2, C2_p1);
			std::complex<numT> C2p   = order / arg2.fun() * C2 - C2_p1;
			std::complex<numT> C2_p  = arg2.diff() * C2p;
			std::complex<numT> C2p_p = (arg2.diff() * (-order * C2 + order * order * C2 - C2 * arg2.fun() * arg2.fun() + C2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			C2_AD    = AD<std::complex<numT> >(C2, C2_p);
			C2_p_AD  = AD<std::complex<numT> >(C2p, C2p_p);
	
			std::complex<numT> D2, D2_p1;
			bessely(doubleValue(order), arg2.fun(), D2, D2_p1);
			std::complex<numT> D2p   = order / arg2.fun() * D2 - D2_p1;
			std::complex<numT> D2_p  = arg2.diff() * D2p;
			std::complex<numT> D2p_p = (arg2.diff() * (-order * D2 + order * order * D2 - D2 * arg2.fun() * arg2.fun() + D2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			D2_AD    = AD<std::complex<numT> >(D2, D2_p);
			D2_p_AD  = AD<std::complex<numT> >(D2p, D2p_p);
		}
		else
		{
			AD<std::complex<numT> > arg2(gamma2_AD * r_AD);
			std::complex<numT> C2, C2_p1;
			besselh2(doubleValue(order), arg2.fun(), C2, C2_p1);
			std::complex<numT> C2p   = order / arg2.fun() * C2 - C2_p1;
			std::complex<numT> C2_p  = arg2.diff() * C2p;
			std::complex<numT> C2p_p = (arg2.diff() * (-order * C2 + order * order * C2 - C2 * arg2.fun() * arg2.fun() + C2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());

			C2_AD    = AD<std::complex<numT> >(C2, C2_p);
			C2_p_AD  = AD<std::complex<numT> >(C2p, C2p_p);
	
			std::complex<numT> D2, D2_p1;
			besselh1(doubleValue(order), arg2.fun(), D2, D2_p1);
			std::complex<numT> D2p   = order / arg2.fun() * D2 - D2_p1;
			std::complex<numT> D2_p  = arg2.diff() * D2p;
			std::complex<numT> D2p_p = (arg2.diff() * (-order * D2 + order * order * D2 - D2 * arg2.fun() * arg2.fun() + D2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			D2_AD    = AD<std::complex<numT> >(D2, D2_p);
			D2_p_AD  = AD<std::complex<numT> >(D2p, D2p_p);
		}
	}
	else
	{
		gamma2_AD = std::sqrt(k0_AD * k0_AD * (nEff_AD * nEff_AD - n2_AD * n2_AD));
		sign2_AD = AD<std::complex<numT> >(static_cast<numT>(-1.0));

		AD<std::complex<numT> > arg2(gamma2_AD * r_AD);
		std::complex<numT> C2, C2_p1;
		besseli(doubleValue(order), arg2.fun(), C2, C2_p1);
		std::complex<numT> C2p   = order / arg2.fun() * C2 + C2_p1;
		std::complex<numT> C2_p  = arg2.diff() * C2p;
		std::complex<numT> C2p_p = (arg2.diff() * (C2 * arg2.fun() * arg2.fun() - C2_p1 * arg2.fun() - order * C2 + order * order * C2)) / (arg2.fun() * arg2.fun());
		C2_AD    = AD<std::complex<numT> >(C2, C2_p);
		C2_p_AD  = AD<std::complex<numT> >(C2p, C2p_p);

		std::complex<numT> D2, D2_p1;
		besselk(doubleValue(order), arg2.fun(), D2, D2_p1);
		std::complex<numT> D2p   = order / arg2.fun() * D2 - D2_p1;		
		std::complex<numT> D2_p  = arg2.diff() * D2p;
		std::complex<numT> D2p_p = (arg2.diff() * (-order * D2 + order * order * D2 + D2 * arg2.fun() * arg2.fun() + D2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
		D2_AD    = AD<std::complex<numT> >(D2, D2_p);
		D2_p_AD  = AD<std::complex<numT> >(D2p, D2p_p);
	}

	AD<std::complex<numT> > sgammaRatio_AD = (sign1_AD * gamma1_AD) / (sign2_AD * gamma2_AD);
	AD<std::complex<numT> > overSgammaDiff_AD = order_AD * nEff_AD / (i_AD * r_AD) * (sign1_AD * gamma1_AD / (sign2_AD * gamma2_AD * gamma2_AD) - AD<std::complex<numT> >(static_cast<numT>(1.0)) / gamma1_AD);

	AD<std::complex<numT> > fa_AD = (e2_AD * e2_AD) / (e1_AD * e1_AD) * sgammaRatio_AD;
	AD<std::complex<numT> > fb_AD = Z0_AD / (e1_AD * e1_AD) * overSgammaDiff_AD;
	AD<std::complex<numT> > fc_AD = AD<std::complex<numT> >(static_cast<numT>(-1.0)) / (Z0_AD * u1_AD * u1_AD) * overSgammaDiff_AD;
	AD<std::complex<numT> > fd_AD = (u2_AD * u2_AD) / (u1_AD * u1_AD) * sgammaRatio_AD;
	
	matrix<AD<std::complex<numT> > > m(4,4);
	
	m(0,0) = (C2_AD * D1_p_AD - fa_AD * C2_p_AD * D1_AD) / W_AD;
	m(0,1) = (D2_AD * D1_p_AD - fa_AD * D2_p_AD * D1_AD) / W_AD;
	m(0,2) = (fb_AD * C2_AD * D1_AD) / W_AD;
	m(0,3) = (fb_AD * D2_AD * D1_AD) / W_AD;
	
	m(1,0) = (fa_AD * C2_p_AD * C1_AD - C2_AD * C1_p_AD) / W_AD;
	m(1,1) = (fa_AD * D2_p_AD * C1_AD - D2_AD * C1_p_AD) / W_AD;
	m(1,2) = (-fb_AD * C2_AD * C1_AD) / W_AD;
	m(1,3) = (-fb_AD * D2_AD * C1_AD) / W_AD;

	m(2,0) = (fc_AD * C2_AD * D1_AD) / W_AD;
	m(2,1) = (fc_AD * D2_AD * D1_AD) / W_AD;
	m(2,2) = (C2_AD * D1_p_AD - fd_AD * C2_p_AD * D1_AD) / W_AD;
	m(2,3) = (D2_AD * D1_p_AD - fd_AD * D2_p_AD * D1_AD) / W_AD;
	
	m(3,0) = (-fc_AD * C2_AD * C1_AD) / W_AD;
	m(3,1) = (-fc_AD * D2_AD * C1_AD) / W_AD;
	m(3,2) = (fd_AD * C2_p_AD * C1_AD - C2_AD * C1_p_AD) / W_AD;
	m(3,3) = (fd_AD * D2_p_AD * C1_AD - D2_AD * C1_p_AD) / W_AD;
	
	return m;
}

template <class numT>
void modeModelHE<numT>::field_radial(const numT& radius, const std::complex<numT>& nEff, const std::complex<numT>& e, const std::complex<numT>& u, const matrix<std::complex<numT> >& fieldExpansionCoefficients, const numT& _lambda, const numT& order, const numT& _scaling, std::complex<numT>* E_phi, std::complex<numT>* E_r, std::complex<numT>* E_z, std::complex<numT>* H_phi, std::complex<numT>* H_r, std::complex<numT>* H_z)
{
	numT lambda = _lambda * _scaling;
	numT r = radius * _scaling;
	numT Z0 = static_cast<numT>(119.9169832 * M_PI) * _scaling * _scaling;
	numT k0 = static_cast<numT>(2.0 * M_PI) / lambda;
	std::complex<numT> n(e * u);
	std::complex<numT> i(static_cast<numT>(0.0), static_cast<numT>(1.0));
	std::complex<numT> beta = nEff * k0;
	std::complex<numT> gamma;
	std::complex<numT> C = static_cast<numT>(0.0);
	std::complex<numT> C_p = static_cast<numT>(0.0);
	std::complex<numT> D = static_cast<numT>(0.0);
	std::complex<numT> D_p = static_cast<numT>(0.0);
	std::complex<numT> sign;
	
	if (std::real(n) > std::real(nEff))
	{
		gamma = std::sqrt(k0 * k0 * (n * n - nEff * nEff));

		if ((std::abs(fieldExpansionCoefficients(0)) != 0.0) && (std::abs(fieldExpansionCoefficients(2)) != 0.0))
		{
			besselj(doubleValue(order), gamma * r, C, C_p);
			C_p = order / (gamma * r) * C - C_p;
		}

		if ((std::abs(fieldExpansionCoefficients(1)) != 0.0) && (std::abs(fieldExpansionCoefficients(3)) != 0.0))
		{
			bessely(doubleValue(order), gamma * r, D, D_p);
			D_p = order / (gamma * r) * D - D_p;
		}
		
		sign = static_cast<numT>(1.0);
	}
	else
	{
		gamma = std::sqrt(k0 * k0 * (nEff * nEff - n * n));

		if ((std::abs(fieldExpansionCoefficients(0)) != 0.0) && (std::abs(fieldExpansionCoefficients(2)) != 0.0))
		{
			besseli(doubleValue(order), gamma * r, C, C_p);
			C_p = order / (gamma * r) * C + C_p;
		}
		
		if ((std::abs(fieldExpansionCoefficients(1)) != 0.0) && (std::abs(fieldExpansionCoefficients(3)) != 0.0))
		{
			besselk(doubleValue(order), gamma * r, D, D_p);
			D_p = order / (gamma * r) * D - D_p;
		}
		
		sign = static_cast<numT>(-1.0);
	}
	
	std::complex<numT> ibeta_gammagamma = sign * i * beta / (gamma * gamma);
	std::complex<numT> k0Z0uu_beta = k0 * Z0 * u * u / beta;
	std::complex<numT> k0ee_betaZ0 = k0 * e * e / (beta * Z0);

	std::complex<numT> dE_z_dr = gamma * (fieldExpansionCoefficients(0) * C_p + fieldExpansionCoefficients(1) * D_p);
	std::complex<numT> dE_z_rdphi = i * order / r * (fieldExpansionCoefficients(0) * C + fieldExpansionCoefficients(1) * D);
	std::complex<numT> dH_z_dr = gamma * (fieldExpansionCoefficients(2) * C_p + fieldExpansionCoefficients(3) * D_p);
	std::complex<numT> dH_z_rdphi = i * order / r * (fieldExpansionCoefficients(2) * C + fieldExpansionCoefficients(3) * D);
	
	if (E_phi != 0)
		*E_phi = ibeta_gammagamma * (-k0Z0uu_beta * dH_z_dr + dE_z_rdphi);

	if (E_r != 0)
		*E_r = ibeta_gammagamma * (dE_z_dr + k0Z0uu_beta * dH_z_rdphi);
	
	if (E_z != 0)
		*E_z = fieldExpansionCoefficients(0) * C + fieldExpansionCoefficients(1) * D;

	if (H_phi != 0)
		*H_phi = ibeta_gammagamma * (k0ee_betaZ0 * dE_z_dr + dH_z_rdphi);

	if (H_r != 0)
		*H_r = ibeta_gammagamma * (dH_z_dr - k0ee_betaZ0 * dE_z_rdphi);

	if (H_z != 0)
		*H_z = i * (fieldExpansionCoefficients(2) * C + fieldExpansionCoefficients(3) * D);
}

template <class numT>
struct couplingIntegralRadialHE_parameters
{
	mode<modeModelHE<numT>, numT>& mode1;
	mode<modeModelHE<numT>, numT>& mode2;
};

template <class numT>
std::complex<numT> coupling_integral_radial_subfunctionHE(const numT& r, void* param)
{
	couplingIntegralRadialHE_parameters<numT>* _param = (couplingIntegralRadialHE_parameters<numT>*)param;
	mode<modeModelHE<numT>, numT>& mode1 = _param->mode1;
	mode<modeModelHE<numT>, numT>& mode2 = _param->mode2;
	layerDescription<numT> layerMode1 = mode1.get_layer_description_for_radius(r);
	layerDescription<numT> layerMode2 = mode2.get_layer_description_for_radius(r);
	numT scaling = mode1.get_scaling();
	
	std::complex<numT> E1_r, E1_phi, E2_r, E2_phi;
	
	mode1.field_radial(r, mode1.get_nEff(), layerMode1.e, layerMode1.u, layerMode1.fieldExpansionCoefficients, mode1.get_lambda(), mode1.get_core_mode_order(), scaling, &E1_phi, &E1_r, 0, 0, 0, 0);
	mode2.field_radial(r, mode2.get_nEff(), layerMode2.e, layerMode2.u, layerMode2.fieldExpansionCoefficients, mode2.get_lambda(), mode2.get_core_mode_order(), scaling, &E2_phi, &E2_r, 0, 0, 0, 0);
	
	return (E1_r * std::conj(E2_r) + E1_phi * std::conj(E2_phi)) * r * scaling;
}

template <class numT>
std::complex<numT> coupling_integral_radial(mode<modeModelHE<numT>, numT>& mode1, mode<modeModelHE<numT>, numT>& mode2, const numT& r_start, const numT& r_end)
{
	unsigned int check;
	couplingIntegralRadialHE_parameters<numT> param = {mode1, mode2};
	return integral(&coupling_integral_radial_subfunctionHE, &param, r_start, r_end, check);
}
#endif

