/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _MODEMODELLP_H
#define _MODEMODELLP_H 1

#include <complex>
#include "algorithms.h"
#include "ad.h"
#include "matrix_template.h"
#include "bessel_amos.h"
#include "mode.h"
#include "modeSolver.h"

template <class numT>
class modeModelLP
{
	private:
		enum outermostLayerPosition
		{
			kNONE,
			kFIRST,
			kSECOND
		};
		
	public:
		// Start Interface for modeSolver
		static numT get_core_mode_order(void);
		static void characteristic_equation(const std::complex<numT>& nEff, void* param, std::complex<numT>* f, std::complex<numT>* fDiff = NULL);
		// End Interface for modeSolver
		// Start Interface for mode
		static std::complex<numT> normalisation_integration_function(const numT& r, void* param);
		static numT normalisation_prefactor(const mode<modeModelLP<numT>, numT>& mode);
		static matrix<std::complex<numT> > field_expansion_coefficients_centre(const mode<modeModelLP<numT>, numT>& mode);
		static matrix<std::complex<numT> > M(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition = kNONE);
		void ensure_bound_modes_in_outermost_layer(std::vector<layerDescription<numT> >& layers);
		// End Interface for mode
		static matrix<AD<std::complex<numT> > > MAD(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition = kNONE);
		static void field_radial(const numT& radius, const std::complex<numT>& nEff, const std::complex<numT>& e, const std::complex<numT>& u, const matrix<std::complex<numT> >& fieldExpansionCoefficients, const numT& _lambda, const numT& order, const numT& _scaling, std::complex<numT>* R);
};

template <class numT>
numT modeModelLP<numT>::get_core_mode_order(void)
{
	return static_cast<numT>(0.0);
}

template <class numT>
void modeModelLP<numT>::characteristic_equation(const std::complex<numT>& nEff, void* param, std::complex<numT>* f, std::complex<numT>* fDiff)
{
	modeSolver<modeModelLP<numT>, numT>* solver = (modeSolver<modeModelLP<numT>, numT>*)param;
	numT lambda_double = solver->get_lambda_double();
	std::vector<layerDescription<numT> > layers_vector = solver->get_layers_vector();
	unsigned int modeOrder_uint = static_cast<unsigned int>(intValue(get_core_mode_order()));
	numT scaling_double = solver->get_scaling_double();
	
	if (fDiff != NULL)
	{
		matrix<AD<std::complex<numT> > > m(2, 2, AD<std::complex<numT> >(static_cast<numT>(1.0)), matrix<AD<std::complex<numT> > >::diagonal_fillPattern);
		
		for (int counter = 0; counter < layers_vector.size() - 1; counter++)
 			if (counter == layers_vector.size() - 2)
	 			m = m * MAD(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, modeOrder_uint, scaling_double, kSECOND);
 			else
	 			m = m * MAD(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, modeOrder_uint, scaling_double, kNONE);
		
		AD<std::complex<numT> > result(m(1,1));

		*f = result.fun();
		*fDiff = result.diff();
	}
	else
	{
		matrix<std::complex<numT> > m(2, 2, std::complex<numT>(static_cast<numT>(1.0)), matrix<std::complex<numT> >::diagonal_fillPattern);
		
		for (int counter = 0; counter < layers_vector.size() - 1; counter++)
			if (counter == layers_vector.size() - 2)
				m = m * M(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, modeOrder_uint, scaling_double, kSECOND);
			else
				m = m * M(nEff, layers_vector[counter].e, layers_vector[counter].u, layers_vector[counter + 1].e, layers_vector[counter + 1].u, layers_vector[counter].r, lambda_double, modeOrder_uint, scaling_double, kNONE);
		
		*f = m(1,1);
	}

}

template <class numT>
std::complex<numT> modeModelLP<numT>::normalisation_integration_function(const numT& r, void* param)
{
	mode<modeModelLP<numT>, numT>* theMode = (mode<modeModelLP<numT>, numT>*)param;

	if (!theMode->are_field_expansion_coeffcients_set())
		theMode->set_field_expansion_coefficients();
	
	std::complex<numT> R;
	
	layerDescription<numT> layer = theMode->get_layer_description_for_radius(r);

	field_radial(r, theMode->get_nEff(), layer.e, layer.u, layer.fieldExpansionCoefficients, theMode->get_lambda(), get_core_mode_order(), theMode->get_scaling(), &R);
	return R * std::conj(R) * r * theMode->get_scaling();
}

template <class numT>
numT modeModelLP<numT>::normalisation_prefactor(const mode<modeModelLP<numT>, numT>& mode)
{
	return static_cast<numT>(M_PI) * std::real(mode.get_nEff()) / std::real(std::complex<numT>(kC * kMU_0) * mode.get_scaling() * mode.get_scaling());
}

template <class numT>
matrix<std::complex<numT> > modeModelLP<numT>::field_expansion_coefficients_centre(const mode<modeModelLP<numT>, numT>& mode)
{
 	matrix<std::complex<numT> > coeffs(2);
	coeffs(0)	= static_cast<numT>(1.0);
	coeffs(1)	= static_cast<numT>(0.0);

	return coeffs;
}

template <class numT>
matrix<std::complex<numT> > modeModelLP<numT>::M(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition)
{
	numT lambda = _lambda * _scaling;
	numT r = _r * _scaling;
	numT k0 = static_cast<numT>(2.0 * M_PI) / lambda;
	std::complex<numT> n1 = e1 * u1;
	std::complex<numT> n2 = e2 * u2;
	std::complex<numT> beta = nEff * k0;
	std::complex<numT> gamma1;
	std::complex<numT> gamma2;
	std::complex<numT> C1, C2;
	std::complex<numT> C1_p, C2_p;
	std::complex<numT> D1, D2;
	std::complex<numT> D1_p, D2_p;
	std::complex<numT> q;
	std::complex<numT> i(static_cast<numT>(0.0), static_cast<numT>(1.0));

	if (std::real(n1) > std::real(nEff))
	{
		if (layerPosition != kFIRST)
		{
			gamma1 = std::sqrt(k0 * k0 * (n1 * n1 - nEff * nEff));
	
			besselj(doubleValue(order), gamma1 * r, C1, C1_p);
			C1_p = order / (gamma1 * r) * C1 - C1_p;
	
			bessely(doubleValue(order), gamma1 * r, D1, D1_p);
			D1_p = order / (gamma1 * r) * D1 - D1_p;
			
			q = static_cast<numT>(2.0) / (static_cast<numT>(M_PI) * r);
		}
		else
		{
			gamma1 = std::sqrt(k0 * k0 * (n1 * n1 - nEff * nEff));
	
			besselh2(doubleValue(order), gamma1 * r, C1, C1_p);
			C1_p = order / (gamma1 * r) * C1 - C1_p;
	
			besselh1(doubleValue(order), gamma1 * r, D1, D1_p);
			D1_p = order / (gamma1 * r) * D1 - D1_p;
			
			q = static_cast<numT>(-4.0) * i / (static_cast<numT>(M_PI) * r);
		}
	}
	else
	{
		gamma1 = std::sqrt(k0 * k0 * (nEff * nEff - n1 * n1));

		besseli(doubleValue(order), gamma1 * r, C1, C1_p);
		C1_p = order / (gamma1 * r) * C1 + C1_p;

		besselk(doubleValue(order), gamma1 * r, D1, D1_p);
		D1_p = order / (gamma1 * r) * D1 - D1_p;
		
		q = static_cast<numT>(-1.0) / r;
	}

	if (std::real(n2) > std::real(nEff))
	{
		if (layerPosition != kSECOND)
		{
			gamma2 = std::sqrt(k0 * k0 * (n2 * n2 - nEff * nEff));

			besselj(doubleValue(order), gamma2 * r, C2, C2_p);
			C2_p = order / (gamma2 * r) * C2 - C2_p;
	
			bessely(doubleValue(order), gamma2 * r, D2, D2_p);
			D2_p = order / (gamma2 * r) * D2 - D2_p;
		}
		else
		{
			gamma2 = std::sqrt(k0 * k0 * (n2 * n2 - nEff * nEff));
	
			besselh2(doubleValue(order), gamma2 * r, C2, C2_p);
			C2_p = order / (gamma2 * r) * C2 - C2_p;
	
			besselh1(doubleValue(order), gamma2 * r, D2, D2_p);
			D2_p = order / (gamma2 * r) * D2 - D2_p;
		}
	}
	else
	{
		gamma2 = std::sqrt(k0 * k0 * (nEff * nEff - n2 * n2));

		besseli(doubleValue(order), gamma2 * r, C2, C2_p);
		C2_p = order / (gamma2 * r) * C2 + C2_p;

		besselk(doubleValue(order), gamma2 * r, D2, D2_p);
		D2_p = order / (gamma2 * r) * D2 - D2_p;
	}
	
	matrix<std::complex<numT> > m(2,2);
	
	m(0, 0) = (gamma1 * D1_p * C2 - gamma2 * D1 * C2_p) / q;
	m(0, 1) = (gamma1 * D1_p * D2 - gamma2 * D1 * D2_p) / q;
	m(1, 0) = (-gamma1 * C1_p * C2 + gamma2 * C1 * C2_p) / q;
	m(1, 1) = (-gamma1 * C1_p * D2 + gamma2 * C1 * D2_p) / q;

	return m;
}

template <class numT>
void modeModelLP<numT>::ensure_bound_modes_in_outermost_layer(std::vector<layerDescription<numT> >& layers)
{
	layers[layers.size() - 1].fieldExpansionCoefficients(0) = static_cast<numT>(0.0);
}

template <class numT>
matrix<AD<std::complex<numT> > > modeModelLP<numT>::MAD(const std::complex<numT>& nEff, const std::complex<numT>& e1, const std::complex<numT>& u1, const std::complex<numT>& e2, const std::complex<numT>& u2, const numT& _r, const numT& _lambda, const numT& order, const numT& _scaling, const outermostLayerPosition& layerPosition)
{
	std::complex<numT> n1(e1 * u1);
	std::complex<numT> n2(e2 * u2);
	AD<std::complex<numT> > e1_AD(e1);
	AD<std::complex<numT> > e2_AD(e2);
	AD<std::complex<numT> > u1_AD(u1);
	AD<std::complex<numT> > u2_AD(u2);
	AD<std::complex<numT> > n1_AD(e1 * u1);
	AD<std::complex<numT> > n2_AD(e2 * u2);
	AD<std::complex<numT> > nEff_AD(nEff, static_cast<numT>(1.0));
	AD<std::complex<numT> > r_AD(_r * _scaling);
	AD<std::complex<numT> > lambda_AD(_lambda * _scaling);
	AD<std::complex<numT> > k0_AD = AD<std::complex<numT> >(static_cast<numT>(2.0 * M_PI)) / lambda_AD;
	AD<std::complex<numT> > beta_AD = nEff_AD * k0_AD;
	AD<std::complex<numT> > order_AD(order);

	AD<std::complex<numT> > i_AD(std::complex<numT>(static_cast<numT>(0.0), static_cast<numT>(1.0)));

	AD<std::complex<numT> > gamma1_AD;
	AD<std::complex<numT> > gamma2_AD;
	AD<std::complex<numT> > C1_AD, C2_AD;
	AD<std::complex<numT> > C1_p_AD, C2_p_AD;
	AD<std::complex<numT> > D1_AD, D2_AD;
	AD<std::complex<numT> > D1_p_AD, D2_p_AD;
	AD<std::complex<numT> > q_AD;

	if (std::real(n1) > std::real(nEff))
	{
		gamma1_AD = std::sqrt(k0_AD * k0_AD * (n1_AD * n1_AD - nEff_AD * nEff_AD));

		if (layerPosition != kFIRST)
		{
			AD<std::complex<numT> > arg1(gamma1_AD * r_AD);
			std::complex<numT> C1, C1_p1;
			besselj(doubleValue(order), arg1.fun(), C1, C1_p1);
			std::complex<numT> C1p   = order / arg1.fun() * C1 - C1_p1;
			std::complex<numT> C1_p  = arg1.diff() * C1p;
			std::complex<numT> C1p_p = (arg1.diff() * (-order * C1 + order * order * C1 - C1 * arg1.fun() * arg1.fun() + C1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
			C1_AD    = AD<std::complex<numT> >(C1, C1_p);
			C1_p_AD  = AD<std::complex<numT> >(C1p, C1p_p);

			std::complex<numT> D1, D1_p1;
			bessely(doubleValue(order), arg1.fun(), D1, D1_p1);
			std::complex<numT> D1p   = order / arg1.fun() * D1 - D1_p1;
			std::complex<numT> D1_p  = arg1.diff() * D1p;
			std::complex<numT> D1p_p = (arg1.diff() * (-order * D1 + order * order * D1 - D1 * arg1.fun() * arg1.fun() + D1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
			D1_AD    = AD<std::complex<numT> >(D1, D1_p);
			D1_p_AD  = AD<std::complex<numT> >(D1p, D1p_p);

			q_AD = AD<std::complex<numT> >(static_cast<numT>(2.0 / M_PI)) / r_AD;
		}
		else
		{
			AD<std::complex<numT> > arg1(gamma1_AD * r_AD);
			std::complex<numT> C1, C1_p1;
			besselh2(doubleValue(order), arg1.fun(), C1, C1_p1);
			std::complex<numT> C1p   = order / arg1.fun() * C1 - C1_p1;
			std::complex<numT> C1_p  = arg1.diff() * C1p;
			std::complex<numT> C1p_p = (arg1.diff() * (-order * C1 + order * order * C1 - C1 * arg1.fun() * arg1.fun() + C1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());

			C1_AD    = AD<std::complex<numT> >(C1, C1_p);
			C1_p_AD  = AD<std::complex<numT> >(C1p, C1p_p);
	
			std::complex<numT> D1, D1_p1;
			besselh1(doubleValue(order), arg1.fun(), D1, D1_p1);
			std::complex<numT> D1p   = order / arg1.fun() * D1 - D1_p1;
			std::complex<numT> D1_p  = arg1.diff() * D1p;
			std::complex<numT> D1p_p = (arg1.diff() * (-order * D1 + order * order * D1 - D1 * arg1.fun() * arg1.fun() + D1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
			D1_AD    = AD<std::complex<numT> >(D1, D1_p);
			D1_p_AD  = AD<std::complex<numT> >(D1p, D1p_p);
			
			q_AD = AD<std::complex<numT> >(static_cast<numT>(-4.0 / M_PI)) * i_AD / r_AD;
		}
	}
	else
	{
		gamma1_AD = std::sqrt(k0_AD * k0_AD * (nEff_AD * nEff_AD - n1_AD * n1_AD));
		AD<std::complex<numT> > arg1(gamma1_AD * r_AD);

		std::complex<numT> C1, C1_p1;
		besseli(doubleValue(order), arg1.fun(), C1, C1_p1);
		std::complex<numT> C1p   = order / arg1.fun() * C1 + C1_p1;
		std::complex<numT> C1_p  = arg1.diff() * C1p;
		std::complex<numT> C1p_p = (arg1.diff() * (C1 * arg1.fun() * arg1.fun() - C1_p1 * arg1.fun() - order * C1 + order * order * C1)) / (arg1.fun() * arg1.fun());
		C1_AD    = AD<std::complex<numT> >(C1, C1_p);
		C1_p_AD  = AD<std::complex<numT> >(C1p, C1p_p);

		std::complex<numT> D1, D1_p1;
		besselk(doubleValue(order), arg1.fun(), D1, D1_p1);
		std::complex<numT> D1p   = order / arg1.fun() * D1 - D1_p1;		
		std::complex<numT> D1_p  = arg1.diff() * D1p;
		std::complex<numT> D1p_p = (arg1.diff() * (-order * D1 + order * order * D1 + D1 * arg1.fun() * arg1.fun() + D1_p1 * arg1.fun())) / (arg1.fun() * arg1.fun());
		D1_AD    = AD<std::complex<numT> >(D1, D1_p);
		D1_p_AD  = AD<std::complex<numT> >(D1p, D1p_p);
		
		q_AD = AD<std::complex<numT> >(static_cast<numT>(-1.0)) / r_AD;
	}

	if (std::real(n2) > std::real(nEff))
	{
		gamma2_AD = std::sqrt(k0_AD * k0_AD * (n2_AD * n2_AD - nEff_AD * nEff_AD));

		if (layerPosition != kSECOND)
		{
			AD<std::complex<numT> > arg2(gamma2_AD * r_AD);
			std::complex<numT> C2, C2_p1;
			besselj(doubleValue(order), arg2.fun(), C2, C2_p1);
			std::complex<numT> C2p   = order / arg2.fun() * C2 - C2_p1;
			std::complex<numT> C2_p  = arg2.diff() * C2p;
			std::complex<numT> C2p_p = (arg2.diff() * (-order * C2 + order * order * C2 - C2 * arg2.fun() * arg2.fun() + C2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			C2_AD    = AD<std::complex<numT> >(C2, C2_p);
			C2_p_AD  = AD<std::complex<numT> >(C2p, C2p_p);
	
			std::complex<numT> D2, D2_p1;
			bessely(doubleValue(order), arg2.fun(), D2, D2_p1);
			std::complex<numT> D2p   = order / arg2.fun() * D2 - D2_p1;
			std::complex<numT> D2_p  = arg2.diff() * D2p;
			std::complex<numT> D2p_p = (arg2.diff() * (-order * D2 + order * order * D2 - D2 * arg2.fun() * arg2.fun() + D2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			D2_AD    = AD<std::complex<numT> >(D2, D2_p);
			D2_p_AD  = AD<std::complex<numT> >(D2p, D2p_p);
		}
		else
		{
			AD<std::complex<numT> > arg2(gamma2_AD * r_AD);
			std::complex<numT> C2, C2_p1;
			besselh2(doubleValue(order), arg2.fun(), C2, C2_p1);
			std::complex<numT> C2p   = order / arg2.fun() * C2 - C2_p1;
			std::complex<numT> C2_p  = arg2.diff() * C2p;
			std::complex<numT> C2p_p = (arg2.diff() * (-order * C2 + order * order * C2 - C2 * arg2.fun() * arg2.fun() + C2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			C2_AD    = AD<std::complex<numT> >(C2, C2_p);
			C2_p_AD  = AD<std::complex<numT> >(C2p, C2p_p);
	
			std::complex<numT> D2, D2_p1;
			besselh1(doubleValue(order), arg2.fun(), D2, D2_p1);
			std::complex<numT> D2p   = order / arg2.fun() * D2 - D2_p1;
			std::complex<numT> D2_p  = arg2.diff() * D2p;
			std::complex<numT> D2p_p = (arg2.diff() * (-order * D2 + order * order * D2 - D2 * arg2.fun() * arg2.fun() + D2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
			D2_AD    = AD<std::complex<numT> >(D2, D2_p);
			D2_p_AD  = AD<std::complex<numT> >(D2p, D2p_p);
		}
	}
	else
	{
		gamma2_AD = std::sqrt(k0_AD * k0_AD * (nEff_AD * nEff_AD - n2_AD * n2_AD));

		AD<std::complex<numT> > arg2(gamma2_AD * r_AD);
		std::complex<numT> C2, C2_p1;
		besseli(doubleValue(order), arg2.fun(), C2, C2_p1);
		std::complex<numT> C2p   = order / arg2.fun() * C2 + C2_p1;
		std::complex<numT> C2_p  = arg2.diff() * C2p;
		std::complex<numT> C2p_p = (arg2.diff() * (C2 * arg2.fun() * arg2.fun() - C2_p1 * arg2.fun() - order * C2 + order * order * C2)) / (arg2.fun() * arg2.fun());
		C2_AD    = AD<std::complex<numT> >(C2, C2_p);
		C2_p_AD  = AD<std::complex<numT> >(C2p, C2p_p);

		std::complex<numT> D2, D2_p1;
		besselk(doubleValue(order), arg2.fun(), D2, D2_p1);
		std::complex<numT> D2p   = order / arg2.fun() * D2 - D2_p1;		
		std::complex<numT> D2_p  = arg2.diff() * D2p;
		std::complex<numT> D2p_p = (arg2.diff() * (-order * D2 + order * order * D2 + D2 * arg2.fun() * arg2.fun() + D2_p1 * arg2.fun())) / (arg2.fun() * arg2.fun());
		D2_AD    = AD<std::complex<numT> >(D2, D2_p);
		D2_p_AD  = AD<std::complex<numT> >(D2p, D2p_p);
	}

	matrix<AD<std::complex<numT> > > m(2,2);
	
	m(0, 0) = (gamma1_AD * D1_p_AD * C2_AD - gamma2_AD * D1_AD * C2_p_AD) / q_AD;
	m(0, 1) = (gamma1_AD * D1_p_AD * D2_AD - gamma2_AD * D1_AD * D2_p_AD) / q_AD;
	m(1, 0) = (-gamma1_AD * C1_p_AD * C2_AD + gamma2_AD * C1_AD * C2_p_AD) / q_AD;
	m(1, 1) = (-gamma1_AD * C1_p_AD * D2_AD + gamma2_AD * C1_AD * D2_p_AD) / q_AD;

	return m;
}

template <class numT>
void modeModelLP<numT>::field_radial(const numT& radius, const std::complex<numT>& nEff, const std::complex<numT>& e, const std::complex<numT>& u, const matrix<std::complex<numT> >& fieldExpansionCoefficients, const numT& _lambda, const numT& order, const numT& _scaling, std::complex<numT>* R)
{
	numT lambda = _lambda * _scaling;
	numT r = radius * _scaling;
	numT k0 = static_cast<numT>(2.0 * M_PI) / lambda;
	std::complex<numT> n(e * u);
	std::complex<numT> beta = nEff * k0;
	std::complex<numT> gamma;
	std::complex<numT> C = static_cast<numT>(0.0);
	std::complex<numT> D = static_cast<numT>(0.0);
	
	if (std::real(n) > std::real(nEff))
	{
		gamma = std::sqrt(k0 * k0 * (n * n - nEff * nEff));

		if (std::abs(fieldExpansionCoefficients(0)) != 0.0)
			C = besselj(doubleValue(order), gamma * r);

		if (std::abs(fieldExpansionCoefficients(1)) != 0.0)
			D = bessely(doubleValue(order), gamma * r);
	}
	else
	{
		gamma = std::sqrt(k0 * k0 * (nEff * nEff - n * n));

		if (std::abs(fieldExpansionCoefficients(0)) != 0.0)
			C = besseli(doubleValue(order), gamma * r);
		
		if (std::abs(fieldExpansionCoefficients(1)) != 0.0)
			D = besselk(doubleValue(order), gamma * r);
	}
	
	if (R != 0)
		*R = fieldExpansionCoefficients(0) * C + fieldExpansionCoefficients(1) * D;
}

template <class numT>
struct couplingIntegralRadialLP_parameters
{
	mode<modeModelLP<numT>, numT>& mode1;
	mode<modeModelLP<numT>, numT>& mode2;
};

template <class numT>
std::complex<numT> coupling_integral_radial_subfunctionLP(const numT& r, void* param)
{
	couplingIntegralRadialLP_parameters<numT>* _param = (couplingIntegralRadialLP_parameters<numT>*)param;
	mode<modeModelLP<numT>, numT>& mode1 = _param->mode1;
	mode<modeModelLP<numT>, numT>& mode2 = _param->mode2;
	layerDescription<numT> layerMode1 = mode1.get_layer_description_for_radius(r);
	layerDescription<numT> layerMode2 = mode2.get_layer_description_for_radius(r);
	numT scaling = mode1.get_scaling();
	
	std::complex<numT> R1, R2;
	
	mode1.field_radial(r, mode1.get_nEff(), layerMode1.e, layerMode1.u, layerMode1.fieldExpansionCoefficients, mode1.get_lambda(), mode1.get_core_mode_order(), scaling, &R1);
	mode2.field_radial(r, mode2.get_nEff(), layerMode2.e, layerMode2.u, layerMode2.fieldExpansionCoefficients, mode2.get_lambda(), mode2.get_core_mode_order(), scaling, &R2);
	
	return R1 * std::conj(R2) * r * scaling;
}

template <class numT>
std::complex<numT> coupling_integral_radial(mode<modeModelLP<numT>, numT>& mode1, mode<modeModelLP<numT>, numT>& mode2, const numT& r_start, const numT& r_end)
{
	unsigned int check;
	couplingIntegralRadialLP_parameters<numT> param = {mode1, mode2};
	return integral(&coupling_integral_radial_subfunctionLP, &param, r_start, r_end, check);
}

#endif

