/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _MODESOLVER_H
#define _MODESOLVER_H 1

#include <cmath>
#include <complex>
#include <stdexcept>
#include <vector>
#include <list>
#include <iostream>
#include <fstream>

#include "numberFunctions.h"
#include "algorithms.h"
#include "matrix_template.h"
#include "mode.h"
#include "pointerWrapper_template.h"

// Interface for modeSolver modeModel parent classes
//	public:
// 		static double get_core_mode_order(void);
// 		static void characteristic_equation(const std::complex<double>& nEff, void* param, std::complex<double>* f, std::complex<double>* fDiff = NULL);

template <class modeModel, class numT>
class modeSolver : public modeModel
{
	private:
		std::vector<layerDescription<numT> > _layerDescription_vector;
		numT _lambda_double;
		numT _scaling_double;
		numT _rootThreshold_double;
		numT _imagOffset_double;
		numT _k0Scaled_double;
		std::map<std::wstring, pointerWrapper<std::ofstream> >* _fileMap;
		
	public:
		modeSolver(const std::vector<layerDescription<numT> >& layerDescription_vector, const numT& lambda, const numT& scaling = static_cast<numT>(1e6), const numT& rootThreshold = static_cast<numT>(1e-10), const numT& imagOffset = static_cast<numT>(1e-5), std::map<std::wstring, pointerWrapper<std::ofstream> >* fileMap = 0); // imagOffset also worked well with 1e-5.
		mode<modeModel, numT> determine_core_mode() throw (std::runtime_error);
		std::vector<mode<modeModel, numT> > determine_modes_in_interval(interval<std::complex<numT> > n_interval);
//		std::vector<mode<modeModel> > determine_modes_in_interval_direct(interval<std::complex<numT> > n_interval);
//		void determine_modes_in_interval_direct_sub(interval<std::complex<numT> > n_interval, std::vector<mode<modeModel> >& modes);
		std::list<interval<std::complex<numT> > > determine_roots_in_interval(interval<std::complex<numT> > n_interval, std::vector<layerDescription<numT> >& layerDescriptions_vector);
		std::list<interval<std::complex<numT> > > split_interval_by_layer_real_n(const interval<std::complex<numT> >& n_interval) throw (std::runtime_error);
		static std::complex<numT> characteristic_equation_path_integration_function(const std::complex<numT>& arg, void* param);
		numT get_lambda_double();
		std::vector<layerDescription<numT> > get_layers_vector();
		numT get_scaling_double();
};

template <class modeModel, class numT>
modeSolver<modeModel, numT>::modeSolver(const std::vector<layerDescription<numT> >& layerDescription_vector, const numT& lambda, const numT& scaling, const numT& rootThreshold, const numT& imagOffset, std::map<std::wstring, pointerWrapper<std::ofstream> >* fileMap)
{
	_layerDescription_vector = layerDescription_vector;
	_lambda_double = lambda;
	_scaling_double = scaling;
	_rootThreshold_double = rootThreshold;
	_imagOffset_double = imagOffset;
	_k0Scaled_double = static_cast<numT>(2.0 * M_PI) / (_lambda_double * _scaling_double);
	_fileMap = fileMap;
}

template <class modeModel, class numT>
mode<modeModel, numT> modeSolver<modeModel, numT>::determine_core_mode() throw (std::runtime_error)
{
	// Set interval
	interval<std::complex<numT> > rootTest_interval;
	rootTest_interval.start_T = (_layerDescription_vector[0].n * (static_cast<numT>(1.0) - _rootThreshold_double) + std::complex<numT>(static_cast<numT>(0.0), -_imagOffset_double));
	rootTest_interval.end_T = (_layerDescription_vector[1].n * (static_cast<numT>(1.0) + _rootThreshold_double) + std::complex<numT>(static_cast<numT>(0.0), _imagOffset_double));

	// Check how many roots are in the core interval
// 	std::vector<mode<modeModel> > mode_vector = determine_modes_in_interval(rootTest_interval);
// 	
// 	if (mode_vector.size() == 1)
// 	{
// 		return mode_vector[0];
// 	}
// 	else
// 	{
// 		if ((mode_vector.size() < 1) && (_fileMap != 0) && (_fileMap->count(L"log") > 0))
// 		{
// 			pointerWrapper<std::ofstream> log = _fileMap->find(L"log")->second;
// 			log.lock();
// 			*(log.getObjectPtr()) << "modeSolver::determine_core_mode: No root found with full geometry trying to recover using only the two innermost layers.\n";
// 			log.unlock();
// 		}					
// 		else if ((mode_vector.size() > 1) && (_fileMap != 0) && (_fileMap->count(L"log") > 0))
// 		{
// 			pointerWrapper<std::ofstream> log = _fileMap->find(L"log")->second;
// 			log.lock();
// 			*(log.getObjectPtr()) << "modeSolver::determine_core_mode: Too many roots found with full geometry trying to recover using only the two innermost layers.\n";
// 			log.unlock();
// 		}					
	
		// Construct new fibre geometry with only core and cladding, where the cladding has the same radius as the
		// outermost layer of the original fibre geometry.
		std::vector<layerDescription<numT> > origLayerDescriptions(_layerDescription_vector);
		typename std::vector<layerDescription<numT> >::iterator start = _layerDescription_vector.begin();
		typename std::vector<layerDescription<numT> >::iterator end = ++(++_layerDescription_vector.begin());
		_layerDescription_vector = std::vector<layerDescription<numT> >(start, end);

		// Try direct method
		std::complex<numT> temp = determine_root_in_interval(rootTest_interval, &modeModel::characteristic_equation, this);
		if (!std::isnan(std::real(temp)) && !std::isnan(std::imag(temp)))
		{
			// Reset layerDescription vector
			std::vector<layerDescription<numT> > newLayerDescriptions(_layerDescription_vector);
			_layerDescription_vector = origLayerDescriptions;			

			if (std::abs(std::real(temp)/std::imag(temp)) > static_cast<numT>(1e15))
				temp = std::real(temp);

			return mode<modeModel, numT>(temp, _lambda_double, _scaling_double, newLayerDescriptions);
		}
//	}
	throw std::runtime_error("modeSolver::determine_core_mode: No core mode found.");
}

template <class modeModel, class numT>
std::vector<mode<modeModel, numT> > modeSolver<modeModel, numT>::determine_modes_in_interval(interval<std::complex<numT> > n_interval)
{
	// Debug start
// 	double realStart = 1.44468;
// 	double realEnd = 1.4447 - _rootThreshold_double;
// 	double imagStart = -_imagOffset_double / 10.0;
// 	double imagEnd = _imagOffset_double * 2.0;
// 	double realStart = std::min(std::real(n_interval.start_T), std::real(n_interval.end_T));
// 	double realEnd = std::max(std::real(n_interval.start_T), std::real(n_interval.end_T));
// 	double imagStart = std::min(std::imag(n_interval.start_T), std::imag(n_interval.end_T));
// 	double imagEnd = std::max(std::imag(n_interval.start_T), std::imag(n_interval.end_T));
// 	std::ofstream testFile("debug.csv", std::ios::out);
// 	testFile.precision(16);
// 	for (double realCounter = realStart; realCounter < realEnd; realCounter += (realEnd - realStart) / 250.0)
// 		for (double imagCounter = imagStart; imagCounter < imagEnd; imagCounter += (imagEnd - imagStart) / 250.0)
// 		{
// 			std::complex<double> fun, diff;
// 			modeModel::characteristic_equation(std::complex<double>(realCounter, imagCounter), this, &fun, &diff);
// 			testFile << realCounter << "\t" << imagCounter << "\t" << std::real(fun) << "\t" << std::imag(fun) << "\t" << std::real(diff) << "\t" << std::imag(diff) << "\n";
// 		}
// 	testFile.close();
	// Debug end

	std::list<interval<std::complex<numT> > > interval_list = split_interval_by_layer_real_n(n_interval);
	std::list<interval<std::complex<numT> > > small_interval_list;
	
	for (typename std::list<interval<std::complex<numT> > >::iterator it = interval_list.begin(); it != interval_list.end(); it++)
	{
		numT smallIntervalSize = static_cast<numT>(5e-3);
		int number = intValue(round(std::abs(it->end_T - it->start_T) / smallIntervalSize));
		if (number > 1)
		{
			std::list<interval<std::complex<numT> > > temp = split_interval(number, *it);
			small_interval_list.splice(small_interval_list.begin(), temp);
		}
		else
			small_interval_list.push_back(*it);
	}
	
	interval_list = small_interval_list;
	
	std::list<interval<std::complex<numT> > > mode_interval_list;
	std::vector<mode<modeModel, numT> > mode_vector;

	for (typename std::list<interval<std::complex<numT> > >::iterator it = interval_list.begin(); it != interval_list.end(); it++)
	{
		std::list<interval<std::complex<numT> > > temp = determine_roots_in_interval(*it, _layerDescription_vector);
		mode_interval_list.splice(mode_interval_list.begin(), temp);
	}

	for (typename std::list<interval<std::complex<numT> > >::iterator it = mode_interval_list.begin(); it != mode_interval_list.end(); it++)
	{
		std::complex<numT> temp = determine_root_in_interval(*it, &modeModel::characteristic_equation, this);

		if (!std::isnan(std::real(temp)) && !std::isnan(std::imag(temp)))
		{
			if (std::abs(std::real(temp)/std::imag(temp)) > static_cast<numT>(1e15))
				temp = std::real(temp);

			if (std::imag(temp) >= static_cast<numT>(0.0))
			{
				// Test if we already got this mode.
				bool existsAlready = false;
				for (int modeCounter = 0; (modeCounter < mode_vector.size()) && !existsAlready; modeCounter++)
				{
					existsAlready = thresholdEqual(static_cast<std::complex<numT> >(mode_vector[modeCounter].get_nEff()), temp, static_cast<numT>(1e-15));
				}
				
				if (!existsAlready)
					mode_vector.push_back(mode<modeModel, numT>(temp, _lambda_double, _scaling_double, _layerDescription_vector));
				else
				{
					if ((_fileMap != 0) && (_fileMap->count(L"log") > 0))
					{
						pointerWrapper<std::ofstream> log = _fileMap->find(L"log")->second;
						log.lock();
						*(log.getObjectPtr()) << "modeSolver::determine_modes_in_interval: Rejected mode " << doubleValue(temp) << " for existing already.\n";
						log.unlock();
					}					
				}
			}
		}
	}
	
	return mode_vector;
}

/*
template <class modeModel, class numT>
std::vector<mode<modeModel> > modeSolver<modeModel, numT>::determine_modes_in_interval_direct(interval<std::complex<double> > n_interval)
{
	std::list<interval<std::complex<double > > > interval_list = split_interval_by_layer_real_n(n_interval);
	std::vector<mode<modeModel> > mode_vector;

	for (std::list<interval<std::complex<double > > >::iterator it = interval_list.begin(); it != interval_list.end(); it++)
	{
		determine_modes_in_interval_direct_sub(*it, mode_vector);
	}
	
	return mode_vector;	
}

template <class modeModel, class numT>
void modeSolver<modeModel, numT>::determine_modes_in_interval_direct_sub(interval<std::complex<double> > n_interval, std::vector<mode<modeModel> >& modes)
{
	std::complex<double> temp = determine_root_in_interval(n_interval, &modeModel::characteristic_equation, this);
	
	if (!std::isnan(std::real(temp)) && !std::isnan(std::imag(temp)))
	{
		if (std::abs(std::real(temp)/std::imag(temp)) > 1e15)
			temp = std::real(temp);

		if (std::imag(temp) >= 0.0)
		{
			// Test if we already got this mode.
			bool existsAlready = false;
			for (int modeCounter = 0; (modeCounter < modes.size()) && !existsAlready; modeCounter++)
			{
				existsAlready = thresholdEqual(modes[modeCounter].get_nEff(), temp, 1e-15);
			}
			
			if (!existsAlready)
				modes.push_back(mode<modeModel>(temp, _lambda_double, _scaling_double, _layerDescription_vector));
			else
			{
				if ((_fileMap != 0) && (_fileMap->count(L"log") > 0))
				{
					pointerWrapper<std::ofstream> log = _fileMap->find(L"log")->second;
					log.lock();
					*(log.getObjectPtr()) << "modeSolver::determine_modes_in_interval: Rejected mode " << temp << " for existing already.\n";
					log.unlock();
				}					
			}
		}
		
		// Split interval and keep looking for roots.
		interval<std::complex<double> > tempInt1, tempInt2;
		if (std::real(n_interval.start_T) < std::real(temp))
		{
			tempInt1.start_T = n_interval.start_T;
			tempInt1.end_T = std::complex<double>(std::real(temp) - _rootThreshold_double, std::imag(n_interval.end_T));
			tempInt2.start_T = std::complex<double>(std::real(temp) + _rootThreshold_double, std::imag(n_interval.start_T));
			tempInt2.end_T = n_interval.end_T;
		}
		else
		{
			tempInt1.start_T = n_interval.start_T;
			tempInt1.end_T = std::complex<double>(std::real(temp) + _rootThreshold_double, std::imag(n_interval.end_T));
			tempInt2.start_T = std::complex<double>(std::real(temp) - _rootThreshold_double, std::imag(n_interval.start_T));
			tempInt2.end_T = n_interval.end_T;
		}
		determine_modes_in_interval_direct_sub(tempInt1, modes);
		determine_modes_in_interval_direct_sub(tempInt2, modes);
	}
}
*/

template <class modeModel, class numT>
std::list<interval<std::complex<numT> > > modeSolver<modeModel, numT>::determine_roots_in_interval(interval<std::complex<numT> > n_interval, std::vector<layerDescription<numT> >& layerDescriptions_vector)
{
	std::list<interval<std::complex<numT> > > interval_list;
	
	// Make sure we are going from higher n to lower
	if (std::real(n_interval.start_T) < std::real(n_interval.end_T))
	{
		std::complex<numT> temp = n_interval.start_T;
		n_interval.start_T = n_interval.end_T;
		n_interval.end_T = temp;
	}

	// Make sure we don't include the real axis, since there are so many roots on this axis.
	if (std::min(std::imag(n_interval.start_T), std::imag(n_interval.end_T)) == static_cast<numT>(0.0))
	{
		if (std::imag(n_interval.start_T) > std::imag(n_interval.end_T))
			n_interval.end_T = std::complex<numT>(std::real(n_interval.end_T), -_imagOffset_double);
		else
			n_interval.start_T = std::complex<numT>(std::real(n_interval.start_T), -_imagOffset_double);
	}
		
	// Check how many root there are in this interval
	int rootTest_int;
	// If the interval is entierly in the negative imaginary part of the complex plane ignore it. This excludes sources.
	if ((std::imag(n_interval.start_T) < static_cast<numT>(0.0)) && (std::imag(n_interval.end_T) < static_cast<numT>(0.0)))
		rootTest_int = 0;
	else
	{
		numT rootTest = root_test(&characteristic_equation_path_integration_function, this, n_interval, static_cast<numT>(1e-2));

		if (((rootTest - floor(rootTest)) > static_cast<numT>(0.35)) && ((rootTest - floor(rootTest)) < static_cast<numT>(0.65)))
		{
			std::list<interval<std::complex<numT> > > tempIntervalList = split_interval_by_layer_real_n(growInterval(n_interval, static_cast<numT>(1e-6)));
		
			for (typename std::list<interval<std::complex<numT> > >::iterator tempListIt = tempIntervalList.begin(); tempListIt != tempIntervalList.end(); tempListIt++)
			{
				std::list<interval<std::complex<numT> > > temp = determine_roots_in_interval(*tempListIt, layerDescriptions_vector);
				interval_list.splice(interval_list.begin(), temp);
			}
			rootTest_int = 0;
		}
		else
			rootTest_int = intValue(round(rootTest));
	}

	if (rootTest_int == 1)
		interval_list.push_back(n_interval);
	else if (rootTest_int > 1)
	{
		// Split interval in as many subintervals as roots were found
		std::list<interval<std::complex<numT> > > tempInterval_list = split_interval(rootTest_int, n_interval);
		
		// Merge all roots found in the subinterval
		unsigned int rootCounter_uint = 0;
		for (typename std::list<interval<std::complex<numT> > >::iterator list_iterator = tempInterval_list.begin(); list_iterator != tempInterval_list.end(); list_iterator++)
		{
			std::list<interval<std::complex<numT> > > temp_list(determine_roots_in_interval(*list_iterator, layerDescriptions_vector));
			
			rootCounter_uint += temp_list.size();
			interval_list.splice(interval_list.begin(), temp_list);
		}

		if (rootCounter_uint != rootTest_int)
			if ((_fileMap != 0) && (_fileMap->count(L"log") > 0))
			{
				pointerWrapper<std::ofstream> log = _fileMap->find(L"log")->second;
				log.lock();
				*(log.getObjectPtr()) << "modeSolver::determine_roots_in_interval: Subintervals gave different number of roots (" << rootCounter_uint << ") than main interval (" << rootTest_int << ", " << doubleValue(n_interval.start_T) << "-" << doubleValue(n_interval.end_T) << ")\n";
				log.unlock();
			}
	}
	else if (rootTest_int < 0)
		if ((_fileMap != 0) && (_fileMap->count(L"log") > 0))
		{
			pointerWrapper<std::ofstream> log = _fileMap->find(L"log")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << "modeSolver::determine_roots_in_interval: rootTest < 0 (" << rootTest_int << ") interval: (" << doubleValue(n_interval.start_T) << "-" << doubleValue(n_interval.end_T) << ")\n";
			log.unlock();
		}

	return interval_list;
}

template <class modeModel, class numT>
std::list<interval<std::complex<numT> > > modeSolver<modeModel, numT>::split_interval_by_layer_real_n(const interval<std::complex<numT> >& n_interval) throw (std::runtime_error)
{
	std::list<interval<std::complex<numT> > > interval_list;
	std::list<numT> n_list;
	numT nRealLower_double = std::min(std::real(n_interval.start_T), std::real(n_interval.end_T));
	numT nRealHigher_double = std::max(std::real(n_interval.start_T), std::real(n_interval.end_T));
	numT nImagLower_double = std::min(std::imag(n_interval.start_T), std::imag(n_interval.end_T));
	numT nImagHigher_double = std::max(std::imag(n_interval.start_T), std::imag(n_interval.end_T));
	
	if (nImagLower_double > -_imagOffset_double * static_cast<numT>(0.9))
		nImagLower_double = -_imagOffset_double * static_cast<numT>(0.9);

	if (nImagHigher_double < _imagOffset_double)
		nImagHigher_double = _imagOffset_double;
	
	// Make a list of the refractive indices of all layers and sort it.
	for (unsigned int layerCounter_uint = 0; layerCounter_uint < _layerDescription_vector.size(); layerCounter_uint++)
		n_list.push_back(std::real(_layerDescription_vector[layerCounter_uint].n));
	
	n_list.sort();
	
	// Take care of the radiative modes.
	if (nRealLower_double < *n_list.begin())
	{
		interval< std::complex<numT> > temp_interval;

		temp_interval.start_T = std::complex<numT>(nRealLower_double, std::min(nImagLower_double, static_cast<numT>(-1e-10)));
		if (nRealHigher_double < *n_list.begin())
			temp_interval.end_T = std::complex<numT>(nRealHigher_double, std::max(nImagHigher_double, static_cast<numT>(1e0)));
		else
			temp_interval.end_T = std::complex<numT>(*n_list.begin() - _rootThreshold_double, std::max(nImagHigher_double, static_cast<numT>(1e-2)));

		interval_list.push_back(temp_interval);
	}

	// Cap the interval so that modes with effective refractive indices smaller than the
	// refractive index of the optically thinnest layer (radiation modes) and modes with
	// a effective refractive index larger than the refractive index of the optically thickest
	// layer (physically impossible?) are excluded.
	nRealLower_double = std::max((*n_list.begin() + _rootThreshold_double), nRealLower_double);
	nRealHigher_double = std::min((*--n_list.end() - _rootThreshold_double), nRealHigher_double);
		
	typename std::list<numT>::iterator list_iterator(n_list.begin());
	typename std::list<numT>::iterator listm1_iterator(list_iterator++);

	for (; (list_iterator != n_list.end()) && (nRealHigher_double != nRealLower_double); listm1_iterator = list_iterator++)
	{
		numT upper_boundary, lower_boundary;
		
		lower_boundary = std::max(nRealLower_double, (*listm1_iterator + _rootThreshold_double));
		upper_boundary = std::min(nRealHigher_double, (*list_iterator - _rootThreshold_double));
		
		if (lower_boundary < upper_boundary) {
			interval<std::complex<numT> > temp_interval;
			temp_interval.start_T = std::complex<numT>(lower_boundary, nImagLower_double);
			temp_interval.end_T = std::complex<numT>(upper_boundary, nImagHigher_double);
			interval_list.push_back(temp_interval);
		}
		
		nRealLower_double = std::max(nRealLower_double, upper_boundary);
	}

	return interval_list;
}

template <class modeModel, class numT>
std::complex<numT> modeSolver<modeModel, numT>::characteristic_equation_path_integration_function(const std::complex<numT>& arg, void* param)
{
	std::complex<numT> f_complexDouble, fPrime_complexDouble;
	modeModel::characteristic_equation(arg, param, &f_complexDouble, &fPrime_complexDouble);

	return fPrime_complexDouble / f_complexDouble;
}

template <class modeModel, class numT>
numT modeSolver<modeModel, numT>::get_lambda_double()
{
	return _lambda_double;
}

template <class modeModel, class numT>
std::vector<layerDescription<numT> > modeSolver<modeModel, numT>::get_layers_vector()
{
	return _layerDescription_vector;
}

template <class modeModel, class numT>
numT modeSolver<modeModel, numT>::get_scaling_double()
{
	return _scaling_double;
}

#endif

