/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _SECTION_H
#define _SECTION_H 1

#include <string>
#include <vector>
#include <complex>
#include <stdexcept>

#include "eo.h"
#include "stringFunctions.h"

//! Class to represent a section of the fibre.
/*!
 * This class holds all the information necessary the represent a section of
 * the fibre. This section can be either a grating, a bare stretch of fibre or
 * another optical element. At the current moment only the grating is
 * implemented. A list of the sections is passed to the sectionSolver together
 * with a list of the modes to then determine how the modes interact due to the
 * element in a particular section.
 */
class section
{
	private:
		std::map<std::wstring, eo<std::complex<double> > >  f_eos; /**< eos representing the equations to calculate the period, accumulated phase change and other characteristics of the section. */
		std::wstring f_type; /**< Type of the section */
		
	public:
		//! Constructor using the type of the section
		/*!
		 * This constructor takes a string containing the type of the grating.
		 */
		section(const std::wstring& type) throw (std::runtime_error);
		//! Copy constructor
		section(const section& obj);
		//! Destructor
		~section();
		//! Function to obtain all variables from the eos
		/*!
		 * Interface to the eo function.
		 */
		std::map<std::wstring, std::complex<double> > get_vars();
		//! Function setting the variables of the eos
		/*!
		 * Interface to the eo function.
		 */
		void set_vars(const std::map<std::wstring, std::complex<double> >& vars);
		//! Function to rename the variables of the eos
		/*!
		 * Interface to the eo function.
		 */
		void rename_vars(const std::map<std::wstring, std::wstring >& vars);
		//! Function to simplify the eos
		/*!
		 * Interface to the eo function.
		 */
		void simplify();
		//! Function to obtain the period of a grating section at a particular position
		double get_period(double position = 0.0) const throw (std::runtime_error);
		//! Function to obtain the accumulated phase of a grating section at a particular position
		double get_accumulated_phase(double position = 0.0) const throw (std::runtime_error);
		//! Function to obtain the apodisation value of a grating section at a particular position
		double get_sigma(const double& position = 0) const throw (std::runtime_error);
		//! Function to obtain the maximum period of grating section
		double get_max_period(void) const throw (std::runtime_error);
		//! Function to obtain the minimum period of a grating section
		double get_min_period(void) const throw (std::runtime_error);
		//! Function to obtain the length of the section
		double get_length(void) const throw (std::runtime_error);
		//! Function to obtain the constant amplitude factor of the modulation function of a grating section
		double get_dc_change(void) const throw (std::runtime_error);
		//! Function to obtain the variable amplitude factor of the modulation function of a grating section
		double get_ac_change(void) const throw (std::runtime_error);
		//! Function to obtain the maximum index variation of a grating section
		double get_p_change(void) const throw (std::runtime_error);
		//! Function to obtain the type of the section
		std::wstring get_type(void) const;
		
	private:
		section();
};

//! Class representing a grating section
/*!
 * This class just provides an initialisation function for the eos of a section.
 */
class section_grating
{
	public:
		//! Function to initialise the eos of a section
		static std::map<std::wstring, eo<std::complex<double> > > init();
};
#endif

