/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "mograsoFileHandler.h"

int mograsoFileHandler::get_numberOfJobs()
{
 	std::wstring query(L"//mograso:job");
	XPathObject obj = XPathQuery(L"//mograso:job", L"mograso", L"http://www.hochreiner.net/mograso.rng");
	
	return obj.get_numberOfNodes();
}

std::wstring mograsoFileHandler::get_jobType(const int& job)
{
	std::wstringstream queryStream;
	queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]";
	
	return XPathQuery(queryStream.str(), L"mograso", L"http://www.hochreiner.net/mograso.rng").get_node(0).get_property(L"type");
}

std::vector<variableStandIn> mograsoFileHandler::get_variableStandInVector(const int& job)
{
	std::vector<variableStandIn> output;

	std::wstringstream queryStream;
	queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]/mograso:variableDefinition";
	std::wstring query(queryStream.str());
	std::map<std::wstring, std::wstring> namesp;
	namesp[L"mograso"] = L"http://www.hochreiner.net/mograso.rng";

	XPathObject resultObj = XPathQuery(query, namesp);

	for (int counter = 0; counter < resultObj.get_numberOfNodes(); counter++)
	{
		std::wstring name = resultObj.get_node(counter).get_property(L"name");

		std::wstringstream prefixQuery;
		prefixQuery << query << L"[" << (counter + 1) << L"]/mograso:prefix";
		std::wstring prefix = XPathQuery(prefixQuery.str(), namesp).get_node(0).get_string();

		std::wstringstream suffixQuery;
		suffixQuery << query << L"[" << (counter + 1) << L"]/mograso:suffix";
		std::wstring suffix = XPathQuery(suffixQuery.str(), namesp).get_node(0).get_string();

		std::wstringstream startQuery;
		startQuery << query << L"[" << (counter + 1) << L"]/mograso:start";
		std::complex<double> start = wstringToComplexDouble(XPathQuery(startQuery.str(), namesp).get_node(0).get_string());

		std::wstringstream endQuery;
		endQuery << query << L"[" << (counter + 1) << L"]/mograso:end";
		std::complex<double> end = wstringToComplexDouble(XPathQuery(endQuery.str(), namesp).get_node(0).get_string());

		std::wstringstream stepQuery;
		stepQuery << query << L"[" << (counter + 1) << L"]/mograso:step";
		std::complex<double> step = wstringToComplexDouble(XPathQuery(stepQuery.str(), namesp).get_node(0).get_string());
		//numberOfSteps = (unsigned long)std::ceil(std::abs((end - start) / step));

		std::wstringstream numberQuery;
		numberQuery << query << L"[" << (counter + 1) << L"]/mograso:numberOfSteps";
		std::complex<double> numberOfSteps = wstringToComplexDouble(XPathQuery(numberQuery.str(), namesp).get_node(0).get_string());

		std::wstringstream valueQuery;
		valueQuery << query << L"[" << (counter + 1) << L"]/mograso:value";
		std::complex<double> value = wstringToComplexDouble(XPathQuery(valueQuery.str(), namesp).get_node(0).get_string());
		//out.add_variable(varName, start, end, numberOfSteps, varPrefix, varSuffix);

		output.push_back(variableStandIn(name, prefix, suffix, value, start, end, step, numberOfSteps));
	}

	return output;
}

variableStorage mograsoFileHandler::get_variableStorage(const int& job)
{
	return variableStandInVectorToVariableStorage(get_variableStandInVector(job));
}

variableStorage mograsoFileHandler::variableStandInVectorToVariableStorage(const std::vector<variableStandIn>& varSIs)
{
	variableStorage output;

	for (std::vector<variableStandIn>::const_iterator it = varSIs.begin(); it != varSIs.end(); it++)
	{
		if (it->get_type() == L"value")
			output.add_variable(it->get_name(), it->get_value(), NAN, static_cast<unsigned long>(NAN), it->get_prefix(), it->get_suffix());
		else
		{
			if (it->get_stepType() == L"step")
				output.add_variable(it->get_name(), it->get_start(), it->get_end(), static_cast<unsigned long>(std::ceil(std::abs((it->get_end() - it->get_start()) / it->get_step()))), it->get_prefix(), it->get_suffix());
			else
				output.add_variable(it->get_name(), it->get_start(), it->get_end(), static_cast<unsigned long>(std::abs(it->get_step())), it->get_prefix(), it->get_suffix());
		}
	}

	return output;
}

std::vector<layerStandIn> mograsoFileHandler::get_layerStandInVector(const int& job)
{
	std::vector<layerStandIn> out;

	std::map<std::wstring, std::wstring> namesp;
	namesp[L"mograso"] = L"http://www.hochreiner.net/mograso.rng";
	
	std::wstringstream queryStream;
	queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]/mograso:layer";
	std::wstring query(queryStream.str());
	
	XPathObject resultObj = XPathQuery(query, namesp);

	for (int counter = 0; counter < resultObj.get_numberOfNodes(); counter++)
	{
		std::wstring type = resultObj.get_node(counter).get_property(L"type");
		std::wstring layerType;
		
		if (type != L"")
			layerType = type;
		
		layerStandIn temp(layerType);
		std::vector<std::wstring> vars(temp.get_varsOrig());
		
		for (std::vector<std::wstring>::iterator it = vars.begin(); it != vars.end(); it++)
		{
			std::wstringstream nodeQuery;
			nodeQuery << query << L"[" << (counter + 1) << L"]";

			XMLNode tempNode = XPathQuery((nodeQuery.str() + std::wstring(L"/mograso:") + *it + std::wstring(L"/mograso:variable")), namesp).get_node(0);

			if (tempNode.get_property(L"name") != L"")
				temp.rename_var(*it, tempNode.get_property(L"name"));
			else
				temp.set_var(*it, wstringToComplexDouble(XPathQuery((nodeQuery.str() + std::wstring(L"/mograso:") + *it), namesp).get_node(0).get_string()));
		}
		
		out.push_back(temp);
	}
	
	return out;
}

std::vector<layer> mograsoFileHandler::get_layerVector(const int& job)
{
	return layerStandInVectorToLayerVector(get_layerStandInVector(job));
}

std::vector<layer> mograsoFileHandler::layerStandInVectorToLayerVector(const std::vector<layerStandIn>& layerSIs)
{
	std::vector<layer> output;

	for (std::vector<layerStandIn>::const_iterator it = layerSIs.begin(); it != layerSIs.end(); it++)
		output.push_back(it->get_layer());

	return output;
}

std::vector<section> mograsoFileHandler::get_sectionVector(const int& job)
{
	std::vector<section> out;

	std::map<std::wstring, std::wstring> namesp;
	namesp[L"mograso"] = L"http://www.hochreiner.net/mograso.rng";
	
	std::wstringstream queryStream;
	queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]/mograso:section";
	std::wstring query(queryStream.str());
	
	XPathObject obj = XPathQuery(query, namesp);

	for (int counter = 0; counter < obj.get_numberOfNodes(); counter++)
	{
		std::wstringstream nodeQuery;
		nodeQuery << query << L"[" << (counter + 1) << L"]";
		std::wstring type = XPathQuery(nodeQuery.str(), namesp).get_node(0).get_property(L"type");
		std::wstring sectionType;
		
		if (type != L"")
			sectionType = type;
		
		section temp(sectionType);
		std::map<std::wstring, std::complex<double> > vars(temp.get_vars());
		std::map<std::wstring, std::wstring> newVars;
		
		for (std::map<std::wstring, std::complex<double> >::iterator it = vars.begin(); it != vars.end(); it++)
		{
			XMLNode tempNode = XPathQuery(nodeQuery.str() + std::wstring(L"/mograso:") + it->first + std::wstring(L"/mograso:variable"), namesp).get_node(0);

			if (tempNode.get_property(L"name") != L"")
				newVars[it->first] = tempNode.get_property(L"name");
			else
				vars[it->first] = wstringToComplexDouble(XPathQuery((nodeQuery.str() + std::wstring(L"/mograso:") + it->first), namesp).get_node(0).get_string());
		}
		
		temp.set_vars(vars);
		temp.rename_vars(newVars);
		out.push_back(temp);
	}
	
	return out;
}

std::map<std::wstring, std::wstring> mograsoFileHandler::get_options(const int& job)
{
 	std::map<std::wstring, std::wstring> retMap;

	std::map<std::wstring, std::wstring> namesp;
	namesp[L"mograso"] = L"http://www.hochreiner.net/mograso.rng";
	
	std::wstringstream queryStream;
	queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]/mograso:option";
	
	XPathObject obj = XPathQuery(queryStream.str(), namesp);

	for (int counter = 0; counter < obj.get_numberOfNodes(); counter++)
		retMap[obj.get_node(counter).get_property(L"name")] = obj.get_node(counter).get_string();

	return retMap;
}

void mograsoFileHandler::get_perJobFiles(const int& job, const std::vector<std::wstring>& filenames, outputStorage* output)
{
	std::map<std::wstring, std::wstring> namesp;
	namesp[L"mograso"] = L"http://www.hochreiner.net/mograso.rng";
	
	for (std::vector<std::wstring>::const_iterator it = filenames.begin(); it != filenames.end(); it++)
	{
		std::wstringstream queryStream;
		queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]/mograso:output[attribute::dataid=\"" << *it <<  "\"]";
	
		std::wstring filename = XPathQuery(queryStream.str(), namesp).get_node(0).get_string();
		if (filename != L"")
			output->add_perJobFile(*it, filename);
	}
}

void mograsoFileHandler::get_perVarSetFiles(const int& job, const std::vector<std::wstring>& filenames, outputStorage* output)
{
	std::map<std::wstring, std::wstring> namesp;
	namesp[L"mograso"] = L"http://www.hochreiner.net/mograso.rng";
	
	for (std::vector<std::wstring>::const_iterator it = filenames.begin(); it != filenames.end(); it++)
	{
		std::wstringstream queryStream;
		queryStream << L"/mograso:mograso/mograso:job[" << (job + 1) << L"]/mograso:output[attribute::dataid=\"" << *it <<  "\"]";
		
		std::wstring filename = XPathQuery(queryStream.str(), namesp).get_node(0).get_string();
		if (filename != L"")
			output->add_perVarSetFile(*it, filename);
	}
}

