/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#import "ConjBesselProductXrInt.h"

std::complex<double> conjBesselProductXrInt(const double& order, const bessel_type& type1, const std::complex<double>& argPart1, const bessel_type& type2, const std::complex<double>& argPart2, const double& rStart, const double& rEnd)
{
	// For the time being we only deal with 0th order bessel functions.
	if (order != 0.0)
		return std::complex<double>(NAN, NAN);
	
	// Take a few precautions
	if ((rStart < 0) || (rStart == INFINITY) || (rEnd <= 0) || (rEnd == INFINITY))
		return std::complex<double>(NAN, NAN);
	
	if (rStart != 0.0)
		return conjBesselProductXrInt(order, type1, argPart1, type2, argPart2, 0.0, rEnd) - conjBesselProductXrInt(order, type1, argPart1, type2, argPart2, 0.0, rStart);
	
	// Real arguments
	if ((std::imag(argPart1) == 0.0) && (std::imag(argPart2) == 0.0))
	{
		// Same real arguments
		if (argPart1 == argPart2)
		{
			// JJ
			if ((type1 == kBesselJ) && (type2 == kBesselJ))
			{
				std::complex<double> J0 = besselj(order, argPart1 * rEnd);
				std::complex<double> J1 = besselj(order + 1.0, argPart1 * rEnd);
				return 0.5 * rEnd * rEnd * (J0 * J0 + J1 * J1);
			}
			// JY and YJ
			else if ((type1 | type2) == (kBesselJ | kBesselY))
			{
				std::complex<double> J0 = besselj(order, argPart1 * rEnd);
				std::complex<double> J1 = besselj(order + 1.0, argPart1 * rEnd);
				std::complex<double> Y0 = bessely(order, argPart1 * rEnd);
				std::complex<double> Y1 = bessely(order + 1.0, argPart1 * rEnd);
				return 0.5 * rEnd * rEnd * (J0 * Y0 + J1 * Y1);
			}
			// YY
			else if ((type1 == kBesselY) && (type2 == kBesselY))
			{
				std::complex<double> Y0 = bessely(order, argPart1 * rEnd);
				std::complex<double> Y1 = bessely(order + 1.0, argPart1 * rEnd);
//				std::cout << "Y0: " << Y0 << "  Y1: " << Y1 << "  arg: " << argPart1 * rEnd << "\n";
				return 0.5 * rEnd * rEnd * (Y0 * Y0 + Y1 * Y1) - 2.0 / (M_PI * M_PI * argPart1 * argPart1);
			}
			// II
			else if ((type1 == kBesselI) && (type2 == kBesselI))
			{
				std::complex<double> I0 = besseli(order, argPart1 * rEnd);
				std::complex<double> I1 = besseli(order + 1.0, argPart1 * rEnd);
				return 0.5 * rEnd * rEnd * (I0 * I0 - I1 * I1);
			}
			// KK
			else if ((type1 == kBesselK) && (type2 == kBesselK))
			{
				std::complex<double> K0 = besselk(order, argPart1 * rEnd);
				std::complex<double> K1 = besselk(order + 1.0, argPart1 * rEnd);
				return 0.5 * rEnd * rEnd * (K0 * K0 - K1 * K1) + 1.0 / (2.0 * argPart1 * argPart1);
			}
		}
		else // Different real arguments
		{
			// JJ
			if ((type1 == kBesselJ) && (type2 == kBesselJ))
			{
				std::complex<double> J0_1 = besselj(order, argPart1 * rEnd);
				std::complex<double> J1_1 = besselj(order + 1.0, argPart1 * rEnd);
				std::complex<double> J0_2 = besselj(order, argPart2 * rEnd);
				std::complex<double> J1_2 = besselj(order + 1.0, argPart2 * rEnd);
				return rEnd * (J1_1 * argPart1 * J0_2 - J0_1 * J1_2 * argPart2) / (argPart1 * argPart1 - argPart2 * argPart2);
			}			
		}
	} else { // complex arguments
		if (argPart1 == argPart2) // same complex arguments
		{
			// JJ
			if ((type1 == kBesselJ) && (type2 == kBesselJ))
			{
				std::complex<double> J0 = besselj(order, argPart1 * rEnd);
				std::complex<double> J1 = besselj(order + 1.0, argPart1 * rEnd);
				std::complex<double> argPart1_conj = std::conj(argPart1);
				std::complex<double> J0_conj = besselj(order, argPart1_conj * rEnd);
				std::complex<double> J1_conj = besselj(order + 1.0, argPart1_conj * rEnd);
				return rEnd * (J1 * argPart1 * J0_conj - J0 * J1_conj * argPart1_conj) / (argPart1 * argPart1 - argPart1_conj * argPart1_conj);
			}
			else if ((type1 == kBesselJ) && (type2 == kBesselY))
			{
				std::complex<double> J0 = besselj(order, argPart1 * rEnd);
				std::complex<double> J1 = besselj(order + 1.0, argPart1 * rEnd);
				std::complex<double> argPart1_conj = std::conj(argPart1);
				std::complex<double> Y0_conj = bessely(order, argPart1_conj * rEnd);
				std::complex<double> Y1_conj = bessely(order + 1.0, argPart1_conj * rEnd);
				return (2.0 + rEnd * M_PI * J0 * Y1_conj * argPart1_conj - rEnd * M_PI * J1 * argPart1 * Y0_conj) / (M_PI * (argPart1_conj * argPart1_conj - argPart1 * argPart1));
			}
			else if ((type1 == kBesselY) && (type2 == kBesselJ))
			{
				std::complex<double> Y0 = bessely(order, argPart1 * rEnd);
				std::complex<double> Y1 = bessely(order + 1.0, argPart1 * rEnd);
				std::complex<double> argPart1_conj = std::conj(argPart1);
				std::complex<double> J0_conj = besselj(order, argPart1_conj * rEnd);
				std::complex<double> J1_conj = besselj(order + 1.0, argPart1_conj * rEnd);
				return (-2.0 - rEnd * M_PI * Y1 * argPart1 * J0_conj + rEnd * M_PI * Y0 * J1_conj * argPart1_conj) / (M_PI * (argPart1_conj * argPart1_conj - argPart1 * argPart1));
			}
// 			else if ((type1 == kBesselY) && (type2 == kBesselY))
// 			{
// 				std::complex<double> Y0 = bessely(order, argPart1 * rEnd);
// 				std::complex<double> Y1 = bessely(order + 1.0, argPart1 * rEnd);
// 				return - 2.0 / (argPart1 * argPart1 * M_PI * M_PI) + 0.5 * rEnd * rEnd * (Y0 * Y0 + Y1 * Y1);
// 			}
		}
		else // different complex arguments
		{
			// JJ
			if ((type1 == kBesselJ) && (type2 == kBesselJ))
			{
				std::complex<double> argPart2_conj = std::conj(argPart2);
				std::complex<double> J0_1 = besselj(order, argPart1 * rEnd);
				std::complex<double> J1_1 = besselj(order + 1.0, argPart1 * rEnd);
				std::complex<double> J0_2_conj = besselj(order, argPart2_conj * rEnd);
				std::complex<double> J1_2_conj = besselj(order + 1.0, argPart2_conj * rEnd);
				return rEnd * (J0_1 * J1_2_conj * argPart2_conj - J1_1 * argPart1 * J0_2_conj) / (argPart2_conj * argPart2_conj - argPart1 * argPart1);
			}
		}
	}

	// If we have not dealt with it already, we can't deal with it.
	return std::complex<double>(NAN, NAN);
}
