/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This header is a wrapper for the FORTRAN functions to calculate the
 * Bessel functions of real order and complex argument written by Donald E.
 * Amos. The particular implementation is taken from the Octave project.
 *
 */

#ifndef _HH_BESSEL_AMOS
#define _HH_BESSEL_AMOS 1

#include <complex>
#include <cmath>
#include "numberFunctions.h"

extern "C" {
	extern void zbesi_(double *ZR, double *ZI, double *FNU, int *KODE, int *N, double *CYR, double *CYI, int *NZ, int *IERR);
	extern void zbesj_(double *ZR, double *ZI, double *FNU, int *KODE, int *N, double *CYR, double *CYI, int *NZ, int *IERR);
	extern void zbesk_(double *ZR, double *ZI, double *FNU, int *KODE, int *N, double *CYR, double *CYI, int *NZ, int *IERR);
	extern void zbesy_(double *ZR, double *ZI, double *FNU, int *KODE, int *N, double *CYR, double *CYI, int *NZ, double *CWRKR, double *CWRKI, int *IERR);
	extern void zbesh_(double *ZR, double *ZI, double *FNU, int *KODE, int *M, int *N, double *CYR, double *CYI, int *NZ, int *IERR);
};

//! Enumeration for the type of the Bessel function
enum bessel_type {
	kBesselJ = 1, /**< Bessel function of the first kind */
	kBesselY = 2, /**< Bessel function of the second kind */
	kBesselN = 2, /**< Bessel function of the second kind */
	kBesselI = 4, /**< modified Bessel function of the first kind */
	kBesselK = 8  /**< modified Bessel function of the second kind */
};

//! Forward for the Bessel function of the second kind
/*!
 * The Bessel function of the second kind is sometime referred to as the
 * Neumann function. Hence, BesselN.
 */
template<class T>
inline std::complex<T> besseln(double order, std::complex<T> argument) {
	return bessely(order, argument);
}

//! Forward to be able to address the Bessel function using the above enumeration
template<class T>
inline std::complex<T> bessel(bessel_type type, double order, std::complex<T> argument) {
	if (type == kBesselJ) {
		return besselj(order, argument);
	} else if (type == kBesselY) {
		return bessely(order, argument);
	} else if (type == kBesselI) {
		return besseli(order, argument);
	} else if (type == kBesselK) {
		return besselk(order, argument);
	}
	return std::complex<T>(NAN, NAN);
}

//! Wrapper for the Bessel I function
template<class T>
std::complex<T> besseli(double order, std::complex<T> argument)
{
	int kode = 1, n = 1, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	I(-FNU,Z) = I(FNU,Z) + (2/PI)*SIN(PI*FNU)*K(FNU,Z)
	if (order < 0)
	{
		order = std::abs(order);
		if (round(order) == order)
			return besseli(order, argument);
		else
			return besseli(order, argument) + besselk(order, argument) * static_cast<T>(std::sin(M_PI * order) * 2.0 / M_PI);
	}
	else
	{
		zbesi_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
//		IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//		IERR=1, INPUT ERROR   - NO COMPUTATION
//		IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//		IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//		IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//		IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			return std::complex<T>(cr[0], ci[0]);
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			SCALING FACTOR: EXP(-ABS(X)), WHERE X=REAL(Z)
			zbesi_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
			{
				return std::complex<T>(cr[0], ci[0]) / std::exp(-std::abs(std::real(argument)));
			}
		}
	}
	return std::complex<T>(NAN, NAN);
}

//! Wrapper for the Bessel I function returning the function value for order and order plus 1
template<class T>
void besseli(double order, std::complex<T> argument, std::complex<T>& retn, std::complex<T>& retnp1)
{
	int kode = 1, n = 2, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	I(-FNU,Z) = I(FNU,Z) + (2/PI)*SIN(PI*FNU)*K(FNU,Z)
	if (order >= 0.0)
	{
		zbesi_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
//		IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//		IERR=1, INPUT ERROR   - NO COMPUTATION
//		IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//		IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//		IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//		IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			retn = std::complex<double>(cr[0], ci[0]);
			retnp1 = std::complex<double>(cr[1], ci[1]);
			return;
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			SCALING FACTOR: EXP(-ABS(X)), WHERE X=REAL(Z)
			zbesi_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
			{
				retn = std::complex<T>(cr[0], ci[0]) / std::exp(-std::abs(std::real(argument)));
				retnp1 = std::complex<T>(cr[1], ci[1]) / std::exp(-std::abs(std::real(argument)));
				return;
			}
		}
	}
	retn = std::complex<T>(NAN, NAN);
	retnp1 = std::complex<T>(NAN, NAN);
	return;
}

//! Wrapper for the Bessel J function
template<class T>
std::complex<T> besselj(double order, std::complex<T> argument)
{
	int kode = 1, n = 1, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	J(-FNU,Z) = J(FNU,Z)*COS(PI*FNU) - Y(FNU,Z)*SIN(PI*FNU)
	if (order < 0)
	{
		order = std::abs(order);
		if (round(order) == order)
		{
			if (((int)order % 2) == 0)
				return besselj(order, argument);
			else
				return -besselj(order, argument);
		}
		else
			return besselj(order, argument) * std::cos(M_PI * (T)order) - bessely(order, argument) * std::sin(M_PI * (T)order);
	}
	else
	{
		zbesj_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
//		IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//		IERR=1, INPUT ERROR   - NO COMPUTATION
//		IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//		IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//		IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//		IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
			return std::complex<T>(cr[0], ci[0]);
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			SCALING FACTOR: EXP(-ABS(Y)), WHERE Y=AIMAG(Z)
			zbesj_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
				return std::complex<T>(cr[0], ci[0]) / std::exp(-std::abs(std::imag(argument)));
		}
	}
	return std::complex<T>(NAN, NAN);
}

//! Wrapper for the Bessel J function returning the function value for order and order plus 1
template<class T>
void besselj(double order, std::complex<T> argument, std::complex<T>& retn, std::complex<T>& retnp1)
{
	int kode = 1, n = 2, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	J(-FNU,Z) = J(FNU,Z)*COS(PI*FNU) - Y(FNU,Z)*SIN(PI*FNU)
	if (order >= 0)
	{
		zbesj_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
//		IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//		IERR=1, INPUT ERROR   - NO COMPUTATION
//		IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//		IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//		IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//		IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			retn = std::complex<T>(cr[0], ci[0]);
			retnp1 = std::complex<T>(cr[1], ci[1]);
			return;
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			SCALING FACTOR: EXP(-ABS(Y)), WHERE Y=AIMAG(Z)
			zbesj_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
			{
				retn = std::complex<T>(cr[0], ci[0]) / std::exp(-std::abs(std::imag(argument)));
				retnp1 = std::complex<T>(cr[1], ci[1]) / std::exp(-std::abs(std::imag(argument)));
				return;
			}
		}
	}
	retn = std::complex<T>(NAN, NAN);
	retnp1 = std::complex<T>(NAN, NAN);
	return;
}

//! Wrapper for the Bessel K function
template <class T>
std::complex<T> besselk(double order, std::complex<T> argument)
{
	int kode = 1, n = 1, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	K(-FNU,Z) = K(FNU,Z)
	if (order < 0)
		order = std::abs(order);

	zbesk_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
//	IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//	IERR=1, INPUT ERROR   - NO COMPUTATION
//	IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//	IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//	IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//	IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
	if (ierr == 0)
		return std::complex<T>(cr[0], ci[0]);
	else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
	{
		kode = 2;
//		SCALING FACTOR: EXP(Z)
		zbesk_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
		if (ierr == 0)
			return std::complex<T>(cr[0], ci[0]) / std::exp(argument);
	}
	return std::complex<T>(NAN, NAN);
}

//! Wrapper for the Bessel K function returning the function value for order and order plus 1
template <class T>
void besselk(double order, std::complex<T> argument, std::complex<T>& retn, std::complex<T>& retnp1)
{
	int kode = 1, n = 2, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	K(-FNU,Z) = K(FNU,Z)
	if (order >= 0.0)
	{
	
		zbesk_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
	//	IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
	//	IERR=1, INPUT ERROR   - NO COMPUTATION
	//	IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
	//	IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
	//	IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
	//	IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			retn = std::complex<T>(cr[0], ci[0]);
			retnp1 = std::complex<T>(cr[1], ci[1]);
			return;
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
	//		SCALING FACTOR: EXP(Z)
			zbesk_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
			{
				retn = std::complex<T>(cr[0], ci[0]) / std::exp(argument);
				retnp1 = std::complex<T>(cr[1], ci[1]) / std::exp(argument);
				return;
			}
		}
	}
	retn = std::complex<T>(NAN, NAN);
	retnp1 = std::complex<T>(NAN, NAN);
	return;
}

//! Wrapper for the Bessel Y function
template <class T>
std::complex<T> bessely(double order, std::complex<T> argument)
{
	int kode = 1, n = 1, nz, ierr;
	double cr[n], ci[n], cwrkr[n], cwrki[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	Y(-FNU,Z) = Y(FNU,Z)*COS(PI*FNU) + J(FNU,Z)*SIN(PI*FNU)
	if (order < 0)
	{
		order = std::abs(order);
		if (round(order) == order)
		{
			if (((int)order % 2) == 0)
				return bessely(order, argument);
			else
				return -bessely(order, argument);
		}
		else
			return bessely(order, argument) * std::cos(M_PI * (T)order) + besselj(order, argument) * std::sin(M_PI * (T)order);			
	}
	else
	{
		zbesy_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, cwrkr, cwrki, &ierr);
//		IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//		IERR=1, INPUT ERROR   - NO COMPUTATION
//		IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//		IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//		IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//		IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
			return std::complex<T>(cr[0], ci[0]);
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			SCALING FACTOR: EXP(-ABS(Y)) WHERE Y=AIMAG(Z)
			zbesy_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, cwrkr, cwrki, &ierr);
			if (ierr == 0)
				return std::complex<T>(cr[0], ci[0]) / std::exp(-std::abs(std::imag(argument)));
		}
	}
	return std::complex<T>(NAN, NAN);
}

//! Wrapper for the Bessel Y function returning the function value for order and order plus 1
template<class T>
void bessely(double order, std::complex<T> argument, std::complex<T>& retn, std::complex<T>& retnp1)
{
	int kode = 1, n = 2, nz, ierr;
	double cr[n], ci[n], cwrkr[n], cwrki[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	Y(-FNU,Z) = Y(FNU,Z)*COS(PI*FNU) + J(FNU,Z)*SIN(PI*FNU)
	if (order >= 0)
	{
		zbesy_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, cwrkr, cwrki, &ierr);
//		IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//		IERR=1, INPUT ERROR   - NO COMPUTATION
//		IERR=2, OVERFLOW      - NO COMPUTATION, AIMAG(Z) TOO LARGE ON KODE=1
//		IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE BUT LOSSES OF SIGNIFCANCE BY ARGUMENT REDUCTION PRODUCE LESS THAN HALF OF MACHINE ACCURACY
//		IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTATION BECAUSE OF COMPLETE LOSSES OF SIGNIFICANCE BY ARGUMENT REDUCTION
//		IERR=5, ERROR - NO COMPUTATION, ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			retn = std::complex<T>(cr[0], ci[0]);
			retnp1 = std::complex<T>(cr[1], ci[1]);
			return;
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			SCALING FACTOR: EXP(-ABS(Y)) WHERE Y=AIMAG(Z)
			zbesy_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &n, cr, ci, &nz, cwrkr, cwrki, &ierr);
			if (ierr == 0)
			{
				retn = std::complex<T>(cr[0], ci[0]) / std::exp(-std::abs(std::imag(argument)));
				retnp1 = std::complex<T>(cr[1], ci[1]) / std::exp(-std::abs(std::imag(argument)));
				return;
			}
		}
	}
	retn = std::complex<T>(NAN, NAN);
	retnp1 = std::complex<T>(NAN, NAN);
	return;
}

//! Wrapper for the Hankel function of the first kind
template <class T>
std::complex<T> besselh1(double order, std::complex<T> argument)
{
	int kode = 1, n = 1, m = 1, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	H(1,-FNU,Z) = H(1,FNU,Z)*CEXP( PI*FNU*I)
	if (order < 0)
	{
		order = std::abs(order);
		return besselh1(order, argument) / std::exp(std::complex<T>(static_cast<T>(0.0), static_cast<T>(M_PI * order)));
	}
	else
	{
		zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
//C                    IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//C                    IERR=1, INPUT ERROR   - NO COMPUTATION
//C                    IERR=2, OVERFLOW      - NO COMPUTATION, FNU TOO
//C                            LARGE OR CABS(Z) TOO SMALL OR BOTH
//C                    IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE
//C                            BUT LOSSES OF SIGNIFCANCE BY ARGUMENT
//C                            REDUCTION PRODUCE LESS THAN HALF OF MACHINE
//C                            ACCURACY
//C                    IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTA-
//C                            TION BECAUSE OF COMPLETE LOSSES OF SIGNIFI-
//C                            CANCE BY ARGUMENT REDUCTION
//C                    IERR=5, ERROR              - NO COMPUTATION,
//C                            ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
			return std::complex<T>(cr[0], ci[0]);
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			Scaling factor: EXP(-MM*Z*I)     MM=3-2*M,   I**2=-1
			zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
				return std::complex<T>(cr[0], ci[0]) / std::exp(-(std::complex<T>(0.0, 3.0 - 2.0 * (T)m) * argument));
		}
	}
	return std::complex<T>(NAN, NAN);
}

//! Wrapper for the Hankel function of the first kind returning the function value for order and order plus 1
template <class T>
void besselh1(double order, std::complex<T> argument, std::complex<T>& retn, std::complex<T>& retnp1)
{
	int kode = 1, n = 2, m = 1, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	H(1,-FNU,Z) = H(1,FNU,Z)*CEXP( PI*FNU*I)
	if (order >= 0)
	{
		zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
//C                    IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//C                    IERR=1, INPUT ERROR   - NO COMPUTATION
//C                    IERR=2, OVERFLOW      - NO COMPUTATION, FNU TOO
//C                            LARGE OR CABS(Z) TOO SMALL OR BOTH
//C                    IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE
//C                            BUT LOSSES OF SIGNIFCANCE BY ARGUMENT
//C                            REDUCTION PRODUCE LESS THAN HALF OF MACHINE
//C                            ACCURACY
//C                    IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTA-
//C                            TION BECAUSE OF COMPLETE LOSSES OF SIGNIFI-
//C                            CANCE BY ARGUMENT REDUCTION
//C                    IERR=5, ERROR              - NO COMPUTATION,
//C                            ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			retn = std::complex<T>(cr[0], ci[0]);
			retnp1 = std::complex<T>(cr[1], ci[1]);
			return;
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			Scaling factor: EXP(-MM*Z*I)     MM=3-2*M,   I**2=-1
			zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
			{
				retn = std::complex<T>(cr[0], ci[0]) / std::exp(-(std::complex<T>(0.0, 3.0 - 2.0 * (T)m) * argument));
				retnp1 = std::complex<T>(cr[1], ci[1]) / std::exp(-(std::complex<T>(0.0, 3.0 - 2.0 * (T)m) * argument));
				return;
			}
		}
	}
	retn = std::complex<T>(NAN, NAN);
	retnp1 = std::complex<T>(NAN, NAN);
	return;
}

//! Wrapper for the Hankel function of the second kind
template <class T>
std::complex<T> besselh2(double order, std::complex<T> argument)
{
	int kode = 1, n = 1, m = 2, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	H(2,-FNU,Z) = H(2,FNU,Z)*CEXP(-PI*FNU*I)
	if (order < 0)
	{
		order = std::abs(order);
		return besselh2(order, argument) / std::exp(std::complex<T>(0.0, -M_PI * (T)order));
	}
	else
	{
		zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
//C                    IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//C                    IERR=1, INPUT ERROR   - NO COMPUTATION
//C                    IERR=2, OVERFLOW      - NO COMPUTATION, FNU TOO
//C                            LARGE OR CABS(Z) TOO SMALL OR BOTH
//C                    IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE
//C                            BUT LOSSES OF SIGNIFCANCE BY ARGUMENT
//C                            REDUCTION PRODUCE LESS THAN HALF OF MACHINE
//C                            ACCURACY
//C                    IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTA-
//C                            TION BECAUSE OF COMPLETE LOSSES OF SIGNIFI-
//C                            CANCE BY ARGUMENT REDUCTION
//C                    IERR=5, ERROR              - NO COMPUTATION,
//C                            ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
			return std::complex<T>(cr[0], ci[0]);
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			Scaling factor: EXP(-MM*Z*I)     MM=3-2*M,   I**2=-1
			zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
				return std::complex<T>(cr[0], ci[0]) / std::exp(-(std::complex<T>(0.0, 3.0 - 2.0 * (T)m) * argument));
		}
	}
	return std::complex<T>(NAN, NAN);
}
//! Wrapper for the Hankel function of the second kind returning the function value for order and order plus 1
template <class T>
void besselh2(double order, std::complex<T> argument, std::complex<T>& retn, std::complex<T>& retnp1)
{
	int kode = 1, n = 2, m = 2, nz, ierr;
	double cr[n], ci[n];
	std::complex<double> tempArg = doubleValue(argument);
	
//	H(2,-FNU,Z) = H(2,FNU,Z)*CEXP(-PI*FNU*I)
	if (order >= 0)
	{
		zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
//C                    IERR=0, NORMAL RETURN - COMPUTATION COMPLETED
//C                    IERR=1, INPUT ERROR   - NO COMPUTATION
//C                    IERR=2, OVERFLOW      - NO COMPUTATION, FNU TOO
//C                            LARGE OR CABS(Z) TOO SMALL OR BOTH
//C                    IERR=3, CABS(Z) OR FNU+N-1 LARGE - COMPUTATION DONE
//C                            BUT LOSSES OF SIGNIFCANCE BY ARGUMENT
//C                            REDUCTION PRODUCE LESS THAN HALF OF MACHINE
//C                            ACCURACY
//C                    IERR=4, CABS(Z) OR FNU+N-1 TOO LARGE - NO COMPUTA-
//C                            TION BECAUSE OF COMPLETE LOSSES OF SIGNIFI-
//C                            CANCE BY ARGUMENT REDUCTION
//C                    IERR=5, ERROR              - NO COMPUTATION,
//C                            ALGORITHM TERMINATION CONDITION NOT MET
		if (ierr == 0)
		{
			retn = std::complex<T>(cr[0], ci[0]);
			retnp1 = std::complex<T>(cr[1], ci[1]);
			return;
		}
		else if ((ierr == 2) || (ierr == 3) || (ierr == 4))
		{
			kode = 2;
//			Scaling factor: EXP(-MM*Z*I)     MM=3-2*M,   I**2=-1
			zbesh_(&(std::real(tempArg)), &(std::imag(tempArg)), &order, &kode, &m, &n, cr, ci, &nz, &ierr);
			if (ierr == 0)
			{
				retn = std::complex<T>(cr[0], ci[0]) / std::exp(-(std::complex<T>(0.0, 3.0 - 2.0 * (T)m) * argument));
				retnp1 = std::complex<T>(cr[1], ci[1]) / std::exp(-(std::complex<T>(0.0, 3.0 - 2.0 * (T)m) * argument));
				return;
			}
		}
	}
	retn = std::complex<T>(NAN, NAN);
	retnp1 = std::complex<T>(NAN, NAN);
	return;
}

#endif

