/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _EO_H
#define _EO_H 1

#include <cmath>
#include <string>
#include <sstream>
#include <stdexcept>
#include <map>
#include <complex>

#include "stringFunctions.h"

namespace std
{
	template <class T>
	bool isnan(const std::complex<T>& op)
	{
		if (isnan(real(op)) || isnan(imag(op)))
			return true;
		else
			return false;
	}
}

template <class T>
class eo
{
	private:
		enum op
		{
			kOP_NONE,
			kOP_PLUS,
			kOP_MINUS,
			kOP_MULT,
			kOP_DIV,
			kOP_SQRT
		};
		eo<T>* f_op1;
		eo<T>* f_op2;
		op f_op;
		std::wstring* f_name;
		T* f_value;

	public:
		eo(const T& op);
		eo(const std::wstring& name);
		eo(const eo<T>& op1, const eo<T>& op2, const op& op);
		eo(const eo<T>& op1, const op& op) throw (std::runtime_error);
		eo(const eo<T>& op);
		~eo();
		void set_vars(const std::map<std::wstring, T>& vars);
		std::map<std::wstring, T> get_vars() const;
		void get_vars(std::map<std::wstring, T>& vars) const;
		void rename_vars(const std::map<std::wstring, std::wstring>& vars);
		void operator=(const eo<T>& op);
		eo<T> operator+(const eo<T>& op) const;
		eo<T> operator+(const T& op) const { return (*this) + (eo<T>)op; };
		eo<T> operator-(const eo<T>& op) const;
		eo<T> operator-(const T& op) const { return (*this) - (eo<T>)op; };
		eo<T> operator*(const eo<T>& op) const;
		eo<T> operator*(const T& op) const { return (*this) * (eo<T>)op; };
		eo<T> operator/(const eo<T>& op) const;
		eo<T> operator/(const T& op) const { return (*this) / (eo<T>)op; };
		eo<T> sqrt() const;
		operator T() const throw (std::runtime_error);
		void expression(std::wstringstream& sstream);
		void simplify(const bool& considerAssignedVarsConst = true);
		bool isOperation() const;
		bool isUnaryOperation() const;
		bool isBinaryOperation() const;
		bool isVariable() const;
		bool isAssignedVariable() const;
		bool isConstant() const;
		
	private:
		eo<T>();
		void simplifyFull(bool& changed);
		void simplifyPartial(bool& changed);
};

template<class T, class K>
eo<T> operator+(const K& op1, const eo<T>& op2)
{
	return eo<T>(op1) + op2;
}

template<class T, class K>
eo<T> operator-(const K& op1, const eo<T>& op2)
{
	return eo<T>(op1) - op2;
}

template<class T, class K>
eo<T> operator*(const K& op1, const eo<T>& op2)
{
	return eo<T>(op1) * op2;
}

template<class T, class K>
eo<T> operator/(const K& op1, const eo<T>& op2)
{
	return eo<T>(op1) / op2;
}

template <>
void eo<std::complex<double> >::set_vars(const std::map<std::wstring, std::complex<double> >& vars);

template <>
void eo<std::complex<double> >::get_vars(std::map<std::wstring, std::complex<double> >& vars) const;

template <>
std::map<std::wstring, std::complex<double> > eo<std::complex<double> >::get_vars() const;

template <class T>
eo<T>::eo(const T& op)
{
	f_op1 = 0;
	f_op2 = 0;
	f_op = kOP_NONE;
	f_name = 0;
	f_value = new T(op);
}

template <class T>
eo<T>::eo(const std::wstring& name)
{
	f_op1 = 0;
	f_op2 = 0;
	f_op = kOP_NONE;
	f_name = new std::wstring(name);
	f_value = 0;
}

template <class T>
eo<T>::eo(const eo<T>& op1, const eo<T>& op2, const op& op)
{
	f_op1 = new eo(op1);
	f_op2 = new eo(op2);
	f_op = op;
	f_name = 0;
	f_value = 0;
}

template <class T>
eo<T>::eo(const eo<T>& op1, const op& op) throw (std::runtime_error)
{
	if (op != kOP_SQRT)
		throw std::runtime_error("Only one operand supplied for binary operation.");

	f_op1 = new eo(op1);
	f_op2 = 0;
	f_op = op;
	f_name = 0;
	f_value = 0;
}

template <class T>
eo<T>::eo(const eo<T>& op)
{
	if (op.f_op1 != 0)
		f_op1 = new eo(*(op.f_op1));
	else
		f_op1 = 0;
		
	if (op.f_op2 != 0)	
		f_op2 = new eo(*(op.f_op2));
	else
		f_op2 = 0;

	f_op = op.f_op;
	
	if (op.f_name != 0)
		f_name = new std::wstring(*(op.f_name));
	else
		f_name = 0;

	if (op.f_value != 0)
		f_value = new T(*(op.f_value));
	else
		f_value = 0;
}

template <class T>
eo<T>::~eo()
{
	delete f_op1;
	delete f_op2;
	delete f_name;
	delete f_value;
}

template <class T>
void eo<T>::set_vars(const std::map<std::wstring, T>& vars)
{
	if ((f_name != 0) && (vars.count(*f_name) != 0) && (!std::isnan(vars.find(*f_name)->second)))
	{
		delete f_value;
		f_value = new T(vars.find(*f_name)->second);
	}
	
	if (f_op1 != 0)
		f_op1->set_vars(vars);

	if (f_op2 != 0)
		f_op2->set_vars(vars);
}

template <class T>
std::map<std::wstring, T> eo<T>::get_vars() const
{
	std::map<std::wstring, T> vars;
	
	if (f_name != 0)
	{
		if (f_value != 0)
			vars.insert(std::pair<std::wstring, T>(*f_name, *f_value));
		else
			vars.insert(std::pair<std::wstring, T>(*f_name, (T)NAN));
	}
	
	if (f_op1 != 0)
		f_op1->get_vars(vars);
	
	if (f_op2 != 0)
		f_op2->get_vars(vars);	
	
	return vars;
}

template <class T>
void eo<T>::get_vars(std::map<std::wstring, T>& vars) const
{
	if (f_name != 0)
	{
		if (f_value != 0)
			vars.insert(std::pair<std::wstring, T>(*f_name, *f_value));
		else
			vars.insert(std::pair<std::wstring, T>(*f_name, (T)NAN));
	}
	
	if (f_op1 != 0)
		f_op1->get_vars(vars);
	
	if (f_op2 != 0)
		f_op2->get_vars(vars);
}

template<class T>
void eo<T>::rename_vars(const std::map<std::wstring, std::wstring>& vars)
{
	if (isVariable())
	{
		if (vars.count(*f_name) != 0)
			*f_name = vars.find(*f_name)->second;
	}
	
	if (f_op1 != 0)
		f_op1->rename_vars(vars);
	
	if (f_op2 != 0)
		f_op2->rename_vars(vars);
}

template <class T>
void eo<T>::operator=(const eo<T>& op)
{
	delete f_op1;
	f_op1 = 0;
	delete f_op2;
	f_op2 = 0;
	f_op = kOP_NONE;
	delete f_name;
	f_name = 0;
	delete f_value;
	f_value = 0;
	
	if (op.f_op1 != 0)
		f_op1 = new eo<T>(*(op.f_op1));
	if (op.f_op2 != 0)
		f_op2 = new eo<T>(*(op.f_op2));
	f_op = op.f_op;
	if (op.f_name != 0)
		f_name = new std::wstring(*(op.f_name));
	if (op.f_value != 0)
		f_value = new T(*(op.f_value));
}

template <class T>
eo<T> eo<T>::operator+(const eo<T>& op) const
{
	return eo<T>((*this), op, kOP_PLUS);
}

template <class T>
eo<T> eo<T>::operator-(const eo<T>& op) const
{
	return eo<T>((*this), op, kOP_MINUS);
}

template <class T>
eo<T> eo<T>::operator*(const eo<T>& op) const
{
	return eo<T>((*this), op, kOP_MULT);
}

template <class T>
eo<T> eo<T>::operator/(const eo<T>& op) const
{
	return eo<T>((*this), op, kOP_DIV);
}

template <class T>
eo<T> eo<T>::sqrt() const
{
	return eo<T>((*this), kOP_SQRT);
}

template <class T>
eo<T> sqrt(const eo<T>& op)
{
	return op.sqrt();
}

namespace std
{
	template <class T>
	eo<T> sqrt(const eo<T>& op)
	{
		return op.sqrt();
	}
}

template <class T>
eo<T>::operator T() const throw (std::runtime_error)
{
	if (isOperation()) // operation
		switch (f_op)
		{
			case kOP_PLUS:
				return (T)(*f_op1) + (T)(*f_op2);
			case kOP_MINUS:
				return (T)(*f_op1) - (T)(*f_op2);
			case kOP_MULT:
				return (T)(*f_op1) * (T)(*f_op2);
			case kOP_DIV:
				return (T)(*f_op1) / (T)(*f_op2);
			case kOP_SQRT:
				return std::sqrt((T)(*f_op1));
		}
	else if (isVariable()) // variable
	{
		if (f_value == 0)
		{
			std::wstringstream temp;
			temp << "Variable \"" << *f_name << "\" was not assigned.";
			throw std::runtime_error(wstos(temp.str()));
		}
		else
			return *f_value;
	}
	else if (isConstant()) // constant value
		return *f_value;
}

template<class T>
void eo<T>::expression(std::wstringstream& sstream)
{
	if (isOperation()) // operation
	{
		switch (f_op)
		{
			case kOP_PLUS:
			{
				sstream << "(";
				f_op1->expression(sstream);
				sstream << " + ";
				f_op2->expression(sstream);
				sstream << ")";
				break;
			}
			case kOP_MINUS:
			{
				sstream << "(";
				f_op1->expression(sstream);
				sstream << " - ";
				f_op2->expression(sstream);
				sstream << ")";
				break;
			}
			case kOP_MULT:
			{
				sstream << "(";
				f_op1->expression(sstream);
				sstream << " * ";
				f_op2->expression(sstream);
				sstream << ")";
				break;
			}
			case kOP_DIV:
			{
				sstream << "(";
				f_op1->expression(sstream);
				sstream << " / ";
				f_op2->expression(sstream);
				sstream << ")";
				break;
			}
			case kOP_DIV:
			{
				sstream << "sqrt(";
				f_op1->expression(sstream);
				sstream << ")";
				break;
			}
		}
	}
	else if (isVariable()) // variable
	{
		sstream << *f_name;
	}
	else if (isConstant()) // constant value
		sstream << *f_value;
}

template<class T>
void eo<T>::simplify(const bool& considerAssignedVarsConst)
{
	bool changed = true;
	
	while (changed)
	{
		changed = false;
		if (considerAssignedVarsConst)
			simplifyFull(changed);
		else
			simplifyPartial(changed);
	}
}

template<class T>
bool eo<T>::isOperation() const
{
	if (isUnaryOperation() || isBinaryOperation())
		return true;
	else
		return false;
}

template<class T>
bool eo<T>::isUnaryOperation() const
{
	if ((f_op1 != 0) && (f_op2 == 0) && (f_op != kOP_NONE))
		return true;
	else
		return false;
}

template<class T>
bool eo<T>::isBinaryOperation() const
{
	if ((f_op1 != 0) && (f_op2 != 0) && (f_op != kOP_NONE))
		return true;
	else
		return false;
}

template<class T>
bool eo<T>::isVariable() const
{
	if (f_name != 0)
		return true;
	else
		return false;
}

template<class T>
bool eo<T>::isAssignedVariable() const
{
	if (isVariable() && (f_value != 0))
		return true;
	else
		return false;
}

template<class T>
bool eo<T>::isConstant() const
{
	if (!isVariable() && (f_value != 0))
		return true;
	else
		return false;
}

template<class T>
void eo<T>::simplifyFull(bool& changed)
{
	if (f_op1 != 0)
		f_op1->simplifyFull(changed);
	if (f_op2 != 0)
		f_op2->simplifyFull(changed);

	if (isOperation())
	{		
		try
		{
			f_value = new (T)(*this);
			delete f_op1;
			f_op1 = 0;
			delete f_op2;
			f_op2 = 0;
			f_op = kOP_NONE;
			delete f_name;
			f_name = 0;		
			changed = true;
		}
		catch (...)
		{
		}
	}
}

template<class T>
void eo<T>::simplifyPartial(bool& changed)
{
	if (f_op1 != 0)
		f_op1->simplifyPartial(changed);
	if (f_op2 != 0)
		f_op2->simplifyPartial(changed);

	if ((isUnaryOperation() && f_op1->isConstant()) || (isBinaryOperation() && f_op1->isConstant() && f_op2->isConstant()))
	{
		try
		{
			f_value = new (T)(*this);
			delete f_op1;
			f_op1 = 0;
			delete f_op2;
			f_op2 = 0;
			f_op = kOP_NONE;
			delete f_name;
			f_name = 0;		
			changed = true;
		}
		catch (...)
		{
		}
	}
}
#endif
