/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _HH_MATRIX_TEMPLATE
#define _HH_MATRIX_TEMPLATE 1

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#include <stdexcept>
#include <cmath>
#include <map>
#include <functional>
#include <set>
#include <iostream>
#include <complex>
#include "matrix_position.h"

#ifdef HAVE_LAPACK
	#include "clapack.h"
#endif

#ifdef HAVE_LAPACK
void indexArray_from_permutationArray(const unsigned int& dimension, __CLPK_integer* permutationArray, unsigned int* indexArray);
#endif

inline double posDoubleRepresentation(const double& op) { return std::abs(op); };
inline double posDoubleRepresentation(const std::complex<double>& op) { return std::abs(op); };
inline double posDoubleRepresentation(const std::complex<long double>& op) { return static_cast<double>(std::abs(op)); };

template <class T>
class matrix
{
	private:
		T** _matrix_TPtrPtr;
		unsigned int _rowCount_uint;
		unsigned int _columnCount_uint;
		unsigned int* _rowPermutation_uintPtr;
		unsigned int* _columnPermutation_uintPtr;
		
		void allocate(const unsigned int& rows, const unsigned int& columns) throw (std::runtime_error);
		void deallocate(void);
	
	public:
		enum fillPattern_enum
		{
			all_fillPattern,
			diagonal_fillPattern
		};

		matrix(void);
		matrix(const unsigned int& rows, const unsigned int& columns = 1, const T& value = static_cast<T>(0.0), const fillPattern_enum& fillPattern = all_fillPattern);
		matrix(const matrix<T>& op) throw (std::runtime_error);
		matrix(const matrix<std::complex<T> >& op) throw (std::runtime_error);
		~matrix(void) { this->deallocate(); };
		void operator= (const matrix<T>& op);
		T& operator() (const unsigned int& row, const unsigned int& column = 0) const throw (std::runtime_error);
		matrix<T> operator* (const matrix<T>& op1) throw (std::runtime_error);
		matrix<T> operator* (const T& op1);
		matrix<T> operator/ (const T& op1);
		matrix<T> operator+ (const T& op1);
		unsigned int rows(void) const;
		unsigned int columns(void) const;
		void fill(const T& value, const fillPattern_enum& fillPattern);
		matrix<T> transpose(void) const;
		matrix<T> permutate(void) throw (std::runtime_error);
		matrix<T> ludecomp(void) throw (std::runtime_error);
		matrix<T> fbsubs(const matrix<T>& b) throw (std::runtime_error);
		matrix<T> inv(void);
		T det(void);
		matrix<T> abs(void);
		std::ostream& output(std::ostream& os) const;
};

#ifdef HAVE_LAPACK
class matrixNumber
{
	private:
		__CLPK_doublecomplex* _number;
	
	public:
		matrixNumber(__CLPK_doublecomplex* const number);
		operator std::complex<double>();
		void operator= (const matrixNumber& number);
		void operator= (const std::complex<double>& number);
		void operator+= (const matrixNumber& number);
		void operator-= (const matrixNumber& number);
		void operator*= (const matrixNumber& number);
		void operator/= (const matrixNumber& number);
		int operator== (const matrixNumber& number) const;
		int operator!= (const matrixNumber& number) const;
		void operator+= (const std::complex<double>& number);
		void operator-= (const std::complex<double>& number);
		void operator*= (const std::complex<double>& number);
		void operator/= (const std::complex<double>& number);
		int operator== (const std::complex<double>& number) const;
		int operator!= (const std::complex<double>& number) const;
		std::complex<double> operator+ (const matrixNumber& number) const;
		std::complex<double> operator- (const matrixNumber& number) const;
		std::complex<double> operator* (const matrixNumber& number) const;
		std::complex<double> operator/ (const matrixNumber& number) const;
		std::complex<double> operator+ (const std::complex<double>& number) const;
		std::complex<double> operator- (const std::complex<double>& number) const;
		std::complex<double> operator* (const std::complex<double>& number) const;
		std::complex<double> operator/ (const std::complex<double>& number) const;
		std::complex<double> operator+ (const double& number) const;
		std::complex<double> operator- (const double& number) const;
		std::complex<double> operator* (const double& number) const;
		std::complex<double> operator/ (const double& number) const;
		std::ostream& output(std::ostream& os) const;
		std::complex<double> conj() const;
		double abs() const;
		double real() const;
		double imag() const;
};

template <>
class matrix<std::complex<double> >
{
	private:
		__CLPK_doublecomplex* _data;
		__CLPK_integer* _ipiv;
		unsigned int* _rowPermutation_uintPtr;
		unsigned int _rowCount_uint;
		unsigned int _columnCount_uint;
		
		void allocate(const unsigned int& rows, const unsigned int& columns) throw (std::runtime_error);
		void deallocate(void);

	public:
		enum fillPattern_enum
		{
			all_fillPattern,
			diagonal_fillPattern
		};
		
		matrix(void);
		matrix(const unsigned int& rows, const unsigned int& columns = 1, const std::complex<double>& value = 0.0, const fillPattern_enum& fillPattern = all_fillPattern);
		matrix(const matrix<std::complex<double> >& op) throw (std::runtime_error);
		~matrix(void);
		void operator= (const matrix<std::complex<double> >& op);
		matrixNumber operator() (const unsigned int& row, const unsigned int& column = 0) const throw (std::runtime_error);
 		matrix<std::complex<double> > operator* (const matrix<std::complex<double> >& op1) throw (std::runtime_error);
		matrix<std::complex<double> > operator* (const std::complex<double>& op1);
		matrix<std::complex<double> > operator+ (const std::complex<double>& op1);
		unsigned int rows(void) const;
		unsigned int columns(void) const;
 		void fill(const std::complex<double>& value, const fillPattern_enum& fillPattern);
 		matrix<std::complex<double> > transpose(void) const;
// 		matrix<T> permutate(void) throw (std::runtime_error);
 		matrix<std::complex<double> > ludecomp(void) throw (std::runtime_error);
 		matrix<std::complex<double> > fbsubs(const matrix<std::complex<double> >& b) throw (std::runtime_error);
 		matrix<std::complex<double> > inv(void);
		std::complex<double> det(void);
		matrix<std::complex<double> > abs(void);
		std::ostream& output(std::ostream& os) const;
};

std::complex<double> operator+ (const std::complex<double> op1, const matrixNumber& op2);
std::complex<double> operator- (const std::complex<double> op1, const matrixNumber& op2);
std::complex<double> operator* (const std::complex<double> op1, const matrixNumber& op2);
std::complex<double> operator/ (const std::complex<double> op1, const matrixNumber& op2);
std::complex<double> operator+ (const double op1, const matrixNumber& op2);
std::complex<double> operator- (const double op1, const matrixNumber& op2);
std::complex<double> operator* (const double op1, const matrixNumber& op2);
std::complex<double> operator/ (const double op1, const matrixNumber& op2);
std::ostream& operator<< (std::ostream& os, const matrixNumber& number);

namespace std
{
	complex<double> conj(const matrixNumber& number);
	double abs(const matrixNumber& number);
	double real(const matrixNumber& number);
	double imag(const matrixNumber& number);
	void operator*= (complex<double>& op1, const matrixNumber& number);
}
#endif

template<class T>
matrix<T>::matrix(void)
{
	this->_rowCount_uint = 0;
	this->_columnCount_uint = 0;
	this->_matrix_TPtrPtr = NULL;
	this->_rowPermutation_uintPtr = NULL;
	this->_columnPermutation_uintPtr = NULL;	
}

template<class T>
matrix<T>::matrix(const unsigned int& rows, const unsigned int& columns, const T& value, const fillPattern_enum& fillPattern)
{
	this->allocate(rows, columns);
	this->fill(value, fillPattern);
}

template<class T>
std::ostream& operator<< (std::ostream& os, const matrix<T>& op) { return op.output(os); };

template<class T>
void matrix<T>::allocate(const unsigned int& rows, const unsigned int& columns) throw (std::runtime_error)
{
	if ((rows > 0) && (columns > 0))
	{
		this->_rowCount_uint = rows;
		this->_columnCount_uint = columns;
		this->_matrix_TPtrPtr = new T*[this->_rowCount_uint];
		this->_rowPermutation_uintPtr = new unsigned int[this->_rowCount_uint];
		this->_columnPermutation_uintPtr = new unsigned int[this->_columnCount_uint];
		for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		{
			this->_matrix_TPtrPtr[rowCounter_uint] = new T[this->_columnCount_uint];
			this->_rowPermutation_uintPtr[rowCounter_uint] = rowCounter_uint;
		}
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			this->_columnPermutation_uintPtr[columnCounter_uint] = columnCounter_uint;
		}
	}
	else
	{
		this->_rowCount_uint = 0;
		this->_columnCount_uint = 0;
		this->_matrix_TPtrPtr = NULL;
		this->_rowPermutation_uintPtr = NULL;
		this->_columnPermutation_uintPtr = NULL;
	}
}

template<class T>
matrix<T>::matrix(const matrix<T>& op) throw (std::runtime_error)
{
	this->allocate(op._rowCount_uint, op._columnCount_uint);
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			(*this)(rowCounter_uint, columnCounter_uint) = op(rowCounter_uint, columnCounter_uint);
		}
	}
}

template<class T>
matrix<T>::matrix(const matrix<std::complex<T> >& op) throw (std::runtime_error)
{
	this->allocate(op.rows(), op.columns());
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			(*this)(rowCounter_uint, columnCounter_uint) = std::real(op(rowCounter_uint, columnCounter_uint));
			if (std::imag(op(rowCounter_uint, columnCounter_uint)) != static_cast<T>(0.0))
				throw std::runtime_error("matrix<T>::matrix: Discarding imaginary part.");
		}
	}
}

template<class T>
void matrix<T>::deallocate(void)
{
	if ((this->_rowCount_uint > 0) && (this->_matrix_TPtrPtr != NULL))
	{
		for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		{
			delete[] this->_matrix_TPtrPtr[rowCounter_uint];
		}
		delete[] this->_matrix_TPtrPtr;
		this->_matrix_TPtrPtr = NULL;
		this->_rowCount_uint = 0;
		this->_columnCount_uint = 0;
		delete[] this->_rowPermutation_uintPtr;
		this->_rowPermutation_uintPtr = NULL;
		delete[] this->_columnPermutation_uintPtr;
		this->_columnPermutation_uintPtr = NULL;
	}
}

template<class T>
void matrix<T>::operator= (const matrix<T>& op)
{
	this->deallocate();
	this->allocate(op._rowCount_uint, op._columnCount_uint);
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			(*this)(rowCounter_uint, columnCounter_uint) = op(rowCounter_uint, columnCounter_uint);
		}
	}	
}

template<class T>
T& matrix<T>::operator() (const unsigned int& row, const unsigned int& column) const throw (std::runtime_error)
{
	if ((row < this->_rowCount_uint) && (column < this->_columnCount_uint))
		return this->_matrix_TPtrPtr[this->_rowPermutation_uintPtr[row]][this->_columnPermutation_uintPtr[column]];
	else
		throw std::runtime_error("matrix operator() indices out of range");
}

template<class T>
matrix<T> matrix<T>::operator* (const matrix<T>& op1) throw (std::runtime_error)
{
	if (this->_columnCount_uint == op1._rowCount_uint)
	{
		matrix<T> temp(this->_rowCount_uint, op1._columnCount_uint);
		for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
		{
			for (unsigned int columnCounter_uint = 0; columnCounter_uint < op1._columnCount_uint; columnCounter_uint++)
			{
				for (unsigned int innerCounter_uint = 0; innerCounter_uint < this->_columnCount_uint; innerCounter_uint++)
				{
					temp(lineCounter_uint, columnCounter_uint) += (*this)(lineCounter_uint, innerCounter_uint) * op1(innerCounter_uint, columnCounter_uint);
				}
			}
		}
		return temp;
	}
	else
	{
		throw std::runtime_error("matrix operator* inner dimensions do not agree");
	}
}

template<class T>
matrix<T> matrix<T>::operator* (const T& op1)
{
	matrix<T> temp(*this);
	for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			temp(lineCounter_uint, columnCounter_uint) *= op1;
		}
	}
	return temp;	
}

template<class T>
matrix<T> matrix<T>::operator/ (const T& op1)
{
	matrix<T> temp(*this);
	for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			temp(lineCounter_uint, columnCounter_uint) /= op1;
		}
	}
	return temp;	
}

template<class T>
matrix<T> matrix<T>::operator+ (const T& op1)
{
	matrix<T> temp(*this);
	for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			temp(lineCounter_uint, columnCounter_uint) += op1;
		}
	}
	return temp;	
}

template<class T>
unsigned int matrix<T>::rows(void) const
{
	return this->_rowCount_uint;
}

template<class T>
unsigned int matrix<T>::columns(void) const
{
	return this->_columnCount_uint;
}

template<class T>
void matrix<T>::fill(const T& value, const fillPattern_enum& fillPattern)
{
	switch (fillPattern)
	{
		case all_fillPattern:
			for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
			{
				for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
				{
					this->_matrix_TPtrPtr[rowCounter_uint][columnCounter_uint] = value;
				}
			}
			break;
		case diagonal_fillPattern:
			for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
			{
				for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
				{
					if (rowCounter_uint == columnCounter_uint)
					{
						this->_matrix_TPtrPtr[rowCounter_uint][columnCounter_uint] = value;
					}
					else
					{
						this->_matrix_TPtrPtr[rowCounter_uint][columnCounter_uint] = static_cast<T>(0.0);
					}
				}
			}
			break;
	}
}

template<class T>
matrix<T> matrix<T>::transpose(void) const
{
	matrix<T> output_matrix(this->_columnCount_uint, this->_rowCount_uint);
	
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < output_matrix._rowCount_uint; rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < output_matrix._columnCount_uint; columnCounter_uint++)
			output_matrix(rowCounter_uint, columnCounter_uint) = (*this)(columnCounter_uint, rowCounter_uint);
	
	return output_matrix;
}

template<class T>
matrix<T> matrix<T>::permutate(void) throw (std::runtime_error)
{
	if (this->_rowCount_uint != this->_columnCount_uint)
	{
		throw std::runtime_error("matrix<T>::permutate: This function is only available for square matrices.");
	}
	// Generate full permutation arrays
	matrix<T> temp_matrixT(*this);
	unsigned int valueCounter_uint = 0;
	std::set< unsigned int > availableRows_setUint, availableColumns_setUint;
	std::multimap< double, matrix_position, std::greater< double > > map;
	
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
		{
			map.insert(std::pair<double, matrix_position>(posDoubleRepresentation((*this)(rowCounter_uint, columnCounter_uint)), matrix_position(rowCounter_uint, columnCounter_uint)));
		}
		availableRows_setUint.insert(rowCounter_uint);
		availableColumns_setUint.insert(rowCounter_uint);
	}

	for (std::multimap< double, matrix_position, std::greater< double > >::iterator it = map.begin(); (it != map.end()) && (valueCounter_uint < this->_rowCount_uint); it++)
	{
		std::set< unsigned int >::iterator row_iterator = availableRows_setUint.find(((*it).second).get_row());
		std::set< unsigned int >::iterator column_iterator = availableColumns_setUint.find(((*it).second).get_column());
		
		if ((row_iterator != availableRows_setUint.end()) && (column_iterator != availableColumns_setUint.end())) {
			temp_matrixT._rowPermutation_uintPtr[valueCounter_uint] = ((*it).second).get_row();
			temp_matrixT._columnPermutation_uintPtr[valueCounter_uint] = ((*it).second).get_column();
			valueCounter_uint += 1;
			availableRows_setUint.erase(row_iterator);
			availableColumns_setUint.erase(column_iterator);
		}
	}
	
	return temp_matrixT;
}

template<class T>
matrix<T> matrix<T>::ludecomp(void) throw (std::runtime_error)
{
	matrix<T> temp_matrixT = this->permutate();
	unsigned int size_uint(temp_matrixT._rowCount_uint);
	
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < size_uint; rowCounter_uint++)
	{
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < size_uint; columnCounter_uint++)
		{
			T tempSum_T(static_cast<T>(0.0));
			if (rowCounter_uint == 0)
			{
//				LU_matrix[counter_row][counter_column] = temp_matrix[counter_row][counter_column];
			}
			else
			{
				if (columnCounter_uint >= rowCounter_uint)
				{
					for (unsigned int sumCounter_uint = 0; sumCounter_uint < rowCounter_uint; sumCounter_uint++)
					{
						if (sumCounter_uint == rowCounter_uint)
						{
//							temp_sum += LU_matrix[counter_sum][counter_column];
							tempSum_T += temp_matrixT(sumCounter_uint, columnCounter_uint);
						}
						else
						{
//							temp_sum += LU_matrix[counter_row][counter_sum] * LU_matrix[counter_sum][counter_column];
							tempSum_T += temp_matrixT(rowCounter_uint, sumCounter_uint) * temp_matrixT(sumCounter_uint, columnCounter_uint);
						}
					}
//					LU_matrix[counter_row][counter_column] = temp_matrix[counter_row][counter_column] - temp_sum;
					temp_matrixT(rowCounter_uint, columnCounter_uint) = temp_matrixT(rowCounter_uint, columnCounter_uint) - tempSum_T;
				}
				else
				{
					for (int sumCounter_uint = 0; sumCounter_uint < columnCounter_uint; sumCounter_uint++)
					{
						if (sumCounter_uint == rowCounter_uint)
						{
//							temp_sum += LU_matrix[counter_sum][counter_column];
							tempSum_T += temp_matrixT(sumCounter_uint, columnCounter_uint);
						}
						else
						{
//							temp_sum += LU_matrix[counter_row][counter_sum] * LU_matrix[counter_sum][counter_column];
							tempSum_T += temp_matrixT(rowCounter_uint, sumCounter_uint) * temp_matrixT(sumCounter_uint, columnCounter_uint);
						}
					}
//					LU_matrix[counter_row][counter_column] = (temp_matrix[counter_row][counter_column] - temp_sum) / LU_matrix[counter_column][counter_column];
					temp_matrixT(rowCounter_uint, columnCounter_uint) = (temp_matrixT(rowCounter_uint, columnCounter_uint) - tempSum_T) / temp_matrixT(columnCounter_uint, columnCounter_uint);
				}
			}
		}
	}
	return temp_matrixT;
}

template<class T>
matrix<T> matrix<T>::fbsubs(const matrix<T>& b) throw (std::runtime_error)
{
	if (this->_rowCount_uint != b._rowCount_uint)
	{
		throw std::runtime_error("matrix<T>::fbsubs: The number of rows in both matrices have to agree.");
	}
	matrix<T> y_matrixT(b);
	matrix<T> result_matrixT(b);
	
	// Permutate result
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
	{
//		y[counter] = b[row_permutation[counter]];
		y_matrixT._rowPermutation_uintPtr[rowCounter_uint] = this->_rowPermutation_uintPtr[rowCounter_uint];
	}
	
	// Forward Substitution
	for (unsigned int resultColumnCounter_uint = 0; resultColumnCounter_uint < result_matrixT._columnCount_uint; resultColumnCounter_uint++)
	{
		for (unsigned int counter1_uint = 0; counter1_uint < y_matrixT._rowCount_uint; counter1_uint++)
		{
			for (unsigned int counter2_uint = 0; counter2_uint < counter1_uint; counter2_uint++) {
//					y[counter1] -= LU_matrix[counter1][counter2] * y[counter2];
					y_matrixT(counter1_uint, resultColumnCounter_uint) -= (*this)(counter1_uint, counter2_uint) * y_matrixT(counter2_uint, resultColumnCounter_uint);
			}
		}
		
		// Backward Substitution and permutation
		for (unsigned int counter1_uint = (result_matrixT._rowCount_uint - 1); (counter1_uint + 1) > 0; counter1_uint--)
		{
//			result[column_permutation[counter1]] = y[counter1];
			result_matrixT(this->_columnPermutation_uintPtr[counter1_uint], resultColumnCounter_uint) = y_matrixT(counter1_uint, resultColumnCounter_uint);
			for (unsigned int counter2_uint = (result_matrixT._rowCount_uint - 1); (counter2_uint + 1) > counter1_uint; counter2_uint--)
			{
				if (counter1_uint == counter2_uint)
				{
//					result[column_permutation[counter1]] /= LU_matrix[counter1][counter2];
					result_matrixT(this->_columnPermutation_uintPtr[counter1_uint], resultColumnCounter_uint) /= (*this)(counter1_uint, counter2_uint);
				}
				else
				{
//					result[column_permutation[counter1]] -= result[column_permutation[counter2]] * LU_matrix[counter1][counter2];
					result_matrixT(this->_columnPermutation_uintPtr[counter1_uint], resultColumnCounter_uint) -= result_matrixT(this->_columnPermutation_uintPtr[counter2_uint], resultColumnCounter_uint) * (*this)(counter1_uint, counter2_uint);
				}
			}
		}
	}
	
	return result_matrixT;
}

template<class T>
matrix<T> matrix<T>::inv(void)
{
	matrix<T> lu_matrixT(this->ludecomp());
	matrix<T> eye_matrixT(*this);
	eye_matrixT.fill(static_cast<T>(1.0), matrix<T>::diagonal_fillPattern);
	return lu_matrixT.fbsubs(eye_matrixT);
}

template<class T>
T matrix<T>::det(void)
{
	matrix<T> LU_matrix(this->ludecomp());
	T determinant;

	unsigned int permutation_sign = 0;
	
	for (unsigned int counter = 0; counter < LU_matrix.rows(); counter++)
	{
		if (counter == 0)
		{
			determinant = LU_matrix(counter, counter);
		} else {
			determinant *= LU_matrix(counter, counter);
		}

		if ((counter + 1) < LU_matrix.rows()) {
			for (unsigned int permutation_counter = (counter + 1); permutation_counter < LU_matrix.rows(); permutation_counter++) {
				if (LU_matrix._rowPermutation_uintPtr[counter] > LU_matrix._rowPermutation_uintPtr[permutation_counter]) {
					permutation_sign++;
				}
				if (LU_matrix._columnPermutation_uintPtr[counter] > LU_matrix._columnPermutation_uintPtr[permutation_counter]) {
					permutation_sign++;
				}
			}
		}
//		if (row_permutation[counter] != counter) {
//			row_sign++;
//		}
//		if (column_permutation[counter] != counter) {
//			column_sign++;
//		}
	}

	if ((permutation_sign) & 1) {
		return determinant * static_cast<T>(-1.0);
	} else {
		return determinant;
	}
}

template<class T>
matrix<T> matrix<T>::abs(void)
{
	matrix<T> output_matrix(*this);
	
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			output_matrix(rowCounter_uint, columnCounter_uint) = std::abs(output_matrix(rowCounter_uint, columnCounter_uint));
	
	return output_matrix;
}

template<class T>
std::ostream& matrix<T>::output(std::ostream& os) const
{
	if ((this->_rowCount_uint > 0) && (this->_matrix_TPtrPtr != NULL))
	{
		for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		{
			for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			{
				os << (*this)(rowCounter_uint, columnCounter_uint) << "\t";
			}
			os << "\n";
		}
	}
	return os;
}
#endif

